/***************************************************************************
 *   (C) 2007 by Tomas Mandys <tomas.mandys at 2p dot cz>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License           *
 *   version 2.1 as published by the Free Software Foundation;             *
 *                                                                         *
 ***************************************************************************/


/* 
 *Shared library add-on to iptables to add RTPPROXY target support.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <netinet/in.h>  /* put it above xtables.h where is made a IPPROTO_xx magic */
#ifdef IPT_RTPPROXY_IPTABLES_API
#include <iptables.h>
#undef IPT_RTPPROXY_IPTABLES_API
#else
#include <xtables.h>
#endif
#include <net/if.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <stddef.h>
#include <errno.h>
//#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter/xt_RTPPROXY.h>
#include <unistd.h>
#include <rpc/pmap_clnt.h>
#include "libipt_RTPPROXY_rpc.h"

#ifdef _IPTABLES_COMMON_H   
/* old iptables API, it uses iptables_common.h (instead of xtables.h) included from iptables.h */
/* #ifndef XTABLES_VERSION ... optional test */
#define XTABLES_VERSION IPTABLES_VERSION
#define IPT_RTPPROXY_IPTABLES_API 1
#ifndef XT_ALIGN
#define XT_ALIGN IPT_ALIGN
#endif
#endif
//#include "../config.h"

/* Function which prints out usage message. */
static void help(void) {
	printf(
"RTPPROXY target v%s options:\n"
"  --addr     ip                 Address of RTP proxy switchboard\n"
"  --port     port               Start port number of switchboard\n"
"  --max-sess num                Max number of sessions\n"
"  --resurrection-timeout sec    Resurrection timeout\n"
"  --ha-state state              0..active(default), 1..standby\n"
"\n",
XTABLES_VERSION);
}

static struct option opts[] = {
	{ .name = "addr",                 .has_arg = 1, .flag = 0, .val = '1' },
	{ .name = "port",                 .has_arg = 1, .flag = 0, .val = '2' },
	{ .name = "max-sess",             .has_arg = 1, .flag = 0, .val = '3' },
	{ .name = "resurrection-timeout", .has_arg = 1, .flag = 0, .val = '4' },
	{ .name = "ha-state",             .has_arg = 1, .flag = 0, .val = '5' },
	{ .name = 0 }
};

#define DEFAULT_RESURRECTION_TIMEOUT 40

/* Initialize the target. */
static void init(
#ifdef IPT_RTPPROXY_IPTABLES_API
	struct ipt_entry_target *t, unsigned int *nfcache
#else
	struct xt_entry_target *t
#endif
) {
	struct xt_rtpproxy_target_info *info = 
		(struct xt_rtpproxy_target_info*)t->data;

	memset(info, 0, sizeof(*info));
	info->resurrection_timeout = DEFAULT_RESURRECTION_TIMEOUT;
	info->ha_state = xt_rtpproxy_ACTIVE;
}

#define IPT_RTPPROXY_OPT_ADDR      0x01
#define IPT_RTPPROXY_OPT_PORT      0x02
#define IPT_RTPPROXY_OPT_MAX_SESS    0x04
#define IPT_RTPPROXY_OPT_RESURRECTION_TIMEOUT 0x08
#define IPT_RTPPROXY_OPT_HA_STATE 0x10

#define ADJUST_U_INT16(x) ((x)>0xFFFF?0xFFFF:(x))

/* Function which parses command options; returns true if it
   ate an option */
static int parse(
#ifdef IPT_RTPPROXY_IPTABLES_API
	int c, char **argv, int invert, unsigned int *flags,
	const struct ipt_entry *entry,
	struct ipt_entry_target **target
#else
	int c, char **argv, int invert, unsigned int *flags,
	const void *entry,
	struct xt_entry_target **target
#endif
	) {
	struct xt_rtpproxy_target_info *info = 
		(struct xt_rtpproxy_target_info*)(*target)->data;
	switch (c) {
	case '1':
		if (*flags & IPT_RTPPROXY_OPT_ADDR)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --addr twice");

		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --addr");

		if (!inet_aton(optarg, (struct in_addr*) &info->addr.ip) ) 
			exit_error(PARAMETER_PROBLEM,
				   "Invalid IP address %s",
				   optarg);
  
		*flags |= IPT_RTPPROXY_OPT_ADDR;
		break;

	case '2':
		if (*flags & IPT_RTPPROXY_OPT_PORT)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --port twice");

		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --port");

		if ( (info->addr.port = ADJUST_U_INT16(atol(optarg))) == 0)
			exit_error(PARAMETER_PROBLEM,
				   "Invalid number %s",
				   optarg);

		*flags |= IPT_RTPPROXY_OPT_PORT;
		break;

	case '3':
		if (*flags & IPT_RTPPROXY_OPT_MAX_SESS)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --max-sess twice");

		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --max-sess");

		if ( (info->max_sess = ADJUST_U_INT16(atol(optarg))) == 0)
			exit_error(PARAMETER_PROBLEM,
				   "Invalid number %s",
				   optarg);

		*flags |= IPT_RTPPROXY_OPT_MAX_SESS;
		break;

	case '4':
		if (*flags & IPT_RTPPROXY_OPT_RESURRECTION_TIMEOUT)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --resurrection-timeout twice");

		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --resurrection-timeout");

		if ( (info->resurrection_timeout = ADJUST_U_INT16(atol(optarg))) == 0)
			exit_error(PARAMETER_PROBLEM,
				   "Invalid number %s",
				   optarg);

		*flags |= IPT_RTPPROXY_OPT_RESURRECTION_TIMEOUT;
		break;

	case '5':
		if (*flags & IPT_RTPPROXY_OPT_HA_STATE)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --ha-state twice");

		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --ha-state");

		if (strlen(optarg) != 1 || *optarg < '0' || *optarg > '1') 
			exit_error(PARAMETER_PROBLEM,
				   "Invalid number %s",
				   optarg);
		
		info->ha_state = *optarg - '0';

		*flags |= IPT_RTPPROXY_OPT_HA_STATE;
		break;

	default:
		return 0;
	}
	info->libipt_flags = *flags;
	return 1;
}


static void final_check(unsigned int flags) {
	/* Grrr... we don't know struct ipt_entry_target* so we cannot make aggregated test here but in ipt_checkentry :-( */
	if ((flags & (IPT_RTPPROXY_OPT_ADDR|IPT_RTPPROXY_OPT_PORT|IPT_RTPPROXY_OPT_MAX_SESS)) != (IPT_RTPPROXY_OPT_ADDR|IPT_RTPPROXY_OPT_PORT|IPT_RTPPROXY_OPT_MAX_SESS)) {
		exit_error(PARAMETER_PROBLEM,
		           "RTPPROXY target: addr, port and max-sess options are required");
	}
}


/* Prints out the targinfo. */
static void print(
#ifdef IPT_RTPPROXY_IPTABLES_API
	const struct ipt_ip *ip,
	const struct ipt_entry_target *target,
	int numeric
#else
	const void *ip,
	const struct xt_entry_target *target,
	int numeric
#endif
      ) {
	const struct xt_rtpproxy_target_info *info
		= (const struct xt_rtpproxy_target_info *)target->data;
	struct in_addr ip_a = { info->addr.ip };

	printf("RTPPROXY ");
	printf("switchboard:%s:[%d-%d] ", inet_ntoa(ip_a), info->addr.port, info->addr.port + 2*info->max_sess);
	printf("max-sess:%d ", info->max_sess);
	printf("resurrection-timeout:%d ", info->resurrection_timeout);
	printf("ha-state:%d", info->ha_state);
}

static void save(
#ifdef IPT_RTPPROXY_IPTABLES_API
	const struct ipt_ip *ip, 
	const struct ipt_entry_target *target
#else
	const void *ip, 
	const struct xt_entry_target *target
#endif
      ) {
	const struct xt_rtpproxy_target_info *info
		= (const struct xt_rtpproxy_target_info *)target->data;

	if (info->libipt_flags & IPT_RTPPROXY_OPT_ADDR) {
		struct in_addr ip2= { info->addr.ip };
		printf("--addr %s ", inet_ntoa(ip2));
	}

	if (info->libipt_flags & IPT_RTPPROXY_OPT_PORT) {
		printf("--port %d ", info->addr.port);
	}

	if (info->libipt_flags & IPT_RTPPROXY_OPT_MAX_SESS) {
		printf("--max-sess %d ", info->max_sess);
	}

	if (info->libipt_flags & IPT_RTPPROXY_OPT_RESURRECTION_TIMEOUT) {
		printf("--resurrection-timeout %d ", info->resurrection_timeout);
	}
	
	if (info->libipt_flags & IPT_RTPPROXY_OPT_HA_STATE) {
		printf("--ha-state %d ", info->ha_state);
	}
}


static struct 
#ifdef IPT_RTPPROXY_IPTABLES_API
	iptables_target rtpproxy 
#else
	xtables_target rtpproxy 
#endif

= { 
	.version	= XTABLES_VERSION,
	.name		= "RTPPROXY",
	.size		= XT_ALIGN(sizeof(struct xt_rtpproxy_target_info)),
	.userspacesize	= offsetof(struct xt_rtpproxy_target_info, resurrection_timeout),
	.help		= &help,
	.init		= &init,
	.parse		= &parse,
	.final_check	= &final_check,
	.print		= &print,
	.save		= &save,
	.extra_opts	= opts
};

#ifdef IPT_RTPPROXY_IPTABLES_API
void __attribute((constructor)) my_init(void) {
	register_target(&rtpproxy);
}
#else
void _init(void) {
	xtables_register_target(&rtpproxy);
}
#endif

/* libipt RTPPROXY interface functions */
int xt_RTPPROXY_open(struct xt_rtpproxy_handle *handle, enum xt_rtpproxy_connection type, void* params) {
	memset(handle, 0, sizeof(*handle));
	handle->type = type;
	handle->sockfd = -1;
	switch (type) {
		case xt_rtpproxy_LOCAL:
			handle->sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
			if (handle->sockfd < 0) {
				handle->sockfd = -1;
				handle->err_str = "Can't open socket to kernel. Permission denied or ipt_RTPPROXY module not loaded";
				handle->err_no = XT_RTPPROXY_ERR_CANNOT_OPEN_SOCKET;
				return -1;
			}
			break;
		case xt_rtpproxy_REMOTE: {
			struct xt_rtpproxy_connection_rpc_params* p = params;
			struct sockaddr_in addr_in;
			struct timeval timeout;
			handle->sockfd = RPC_ANYSOCK;  /* i.e. -1 */
			handle->u.rpc.total_timeout = p->total_timeout?p->total_timeout:XT_RTPPROXY_RPC_DEFAULT_TOTAL_TIMEOUT;
			addr_in.sin_family = AF_INET;
			addr_in.sin_addr.s_addr = p->addr;
			addr_in.sin_port = htons(p->port); /* if zero then use portmapper */
			switch (p->proto) {
				case xt_rtpproxy_connection_UDP:
					timeout.tv_sec = (p->udp_retry_timeout?p->udp_retry_timeout:XT_RTPPROXY_RPC_DEFAULT_UDP_REPLY_TIMEOUT) / 1000;
					timeout.tv_usec = ((p->udp_retry_timeout?p->udp_retry_timeout:XT_RTPPROXY_RPC_DEFAULT_UDP_REPLY_TIMEOUT) % 1000) * 1000;
					handle->u.rpc.client = clntudp_bufcreate(&addr_in, IPTRTPPROXY_RPC, IPTRTPPROXY_RPC_VER, timeout, &handle->sockfd, 
						p->request_size?p->request_size:XT_RTPPROXY_RPC_DEFAULT_REQUEST_SIZE,
						p->reply_size?p->reply_size:XT_RTPPROXY_RPC_DEFAULT_REPLY_SIZE
					);
					if (!handle->u.rpc.client) {
						handle->err_str = clnt_spcreateerror("clntudp_bufcreate error");
						handle->err_no = XT_RTPPROXY_ERR_CANNOT_OPEN_SOCKET;
						return -1;
					}
					handle->sockfd = RPC_ANYSOCK; /* clnt_destroy() will close socket */
					break;
				case xt_rtpproxy_connection_TCP:
					handle->u.rpc.client = clnttcp_create(&addr_in, IPTRTPPROXY_RPC, IPTRTPPROXY_RPC_VER, &handle->sockfd, 
						p->request_size?p->request_size:XT_RTPPROXY_RPC_DEFAULT_REQUEST_SIZE,
						p->reply_size?p->reply_size:XT_RTPPROXY_RPC_DEFAULT_REPLY_SIZE
					);
					if (!handle->u.rpc.client) {
						handle->err_str = clnt_spcreateerror("clnttcp_bufcreate error");
						handle->err_no = XT_RTPPROXY_ERR_CANNOT_OPEN_SOCKET;
						return -1;
					}
					handle->sockfd = RPC_ANYSOCK; /* clnt_destroy() will close socket */
					break;
				case xt_rtpproxy_connection_RAW:
					handle->u.rpc.client = clntraw_create(IPTRTPPROXY_RPC, IPTRTPPROXY_RPC_VER);
					if (!handle->u.rpc.client) {
						handle->err_str = clnt_spcreateerror("clntraw_bufcreate error");
						handle->err_no = XT_RTPPROXY_ERR_CANNOT_OPEN_SOCKET;
						return -1;
					}
					break;
			}
			break;
		}
	}
	return 0;
}

void xt_RTPPROXY_close(struct xt_rtpproxy_handle *handle) {
	switch (handle->type) {
		case xt_rtpproxy_LOCAL:
			break;
		case xt_rtpproxy_REMOTE:
			if (handle->u.rpc.client) {
				clnt_destroy(handle->u.rpc.client);
				handle->u.rpc.client = NULL;
			}
			break;
		default:
			;
	}
	if (handle->sockfd >= 0) {
		close(handle->sockfd);
	}
	handle->sockfd = -1;
	if (handle->buf) {
		free(handle->buf);
		handle->buf = NULL;
		handle->buf_size = 0;
	}
}

static inline int adjust_buf(struct xt_rtpproxy_handle *handle, unsigned int size) {
	size += sizeof(struct xt_rtpproxy_sockopt_header);
	if (size > handle->buf_size) {
		void *p;
		p = realloc(handle->buf, size);
		if (!p) {
			handle->err_str = "Can't allocate buffer memory";
			handle->err_no = XT_RTPPROXY_ERR_OUT_OF_MEMORY; 
			return -1;
		}
		handle->buf = p;
		handle->buf_size = size;
	}
	return 0;
}

static int do_sockopt(struct xt_rtpproxy_handle *handle, 
		int cmd,
		void *value, unsigned int len, 
		unsigned int *res_len, /* NULL to setsockopt */
		char *err_str
	) {
	switch (handle->type) {
		case xt_rtpproxy_LOCAL:
			if (res_len) {
				if (*res_len < len) {
					*res_len = len;
				}
				handle->err_no = getsockopt(handle->sockfd, IPPROTO_IP, cmd, value, res_len);
			}
			else {
				handle->err_no = setsockopt(handle->sockfd, IPPROTO_IP, cmd, value, len);
			}
			if (handle->err_no < 0) {
				handle->err_str = err_str;
			}
			break;
		case xt_rtpproxy_REMOTE: {
			struct timeval timeout;
			enum clnt_stat stat;
			xt_rtpproxy_sockopt_params params;
			xt_rtpproxy_sockopt_result res;
			timeout.tv_sec = handle->u.rpc.total_timeout / 1000;
			timeout.tv_usec = (handle->u.rpc.total_timeout % 1000) * 1000;
			params.command = cmd;
			params.data.data_len = len;
			params.data.data_val = value;			
			params.result_size = res_len?*res_len:0;
			memset(&res, 0, sizeof(res)); /* needed ! */
	
			stat = clnt_call(handle->u.rpc.client, res_len?GETSOCKOPT:SETSOCKOPT, 
				(xdrproc_t) xdr_xt_rtpproxy_sockopt_params, (caddr_t) &params,
				(xdrproc_t) xdr_xt_rtpproxy_sockopt_result, (caddr_t) &res, 
				timeout);	
			if (stat != RPC_SUCCESS) {
				handle->err_str = clnt_sperror(handle->u.rpc.client, "rpc");
				handle->err_no = XT_RTPPROXY_ERR_RPC;
				break;
			}
			handle->err_no = res.err_no;
			if (res.err_no < 0) {
				handle->err_str = res.xt_rtpproxy_sockopt_result_u.err_str;
			}
			else if (res_len) {
				if (res.xt_rtpproxy_sockopt_result_u.data.data_len > *res_len) {
					handle->err_no = XT_RTPPROXY_ERR_OUT_OF_MEMORY;
					handle->err_str = "result buffer overflow";
				}
				else {
					*res_len = res.xt_rtpproxy_sockopt_result_u.data.data_len;
					memcpy(value, res.xt_rtpproxy_sockopt_result_u.data.data_val, *res_len);
				}
				clnt_freeres(handle->u.rpc.client, (xdrproc_t) xdr_xt_rtpproxy_sockopt_result, (caddr_t) &res);
			}
			break;
		}
		default:
			handle->err_no = -1;  /* never happens */
	}
	return handle->err_no;
}


#define RESET_BUF(handle, skip_header) \
	do { \
		(handle)->buf_ptr = (handle)->buf + (skip_header)*sizeof(struct xt_rtpproxy_sockopt_header); \
		(handle)->buf_iter = 0; \
	} while (0);

#define COPY_TO_BUF(handle, val) \
	do { \
		memcpy(handle->buf_ptr, &(val), sizeof(val)); \
		handle->buf_ptr += sizeof(val); \
	} while (0);


#define COPY_FROM_BUF(handle, val) \
	do { \
		memcpy(&(val), handle->buf_ptr, sizeof(val)); \
		handle->buf_ptr += sizeof(val); \
	} while (0);

static inline void fill_in_h(struct xt_rtpproxy_handle *handle, u_int32_t flags) {
	struct xt_rtpproxy_sockopt_header *in_h;
	in_h = (void*) handle->buf;
	in_h->magic =  XT_RTPPROXY_SOCKOPT_MAGIC;
	in_h->version = XT_RTPPROXY_SOCKOPT_VERSION;
	in_h->flags = flags;
}

int xt_RTPPROXY_get_config(
	struct xt_rtpproxy_handle *handle,
	xt_rtpproxy_sockopt_count switchboard_count,
	struct xt_rtpproxy_switchboard_id (*switchboards)[],

	enum xt_rtpproxy_out_level out_level,
	unsigned int flags,
	struct xt_rtpproxy_sockopt_out_config *config
) {
	unsigned int sz;
	xt_rtpproxy_sockopt_count out_switchboard_count, out_session_count;

	if (switchboard_count && !switchboards) {
		handle->err_no = XT_RTPPROXY_ERR_PARAM;
		handle->err_str = "Switchboards param is null";
		return -1;
	}
	sz = 3*sizeof(out_session_count) + sizeof(*config);
	if (adjust_buf(handle, 3*sizeof(out_session_count) + 
				   sizeof(*config)) < 0) {
		return -1;
	}

	out_switchboard_count = out_session_count = 0;
	if ((out_level >= xt_rtpproxy_SWITCHBOARDS && (flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS)) ||
		(out_level >= xt_rtpproxy_SESSIONS && (flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS) && switchboard_count > 0)) {
		/* we need know how many switchboards/sessions to expect. The fastest method is return global config */
		fill_in_h(handle, 0);
		RESET_BUF(handle, 1);
		COPY_TO_BUF(handle, out_switchboard_count);
		COPY_TO_BUF(handle, out_session_count);
		COPY_TO_BUF(handle, out_switchboard_count); /* as switchboard_count */
		if (do_sockopt(handle, XT_RTPPROXY_SO_GET_CONFIG, handle->buf, handle->buf_ptr - handle->buf, &sz,
			"XT_RTPPROXY_SO_GET_CONFIG error (1st pass)") < 0) {
			return -1;
		}
		if (flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS) {
			out_switchboard_count = ((struct xt_rtpproxy_sockopt_out_config *)handle->buf)->total.num_switchboard;
		}
		else {
			out_switchboard_count = switchboard_count;
		}
		if ((out_level >= xt_rtpproxy_SESSIONS) && (flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS) && out_switchboard_count > 0) {
			out_session_count = ((struct xt_rtpproxy_sockopt_out_config *)handle->buf)->total.num_pair_session * 2;
		}
	}
	else if (out_level >= xt_rtpproxy_SWITCHBOARDS) {
		out_switchboard_count = switchboard_count;
	}
	sz = 3*sizeof(out_session_count) +
			out_switchboard_count*(sizeof(struct xt_rtpproxy_sockopt_switchboard)+sizeof(out_session_count))+
			out_session_count*(sizeof(struct xt_rtpproxy_sockopt_session))+
			sizeof(*config);
	if (adjust_buf(handle, sz) < 0) {
		return -1;
	}
	fill_in_h(handle, flags);
	RESET_BUF(handle, 1);

	COPY_TO_BUF(handle, out_switchboard_count);
	COPY_TO_BUF(handle, out_session_count);
	if ((flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS) == 0) {
		int i;
		COPY_TO_BUF(handle, switchboard_count);
		for (i=0; i<switchboard_count; i++) {
			COPY_TO_BUF(handle, (*switchboards)[i]);
		}
	}
	if (do_sockopt(handle, XT_RTPPROXY_SO_GET_CONFIG, handle->buf, handle->buf_ptr - handle->buf, &sz,
		"XT_RTPPROXY_SO_GET_CONFIG error (2nd pass)") < 0) {
		return -1;
	}
	RESET_BUF(handle, 0);
	COPY_FROM_BUF(handle, *config);
	handle->buf_iter = 1;
	return 0;
}


int xt_RTPPROXY_read_from_buf(struct xt_rtpproxy_handle *handle, unsigned int sz, void *val) {
	if (!handle->buf_iter) {
		handle->err_str = "cannot read from buffer";
		handle->err_no = XT_RTPPROXY_ERR_PARAM;
		return -1;
	}
	memcpy(val, handle->buf_ptr, sz);
	handle->buf_ptr += sz;
	return 0;
}


int xt_RTPPROXY_alloc_sessions(
	struct xt_rtpproxy_handle *handle,
	xt_rtpproxy_sockopt_count switchboard_count1,  /* if >0 allocate all sessions at one of provided switchboard, otherwise alloc at switchboard provided in sessions */
	struct xt_rtpproxy_switchboard_id (*switchboards1)[],
	xt_rtpproxy_sockopt_count switchboard_count2,
	struct xt_rtpproxy_switchboard_id (*switchboards2)[],
	xt_rtpproxy_sockopt_count session_count,
	struct xt_rtpproxy_sockopt_session (*sessions)[]    /* in-out param */
) {
	int i, j, k;
	unsigned int sz;
	xt_rtpproxy_sockopt_count out_session_count;
	if (!session_count) {
		return 0;
	}
	if ((switchboard_count1 && !switchboard_count2) || (!switchboard_count1 && switchboard_count2)) {
		handle->err_no = XT_RTPPROXY_ERR_PARAM;
		handle->err_str = "Both or none switchboards required";
		return -1;		
	}
	if ((switchboard_count1 && !switchboards1) || (switchboard_count2 && !switchboards2)) {
		handle->err_no = XT_RTPPROXY_ERR_PARAM;
		handle->err_str = "Switchboards param is null";
		return -1;
	}
	if (!sessions) {
		handle->err_no = XT_RTPPROXY_ERR_PARAM;
		handle->err_str = "Sessions param is null";
		return -1;
	}

	for (k=0; k<session_count; k++) {
		
		if (!switchboard_count1) {
			for (i=0; i<2; i++) {
				if ((*sessions)[k].dir[i].switchboard.addr.ip && (*sessions)[k].dir[i].switchboard.addr.port) {
					goto sess_ok;
				}
			}
			handle->err_str = "switchboard addr not specified";
			handle->err_no = XT_RTPPROXY_ERR_PARAM;
			return -1;
		sess_ok:
			for (i=0; i<2; i++) {
				if (!(*sessions)[k].dir[i].switchboard.addr.ip) {
					(*sessions)[k].dir[i].switchboard.addr.ip = (*sessions)[k].dir[i^1].switchboard.addr.ip;
				}
				if (!(*sessions)[k].dir[i].switchboard.addr.port) {
					(*sessions)[k].dir[i].switchboard.addr.port = (*sessions)[k].dir[i^1].switchboard.addr.port;
				}
			}
		}
		/* test if there is at least one ip or learning timeout */
		for (i=0; i<2; i++) {
			for (j=0; j<2; j++) {
				if ( ((*sessions)[k].dir[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR &&
					  (*sessions)[k].dir[i].stream[j].source.ip) || 
				     ((*sessions)[k].dir[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_LEARNING_TIMEOUT &&
				      (*sessions)[k].dir[i].stream[j].learning_timeout) ) 
					goto ip_test_ok;
			}
		}
		handle->err_str = "at least one ip or learning timeout must be specified for each session";
		handle->err_no = XT_RTPPROXY_ERR_PARAM;
		return -1;

	ip_test_ok:		
		;
	}


	if (switchboard_count1 > 0) {
		sz = sizeof(switchboard_count1) + 
			  switchboard_count1*sizeof((*switchboards1)[0])+
			  sizeof(switchboard_count2) +
			  switchboard_count2*sizeof((*switchboards2)[0])+
			  sizeof(session_count)+
			  session_count*sizeof((*sessions)[0]);
		if (adjust_buf(handle, sz) < 0) {
			return -1;
		}
		fill_in_h(handle, XT_RTPPROXY_SOCKOPT_FLAG_SWITCHBOARD_GROUPS);
		RESET_BUF(handle, 1);
		COPY_TO_BUF(handle, session_count);
		COPY_TO_BUF(handle, switchboard_count1);
		for (i=0; i<switchboard_count1; i++) {
			COPY_TO_BUF(handle, (*switchboards1)[i]);
		}
		COPY_TO_BUF(handle, switchboard_count2);
		for (i=0; i<switchboard_count2; i++) {
			COPY_TO_BUF(handle, (*switchboards2)[i]);
		}
		
	} else {
		sz = sizeof(session_count)+
			 session_count*sizeof((*sessions)[0]);
		if (adjust_buf(handle, sz) < 0) {
			return -1;
		}
		fill_in_h(handle, 0); 
		RESET_BUF(handle, 1);
		COPY_TO_BUF(handle, session_count);
	}
	for (i=0; i<session_count; i++) {
		COPY_TO_BUF(handle, (*sessions)[i]);
	}

	if (do_sockopt(handle, XT_RTPPROXY_SO_GET_ALLOC_SESSIONS, handle->buf, handle->buf_ptr - handle->buf, &sz,
		"XT_RTPPROXY_SO_GET_ALLOC_SESSIONS error") < 0) {
		return -1;
	}

	RESET_BUF(handle, 0);
	COPY_FROM_BUF(handle, out_session_count);
	if (session_count != out_session_count) {
		handle->err_str = "BUG: out_session_count differs";
		handle->err_no = XT_RTPPROXY_ERR_UNEXPECTED;
		return -1;
	}
	for (i=0; i<out_session_count; i++) {
		COPY_FROM_BUF(handle, (*sessions)[i]);
	}
	return 0;
}

int xt_RTPPROXY_update_sessions(
	struct xt_rtpproxy_handle *handle,
	xt_rtpproxy_sockopt_count session_count,
	struct xt_rtpproxy_sockopt_session (*sessions)[] /* in-out param */
) {
	int i;
	unsigned int sz;
	void *save_buf_ptr;
	if (!session_count) {
		return 0;
	}
	if (!sessions) {
		handle->err_no = XT_RTPPROXY_ERR_PARAM;
		handle->err_str = "Sessions param is null";
		return -1;
	}
	if (adjust_buf(handle, 
		sizeof(session_count) +
		session_count*sizeof((*sessions)[0])
	) < 0) {
		return -1;
	}
	fill_in_h(handle, 0);
	RESET_BUF(handle, 1);
	COPY_TO_BUF(handle, session_count);
	save_buf_ptr = handle->buf_ptr;
	for (i=0; i<session_count; i++) {
		COPY_TO_BUF(handle, (*sessions)[i]);
	}
	
	sz =  handle->buf_ptr - handle->buf;	
	if (do_sockopt(handle, XT_RTPPROXY_SO_GET_UPDATE_SESSIONS, handle->buf, handle->buf_ptr - handle->buf, &sz,
		"XT_RTPPROXY_SO_GET_UPDATE_SESSIONS error") < 0) {
		return -1;
	}
	handle->buf_ptr = save_buf_ptr;
	for (i=0; i<session_count; i++) {
		COPY_FROM_BUF(handle, (*sessions)[i]);
	}
	return 0;
}

int xt_RTPPROXY_delete_switchboard_sessions(
	struct xt_rtpproxy_handle *handle,
	xt_rtpproxy_sockopt_count switchboard_count,   /* 0 to delete all switchboards */
	struct xt_rtpproxy_switchboard_id (*switchboards)[]
) {
	if (switchboard_count > 0) {
		int i;
		if (!switchboards) {
			handle->err_no = XT_RTPPROXY_ERR_PARAM;
			handle->err_str = "Switchboards param is null";
			return -1;
		}
		if (adjust_buf(handle, sizeof(switchboard_count) + 
							   switchboard_count*sizeof((*switchboards)[0])) < 0) {
			return -1;
		}
		fill_in_h(handle, 0);
		RESET_BUF(handle, 1);
		COPY_TO_BUF(handle, switchboard_count);
		for (i=0; i<switchboard_count; i++) {
			COPY_TO_BUF(handle, (*switchboards)[i]);
		}
		
	} else {
		if (adjust_buf(handle, 0) < 0) {
			return -1;
		}
		fill_in_h(handle, XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS); 
		RESET_BUF(handle, 1);
	}
		
	if (do_sockopt(handle, XT_RTPPROXY_SO_SET_DELETE_SWITCHBOARD_SESSIONS, handle->buf, handle->buf_ptr-handle->buf, NULL,
		"XT_RTPPROXY_SO_SET_DELETE_SWITCHBOARD_SESSIONS error") < 0) {
		return -1;
	}
	return 0;
}

int xt_RTPPPROXY_retrieve_sessions(
	struct xt_rtpproxy_handle *handle,
	struct xt_rtpproxy_sockopt_in_interval *interval,  /* NULL .. all sessions */
	xt_rtpproxy_sockopt_count switchboard_count,   /* 0 .. all switchboards */
	struct xt_rtpproxy_switchboard_id (*switchboards)[],
	u_int32_t *buf_size,                                 /* out param */
	void **buf
) {
	#define MIN_BUF_SIZE 65535 
	u_int16_t magic;
	unsigned int sz;
	
	if (switchboard_count && !switchboards) {
		handle->err_no = XT_RTPPROXY_ERR_PARAM;
		handle->err_str = "Switchboards param is null";
		return -1;
	}
	if (!buf_size) {
		handle->err_no = XT_RTPPROXY_ERR_PARAM;
		handle->err_str = "Buf_size param is null";
		return -1;
	}
	if (!buf) {
		handle->err_no = XT_RTPPROXY_ERR_PARAM;
		handle->err_str = "Buf param is null";
		return -1;
	}

	if (adjust_buf(handle, 
			sizeof(struct xt_rtpproxy_sockopt_in_interval)+
			sizeof(switchboard_count)+
			switchboard_count*(sizeof(struct xt_rtpproxy_switchboard_id))
			) < 0) {
		return -1;
	}
	if (adjust_buf(handle, MIN_BUF_SIZE) < 0) {
		return -1;
	}
try_next:
	fill_in_h(handle, (switchboard_count?0:XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS)|(!interval?XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS:0));
	RESET_BUF(handle, 1);
	if (switchboard_count) {
		int i;
		COPY_TO_BUF(handle, switchboard_count);
		for (i=0; i<switchboard_count; i++) {
			COPY_TO_BUF(handle, (*switchboards)[i]);
		}
	}
	if (interval) {
		COPY_TO_BUF(handle, *interval);
	}

	sz = handle->buf_size;
	if (do_sockopt(handle, XT_RTPPROXY_SO_GET_RETRIEVE_SESSIONS, handle->buf, handle->buf_ptr - handle->buf, &sz, 
		"XT_RTPPROXY_SO_GET_RETRIEVE_SESSIONS error") < 0) {
		return -1;
	}
	RESET_BUF(handle, 0);
	COPY_FROM_BUF(handle, magic);
	if (magic == 0) {
		u_int32_t req_mem;
		COPY_FROM_BUF(handle, req_mem);
		if (adjust_buf(handle, req_mem) < 0) {
			return -1;
		}
		goto try_next;
	}
	*buf_size = sz;
	*buf = handle->buf;
	handle->buf_iter = 1;
	return 0;
}

int xt_RTPPPROXY_replicate_sessions(
	struct xt_rtpproxy_handle *handle,
	u_int32_t buf_size,
	void *buf
) {
	if (do_sockopt(handle, XT_RTPPROXY_SO_SET_REPLICATE_SESSIONS, buf, buf_size, NULL,
		"XT_RTPPROXY_SO_SET_REPLICATE_SESSIONS error") < 0) {
		return -1;
	}
	return 0;
}

int xt_RTPPPROXY_get_replication_package_info(
	struct xt_rtpproxy_handle *handle,
	u_int32_t buf_size,
	void *buf,
  	struct xt_rtpproxy_replication_package_info *info
) {
#define COPY_FROM_BUF2(buf, val) \
	do { \
		memcpy(&(val), buf, sizeof(val)); \
		buf += sizeof(val); \
	} while (0);

	COPY_FROM_BUF2(buf, *info);
	if (info->header.magic != XT_RTPPROXY_SOCKOPT_MAGIC) {
		handle->err_no = XT_RTPPROXY_ERR_BAD_MAGIC;
		handle->err_str = "bad magic";
		return -1;
	}
	return 0;
}

int xt_RTPPROXY_set_ha_state(
	struct xt_rtpproxy_handle *handle,
	u_int8_t ha_state,
	xt_rtpproxy_sockopt_count switchboard_count,   /* 0 .. all switchboards */
	struct xt_rtpproxy_switchboard_id (*switchboards)[]
) {
	if (switchboard_count && !switchboards) {
		handle->err_no = XT_RTPPROXY_ERR_PARAM;
		handle->err_str = "Switchboards param is null";
		return -1;
	}
	if (adjust_buf(handle, sizeof(ha_state) + 
		   switchboard_count*(sizeof(struct xt_rtpproxy_sockopt_switchboard))
		) < 0) {
		return -1;
	}
	fill_in_h(handle, switchboard_count?0:XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS);
	RESET_BUF(handle, 1);

	COPY_TO_BUF(handle, ha_state);
	if (switchboard_count) {
		int i;
		COPY_TO_BUF(handle, switchboard_count);
		for (i=0; i<switchboard_count; i++) {
			COPY_TO_BUF(handle, (*switchboards)[i]);
		}
	}

	if (do_sockopt(handle, XT_RTPPROXY_SO_SET_HA_STATE, handle->buf, handle->buf_ptr-handle->buf, NULL,
		"XT_RTPPROXY_SO_SET_HA_STATE error") < 0) {
		return -1;
	}
	return 0;
}


/* rpc server functions */

extern void iptrtpproxy_rpc_1(struct svc_req *rqstp, register SVCXPRT *transp);

static __thread struct xt_rtpproxy_handle rpc_svs_handle;
//static struct xt_rtpproxy_handle rpc_svs_handle;

int xt_RTPPROXY_rpc_svc_run(struct xt_rtpproxy_handle *handle, struct xt_rtpproxy_connection_rpc_params *params) {

	SVCXPRT *transp;
	int proto;
	int sockfd;
	struct sockaddr_in addr_in;
	static char buf[30];
	if (xt_RTPPROXY_open(&rpc_svs_handle, xt_rtpproxy_LOCAL, NULL) < 0) {
		goto fail2;
	}
	sockfd = RPC_ANYSOCK; /* -1 */
	rpc_svs_handle.err_no = XT_RTPPROXY_ERR_RPC; 
	switch (params->proto) {
		case xt_rtpproxy_connection_UDP: 
			proto = IPPROTO_UDP;
			if (params->addr || params->port) {
				addr_in.sin_family = AF_INET;
				addr_in.sin_addr.s_addr = params->addr;
				addr_in.sin_port = htons(params->port); /* if zero then use portmapper */
				sockfd = socket(AF_INET, SOCK_DGRAM, proto);
				if (sockfd < 0) {
					snprintf(buf, sizeof(buf)-1, "socket error (%d)", errno);
					rpc_svs_handle.err_str = buf;
					goto fail;
				}
				if (bind(sockfd, (struct sockaddr *)&addr_in, sizeof(addr_in)) < 0) {
					snprintf(buf, sizeof(buf)-1, "socket error (%d)", errno);
					rpc_svs_handle.err_str = buf;
					goto fail;
				}
			}
			transp = svcudp_bufcreate(sockfd, 
					params->reply_size?params->reply_size:XT_RTPPROXY_RPC_DEFAULT_REPLY_SIZE, 
					params->request_size?params->request_size:XT_RTPPROXY_RPC_DEFAULT_REQUEST_SIZE
			);
			break;
		case xt_rtpproxy_connection_TCP:
			proto = IPPROTO_TCP;
			if (params->addr || params->port) {
				addr_in.sin_family = AF_INET;
				addr_in.sin_addr.s_addr = params->addr;
				addr_in.sin_port = htons(params->port); /* if zero then use portmapper */
				sockfd = socket(AF_INET, SOCK_STREAM, proto);
				if (sockfd < 0) {
					snprintf(buf, sizeof(buf)-1, "socket error (%d)", errno);
					rpc_svs_handle.err_str = buf;
					goto fail;
				}
				if (bind(sockfd, (struct sockaddr *)&addr_in, sizeof(addr_in)) < 0) {
					snprintf(buf, sizeof(buf)-1, "socket error (%d)", errno);
					rpc_svs_handle.err_str = buf;
					goto fail;
				}
			}
			transp = svctcp_create(sockfd,
					params->reply_size?params->reply_size:XT_RTPPROXY_RPC_DEFAULT_REPLY_SIZE, 
					params->request_size?params->request_size:XT_RTPPROXY_RPC_DEFAULT_REQUEST_SIZE
			);
			break;
		case xt_rtpproxy_connection_RAW:
			proto = 1;
			transp = svcraw_create();
			break;
		default:
			rpc_svs_handle.err_str = "unknown RPC protocol";
			goto fail;
	}
	if (!transp) {
		rpc_svs_handle.err_str = "svcXXX_bufcreate error";
		goto fail;
	}
	if (params->port) {
		proto = 0; /* do not use portmapper */
	}

	pmap_unset(IPTRTPPROXY_RPC, IPTRTPPROXY_RPC_VER);
	if (!svc_register(transp, IPTRTPPROXY_RPC, IPTRTPPROXY_RPC_VER, iptrtpproxy_rpc_1, proto)) {
		rpc_svs_handle.err_str = "svc_register error";
		goto fail;
	}
	svc_run(); /* never returns unless raw */
	return 0;
fail:
	if (sockfd >= 0) close(sockfd);
fail2:
	*handle = rpc_svs_handle;  /* caller is responsible to destroy it (to get reason of error) */
	return -1;
}

struct xt_rtpproxy_sockopt_result *getsockopt_1_svc(struct xt_rtpproxy_sockopt_params *argp, struct svc_req *rqstp) {
        static struct xt_rtpproxy_sockopt_result result;
	static char errs[30];
	unsigned int sz;

	sz = argp->result_size;
	if (sz < argp->data.data_len) {
		sz = argp->data.data_len;
	}
	memset(&result, 0, sizeof(result));
	if (adjust_buf(&rpc_svs_handle, sz) == 0) {
		memcpy(rpc_svs_handle.buf, argp->data.data_val, argp->data.data_len);
		snprintf(errs, sizeof(errs)-1, "getsockopt(%d) error", argp->command);
		result.err_no = do_sockopt(&rpc_svs_handle, argp->command, rpc_svs_handle.buf, argp->data.data_len, &sz, errs);
	}
	else {
		result.err_no = rpc_svs_handle.err_no;
	}
	if (result.err_no < 0) {
		result.xt_rtpproxy_sockopt_result_u.err_str = rpc_svs_handle.err_str;
	}
	else {
		result.xt_rtpproxy_sockopt_result_u.data.data_len = sz;
		result.xt_rtpproxy_sockopt_result_u.data.data_val = rpc_svs_handle.buf;
	}	
        return &result;
}

struct xt_rtpproxy_sockopt_result *setsockopt_1_svc(struct xt_rtpproxy_sockopt_params *argp, struct svc_req *rqstp) {
        static struct xt_rtpproxy_sockopt_result result;
	static char errs[30];

	memset(&result, 0, sizeof(result));
	snprintf(errs, sizeof(errs)-1, "setsockopt(%d) error", argp->command);
	result.err_no = do_sockopt(&rpc_svs_handle, argp->command, argp->data.data_val, argp->data.data_len, NULL, errs);
	if (result.err_no < 0) {
		result.xt_rtpproxy_sockopt_result_u.err_str = rpc_svs_handle.err_str;
	}
        return &result;
}
 
