/***************************************************************************
 *   (C) 2007 by Tomas Mandys <tomas.mandys at 2p dot cz>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License           *
 *   version 2.1 as published by the Free Software Foundation;             *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <libgen.h>
#include <errno.h>

#define _GNU_SOURCE
#include <getopt.h>
#include <signal.h>
#include <string.h>
#include <arpa/inet.h>

#include <linux/netfilter/xt_RTPPROXY.h>

static int verbose = 0;
static enum {fmtDefault=0, fmtXml, fmtTab} output_format = fmtDefault;
const char* program_name;

#define CMDVAL_ADDR_PORT 0x0001
#define CMDVAL_A_B 0x0002
#define CMDVAL_RTP_RTCP 0x0004
#define CMDVAL_LEARN_PORT_ONLY 0x0008
#define CMDVAL_BYTES_PACKETS 0x0010

#define CMDVAL_RPC_PROTO 0x0001
#define CMDVAL_RPC_ADDR 0x0002
#define CMDVAL_RPC_PORT 0x0003
#define CMDVAL_RPC_REQUEST_SIZE 0x0004
#define CMDVAL_RPC_REPLY_SIZE 0x0005
#define CMDVAL_RPC_TOTAL_TIMEOUT 0x0006
#define CMDVAL_RPC_UDP_RETRY_TIMEOUT 0x0007

#define CMDVAL_GRP_MASK 0x1F00
#define CMDVAL_GRP_ADDR 0x0100
#define CMDVAL_GRP_RADDR 0x0200
#define CMDVAL_GRP_LEARNING_TIMEOUT 0x0300
#define CMDVAL_GRP_ALWAYS_LEARN 0x0400
#define CMDVAL_GRP_SESSION_ID 0x0500
#define CMDVAL_GRP_LIST 0x0600
#define CMDVAL_GRP_RESET 0x0700
#define CMDVAL_GRP_SESSION_CREATED 0x0800
#define CMDVAL_GRP_FORMAT 0x0900
#define CMDVAL_GRP_EXPIRES 0x0A00
#define CMDVAL_GRP_NEW_SESSION 0x0B00
#define CMDVAL_GRP_DESTROY 0x0C00
#define CMDVAL_GRP_THROTTLE_MARK 0x0D00
#define CMDVAL_GRP_THROTTLE_BANDWIDTH 0x0E00
#define CMDVAL_GRP_HA_STATE 0x0F00
#define CMDVAL_GRP_INTERVAL 0x1000
#define CMDVAL_GRP_PACKAGE_FILE 0x1100
#define CMDVAL_GRP_RPC 0x1200
#define CMDVAL_GRP_TTL 0x1300

static struct option long_options[] = {
	{ "addr-a",			1, 0, CMDVAL_GRP_ADDR },
	{ "addr-b",			1, 0, CMDVAL_GRP_ADDR|CMDVAL_A_B },
	{ "port-a",			1, 0, CMDVAL_GRP_ADDR|CMDVAL_ADDR_PORT },
	{ "port-b",			1, 0, CMDVAL_GRP_ADDR|CMDVAL_ADDR_PORT|CMDVAL_A_B },
	{ "rtp-addr-a",			1, 0, CMDVAL_GRP_RADDR },
	{ "rtcp-addr-a",		1, 0, CMDVAL_GRP_RADDR|CMDVAL_RTP_RTCP },
	{ "rtp-port-a",			1, 0, CMDVAL_GRP_RADDR|CMDVAL_ADDR_PORT },
	{ "rtcp-port-a",		1, 0, CMDVAL_GRP_RADDR|CMDVAL_RTP_RTCP|CMDVAL_ADDR_PORT },
	{ "rtp-learn-port-only-a",	0, 0, CMDVAL_GRP_RADDR|CMDVAL_LEARN_PORT_ONLY },
	{ "rtcp-learn-port-only-a",	0, 0, CMDVAL_GRP_RADDR|CMDVAL_RTP_RTCP|CMDVAL_LEARN_PORT_ONLY },
	{ "rtp-addr-b",			1, 0, CMDVAL_GRP_RADDR|CMDVAL_A_B },
	{ "rtcp-addr-b",		1, 0, CMDVAL_GRP_RADDR|CMDVAL_RTP_RTCP|CMDVAL_A_B },
	{ "rtp-port-b",			1, 0, CMDVAL_GRP_RADDR|CMDVAL_ADDR_PORT|CMDVAL_A_B },
	{ "rtcp-port-b",		1, 0, CMDVAL_GRP_RADDR|CMDVAL_RTP_RTCP|CMDVAL_ADDR_PORT|CMDVAL_A_B },
	{ "rtp-learn-port-only-b",	0, 0, CMDVAL_GRP_RADDR|CMDVAL_LEARN_PORT_ONLY|CMDVAL_A_B },
	{ "rtcp-learn-port-only-b",	0, 0, CMDVAL_GRP_RADDR|CMDVAL_RTP_RTCP|CMDVAL_LEARN_PORT_ONLY|CMDVAL_A_B },
	{ "rtp-learning-timeout-a",	1, 0, CMDVAL_GRP_LEARNING_TIMEOUT },
	{ "rtp-learning-timeout-b",	1, 0, CMDVAL_GRP_LEARNING_TIMEOUT|CMDVAL_A_B },
	{ "rtcp-learning-timeout-a",	1, 0, CMDVAL_GRP_LEARNING_TIMEOUT|CMDVAL_RTP_RTCP },
	{ "rtcp-learning-timeout-b",	1, 0, CMDVAL_GRP_LEARNING_TIMEOUT|CMDVAL_RTP_RTCP|CMDVAL_A_B },
	{ "always-learn-a",		1, 0, CMDVAL_GRP_ALWAYS_LEARN },
	{ "always-learn-b",		1, 0, CMDVAL_GRP_ALWAYS_LEARN|CMDVAL_A_B },
	{ "expires",			1, 0, CMDVAL_GRP_EXPIRES },
	{ "ttl",				1, 0, CMDVAL_GRP_TTL },
	{ "destroy",			0, 0, CMDVAL_GRP_DESTROY },
	{ "new-session",		0, 0, CMDVAL_GRP_NEW_SESSION },

	{ "rtp-throttle-mark-a",	1, 0, CMDVAL_GRP_THROTTLE_MARK },
	{ "rtcp-throttle-mark-a",	1, 0, CMDVAL_GRP_THROTTLE_MARK|CMDVAL_RTP_RTCP },
	{ "rtp-throttle-mark-b",	1, 0, CMDVAL_GRP_THROTTLE_MARK|CMDVAL_A_B },
	{ "rtcp-throttle-mark-b",	1, 0, CMDVAL_GRP_THROTTLE_MARK|CMDVAL_RTP_RTCP|CMDVAL_A_B },
	{ "rtp-throttle-bandwidth-bytes-a",	1, 0, CMDVAL_GRP_THROTTLE_BANDWIDTH },
	{ "rtcp-throttle-bandwidth-bytes-a",	1, 0, CMDVAL_GRP_THROTTLE_BANDWIDTH|CMDVAL_RTP_RTCP },
	{ "rtp-throttle-bandwidth-bytes-b",	1, 0, CMDVAL_GRP_THROTTLE_BANDWIDTH|CMDVAL_A_B },
	{ "rtcp-throttle-bandwidth-bytes-b",	1, 0, CMDVAL_GRP_THROTTLE_BANDWIDTH|CMDVAL_RTP_RTCP|CMDVAL_A_B },
	{ "rtp-throttle-bandwidth-packets-a",	1, 0, CMDVAL_GRP_THROTTLE_BANDWIDTH|CMDVAL_BYTES_PACKETS },
	{ "rtcp-throttle-bandwidth-packets-a",	1, 0, CMDVAL_GRP_THROTTLE_BANDWIDTH|CMDVAL_RTP_RTCP|CMDVAL_BYTES_PACKETS },
	{ "rtp-throttle-bandwidth-packets-b",	1, 0, CMDVAL_GRP_THROTTLE_BANDWIDTH|CMDVAL_A_B|CMDVAL_BYTES_PACKETS },
	{ "rtcp-throttle-bandwidth-packets-b",	1, 0, CMDVAL_GRP_THROTTLE_BANDWIDTH|CMDVAL_RTP_RTCP|CMDVAL_A_B|CMDVAL_BYTES_PACKETS },

	{ "sess-id-a",			1, 0, CMDVAL_GRP_SESSION_ID },
	{ "sess-id-b",			1, 0, CMDVAL_GRP_SESSION_ID|CMDVAL_A_B },
	{ "sess-id-created",		1, 0, CMDVAL_GRP_SESSION_CREATED},

	{ "ha-state",			1, 0, CMDVAL_GRP_HA_STATE},
	
	{ "stamp-from",			1, 0, CMDVAL_GRP_INTERVAL|0 },
	{ "stamp-to",			1, 0, CMDVAL_GRP_INTERVAL|1 },
	{ "package",			1, 0, CMDVAL_GRP_PACKAGE_FILE},

	{ "sessions",			0, 0, CMDVAL_GRP_LIST|0 },

	{ "rpc-proto",			1, 0, CMDVAL_GRP_RPC|CMDVAL_RPC_PROTO },
	{ "rpc-addr",			1, 0, CMDVAL_GRP_RPC|CMDVAL_RPC_ADDR },
	{ "rpc-port",			1, 0, CMDVAL_GRP_RPC|CMDVAL_RPC_PORT },
	{ "rpc-request-size",		1, 0, CMDVAL_GRP_RPC|CMDVAL_RPC_REQUEST_SIZE },
	{ "rpc-reply-size",		1, 0, CMDVAL_GRP_RPC|CMDVAL_RPC_REPLY_SIZE },
	{ "rpc-total-timeout",		1, 0, CMDVAL_GRP_RPC|CMDVAL_RPC_TOTAL_TIMEOUT },
	{ "rpc-udp-retry-timeout",	1, 0, CMDVAL_GRP_RPC|CMDVAL_RPC_UDP_RETRY_TIMEOUT },

	{ "reset-global-stat",		0, 0, CMDVAL_GRP_RESET|0 },
	{ "reset-switchboard-stat",	0, 0, CMDVAL_GRP_RESET|1 },
	{ "reset-packet-stat",		0, 0, CMDVAL_GRP_RESET|2 },

	{ "format",			1, 0, CMDVAL_GRP_FORMAT },
	{ "verbose",			0, 0, 'v' },
	{ "help",			0, 0, 'h' },
	{ 0 }
};

struct enum_entry {
        int id;
        char* name;
        char* descr;
};

enum {cmdNone=0, cmdInfo, cmdList, cmdAlloc, cmdAllocGroup, cmdSessionInfo, cmdUpdate, cmdDelete, cmdHaState, cmdRetrieve, cmdReplicate, cmdRpcServer};

static struct enum_entry enum_command[] = {
	{id: cmdInfo, 	name: "info", descr: 
		"get RTP proxy global information"
	},
	{id: cmdList, 	name: "list",	descr: 
		"list specified switchboard or all switchboards if no switchboard(s) specified by one or multiple addr-a+port-a (addr-b,port-b are aliases). "
		"List also all sessions if --sessions options."
	},
	{id: cmdAlloc,	name: "alloc",	descr: 
		"alloc new RTP session(s). Switchboard must be specified by addr-a+port-a+addr-b+port-b. "
		"At least one remote client rtp/rtcp-(addr/port/learning-timeout)-a/b parameter is mandatory. "
		"Multiple sessions may be provided and each may use different switchboard."
	},
	{id: cmdAllocGroup, name: "alloc_group", descr:
		"alloc new RTP session(s). "
		"Typically two groups of switchboards are specified by multiple addr-a+port-a and addr-b+port-b sets. "
		"RTP proxy iterates through switchboard list and tries to allocate all sessions using one switchboard pair, "
		"i.e. switchboards are grouped to cluster which help to overcome limited number of ports. "
		"At least one remote client rtp/rtcp-(addr/port/learning-timeout)-a/b parameter is mandatory."
	},
	{id: cmdUpdate,	name: "update",	descr: 
		"update RTP session(s), session is specified by addr-a/b, port-a/b, sess-id-a/b."
	},
	{id: cmdSessionInfo, name: "session_info",	descr: 
		"get RTP session(s) info, session is specified by addr-a/b, port-a/b, sess-id-a/b."
	},
	{id: cmdDelete,	name: "delete_sessions",	descr: 
		"delete all RTP session(s) at one or all switchboards unless specified addr-a+port-a (addr-b,port-b are aliases). The sessions could be reused immediately."
	},
	{id: cmdHaState, name: "ha", descr:
		"set switchboard(s) HA state. Switchboards in standby are completely read-only with one exception-replicate."
	},
	{id: cmdRetrieve, name: "retrieve", descr:
		"retrieve switchboard(s) session state for replication purposes. If interval not provided then all sessions are retrieved. The command is"
		"valid at switchboards in active HA state only."
	},
	{id: cmdReplicate, name: "replicate", descr:
		"replicate session states. The switchboard(s) being replicated must be in HA standby state."
	},
	{id: cmdRpcServer,  name: "rpc_server", descr:
		"daemonize itself as RPC server to service remote requests."
	},
	{id: cmdNone, name: NULL}
};

static struct enum_entry enum_format_option[] = {
	{id: fmtDefault, name: "default", descr:
		"human readable (default)"
	},
	{id: fmtXml, name: "xml", descr:
		"XML"
	},
	{id: fmtTab, name: "tab", descr:
		"tab-delimited"
	},
	{id: -1, name: NULL}
};

static struct enum_entry *enum_by_id(int id, struct enum_entry *e) {
	int i;
	for (i=0; e[i].name != NULL; i++) {
		if (e[i].id == id) 
			return e+i;
	}
	return NULL;
}

static struct enum_entry *enum_by_name(char *name, struct enum_entry *e) {
	int i;
	for (i=0; e[i].name != NULL; i++) {
		if (strcmp(name, e[i].name)==0) 
			return e+i;
	}
	return NULL;
}

static void print_opt (int cmd, int opt) {
	struct option *o;
	int i;
	for (o=long_options; o->name; o++) {
		if ((o->val & CMDVAL_GRP_MASK) == opt) {
			switch (opt) {
				case CMDVAL_GRP_RPC:
					switch (o->val & 0x0F) {
						case CMDVAL_RPC_TOTAL_TIMEOUT:
						case CMDVAL_RPC_UDP_RETRY_TIMEOUT:
							if (cmd == cmdRpcServer) 
								continue;
							break;
						default: 
							;
					}
				break;
				default: 
					;
			}
			printf("  --%s ", o->name);
			switch (opt) {
				case CMDVAL_GRP_ADDR:
					printf("<%s  RTP proxy switchboard %c %s", 
						(o->val & CMDVAL_ADDR_PORT)?"port>":"ip>  ",
						'a'+((o->val & CMDVAL_A_B) != 0),
						(o->val & CMDVAL_ADDR_PORT)?"port":"ip"
					);
					break;

				case CMDVAL_GRP_RADDR:
					if (o->val & CMDVAL_LEARN_PORT_ONLY) {
						printf("   learn only %s client-%c port", 
							(o->val & CMDVAL_RTP_RTCP)?"RTCP":" RTP",
							'a'+((o->val & CMDVAL_A_B) != 0)
						);
					} else {
						printf("<%s  %s client-%c %s", 
							(o->val & CMDVAL_ADDR_PORT)?"port>":"ip>  ",
							(o->val & CMDVAL_RTP_RTCP)?"RTCP":" RTP",
							'a'+((o->val & CMDVAL_A_B) != 0),
							(o->val & CMDVAL_ADDR_PORT)?"port":"ip"
						);
					}
					break;

				case CMDVAL_GRP_LEARNING_TIMEOUT:
					printf("<sec>  %s client-%c learning timeout", 
						(o->val & CMDVAL_RTP_RTCP)?"RTCP":" RTP",
						'a'+((o->val & CMDVAL_A_B) != 0)
					);
					break;

				case CMDVAL_GRP_ALWAYS_LEARN:
					printf("<on/off>   always learn RTP client-%c", 
						'a'+((o->val & CMDVAL_A_B) != 0)
					);
					break;

				case CMDVAL_GRP_THROTTLE_MARK:
					printf("<bits>  mark %s client-%c packets", 
						(o->val & CMDVAL_RTP_RTCP)?"RTCP":" RTP",
						'a'+((o->val & CMDVAL_A_B) != 0)
					);
					break;
				
				case CMDVAL_GRP_THROTTLE_BANDWIDTH:
					printf("<%s/sec>  %s client-%c max.bandwidth",
						(o->val & CMDVAL_BYTES_PACKETS)?"packets":"bytes",
						(o->val & CMDVAL_RTP_RTCP)?"RTCP":" RTP",
						'a'+((o->val & CMDVAL_A_B) != 0)
					);
					break;
				
				case CMDVAL_GRP_EXPIRES:
					printf("<sec>       expiration timeout");
					break;

				case CMDVAL_GRP_TTL:
					printf("<sec>       max. time to live");
					break;

				case CMDVAL_GRP_DESTROY:
					printf("  destroy session");
					break;

				case CMDVAL_GRP_SESSION_ID:
					printf("<id>  session-id related to switchboard-%c", 
						'a'+((o->val & CMDVAL_A_B) != 0)
					);
					break;

				case CMDVAL_GRP_SESSION_CREATED:
					printf("<tick>  additional session identification key (optional)"
					);
					break;

				case CMDVAL_GRP_NEW_SESSION:
					printf("   delimits session parameters of next session");
					break;

				case CMDVAL_GRP_LIST:
					for (i=strlen(o->name); i<16; i++) printf(" ");
					switch (o->val & 0x0F) {
						case 0:
							printf("  list all sessions");
							break; 
						default: ;
					}
					break;

				case CMDVAL_GRP_RESET:
					for (i=strlen(o->name); i<23; i++) printf(" ");
					switch (o->val & 0x0F) {
						case 0:
							printf("  reset global statistics");
							break; 
						case 1:
							printf("  reset switchboard statistics");
							break; 
						case 2:
							printf("  reset switchboard packet statistics");
							break; 
						default: ;
					}
					break;

				case CMDVAL_GRP_INTERVAL:
					printf("<tick>   ");
					switch (o->val & 0x0F) {
						case 0:
							printf("  from time, default 0");
							break; 
						case 1:
							printf("  to time, default 0");
							break; 
						default: ;
					}
					break;

				case CMDVAL_GRP_PACKAGE_FILE:
					printf("<file>  replication package file"
					);
					break;

				case CMDVAL_GRP_FORMAT:
					printf("<fmt>      output format");
					for (i=0; enum_format_option[i].name; i++) {
						printf("\n            %s .. %s", enum_format_option[i].name,  enum_format_option[i].descr);
					}
					break;	

				case CMDVAL_GRP_HA_STATE:
					printf("<ha_state>       0..active, 1..standby");
					break;

				case CMDVAL_GRP_RPC:
					switch (o->val & 0x0F) {
						case CMDVAL_RPC_PROTO:
							printf("<udp|tcp>   protocol, default: udp");
							break;
						case CMDVAL_RPC_ADDR:
							if (cmd != cmdRpcServer) {
								printf("<ip>   address where RPC server is running");
							}
							else {
								printf("<ip>   address where RPC server service is provided (optional)");
							}
							break;
						case CMDVAL_RPC_PORT:
							printf("<port>   port where service is running, default: 0..use portmapper to get correct port");
							break;
						case CMDVAL_RPC_REQUEST_SIZE:
							printf("<bytes>   request buffer size, default: %u", XT_RTPPROXY_RPC_DEFAULT_REQUEST_SIZE);
							break;
						case CMDVAL_RPC_REPLY_SIZE:
							printf("<bytes>   reply buffer size, default: %u", XT_RTPPROXY_RPC_DEFAULT_REPLY_SIZE);
							break;
						case CMDVAL_RPC_TOTAL_TIMEOUT:
							printf("<msec>   time to get result, default: %u", XT_RTPPROXY_RPC_DEFAULT_TOTAL_TIMEOUT);
							break;
						case CMDVAL_RPC_UDP_RETRY_TIMEOUT:
							printf("<msec>   time wait for result before retrasmission is sent, default: %u", XT_RTPPROXY_RPC_DEFAULT_UDP_REPLY_TIMEOUT);
							break;
						default:
							;
					}
					break;
				default: ;
			}
			printf("\n");
		}
	}
}

static void print_version() {
	printf("iptables RTPPROXY helper, v%s\n", XT_RTPPROXY_VERSION);
}

static void help(int cmd) {
	struct enum_entry *e;
	print_version();
	if (!cmd) {
		printf("Usage: %s [options] <command> [options]\n", program_name);
		printf("Type '%s -h <command>' for help on a specific command.\n", program_name);

		printf("\n");
		printf("Command list:\n");
		for (e=enum_command; e->name; e++) {
			printf("  %-8s  %s\n", e->name, e->descr);
		}
	}
	else {
		e = enum_by_id(cmd, enum_command);
		printf("Usage: %s %s [options]\n", program_name, e->name);

		printf("%s: %s\n", e->name, e->descr);

		switch (cmd) {
			case cmdAlloc:
			case cmdUpdate:
			case cmdSessionInfo:
				printf("\n"
					"Session identification:\n");
				print_opt(cmd, CMDVAL_GRP_ADDR);
				print_opt(cmd, CMDVAL_GRP_SESSION_ID);
				print_opt(cmd, CMDVAL_GRP_SESSION_CREATED);
				print_opt(cmd, CMDVAL_GRP_NEW_SESSION);
				break;
			case cmdAllocGroup:
			case cmdDelete:
			case cmdList:
			case cmdHaState:
			case cmdRetrieve:
				printf("\n"
					"Switchboard identification:\n");
				print_opt(cmd, CMDVAL_GRP_ADDR);
				break;
		}
	
		if (cmd == cmdAlloc || cmd == cmdAllocGroup || cmd == cmdUpdate) {
			printf("\n"
				"RTP client parameters:\n");
			print_opt(cmd, CMDVAL_GRP_RADDR);
			print_opt(cmd, CMDVAL_GRP_EXPIRES);
			print_opt(cmd, CMDVAL_GRP_TTL);
			print_opt(cmd, CMDVAL_GRP_LEARNING_TIMEOUT);
			print_opt(cmd, CMDVAL_GRP_ALWAYS_LEARN);
			if (cmd == cmdUpdate) {
				print_opt(cmd, CMDVAL_GRP_DESTROY);
			}
			print_opt(cmd, CMDVAL_GRP_THROTTLE_MARK);
			print_opt(cmd, CMDVAL_GRP_THROTTLE_BANDWIDTH);
		}		

		printf("\n"
			"RPC parameters:\n");
		print_opt(cmd, CMDVAL_GRP_RPC);

		if (cmd == cmdList) {
			printf("\n"
				"List verbosity:\n");
			print_opt(cmd, CMDVAL_GRP_LIST);
		}
		if (cmd == cmdHaState) {
			printf("\n");
			print_opt(cmd, CMDVAL_GRP_HA_STATE);
		};
		if (cmd == cmdRetrieve || cmd == cmdReplicate) {
			printf("\n");
			print_opt(cmd, CMDVAL_GRP_PACKAGE_FILE);
		}
		if (cmd == cmdRetrieve) {
			print_opt(cmd, CMDVAL_GRP_INTERVAL);
		}
		printf("\n"
			"Other options:\n");
		print_opt(cmd, CMDVAL_GRP_RESET);
		if (cmd == cmdList || cmd == cmdAlloc || cmd == cmdSessionInfo || cmd == cmdInfo || cmd == cmdRetrieve || cmd == cmdReplicate) {
			print_opt(cmd, CMDVAL_GRP_FORMAT);
		}
	}

	printf("\n");
	printf("Common options:\n");
	printf("  -h,--help      print this help\n");
	printf("  -v,--verbose   print more info, mostly to stderr\n");
	printf("  -V             print version\n");
	printf("\n");
}

static char *ip2s(u_int32_t ip) {
	struct in_addr ip2 = { ip };
	return inet_ntoa(ip2);
} 
/*static char *ip2s(u_int32_t addr) {
    static char buf[17];
    const unsigned char *bytep;

    bytep = (const unsigned char *) &addr;
    snprintf(buf, 16, "%u.%u.%u.%u", bytep[0], bytep[1], bytep[2], bytep[3]);
    buf[16] = 0;
    return buf;
}
*/

static char* state_s[] = {"none", "init1", "init2", "forward1", "forward2", "destroyed", "staled", "expired"};

static void print_header(struct xt_rtpproxy_sockopt_out_config *config) {
	switch (output_format) {
		case fmtXml:
			printf("<header version=\"0x%x\" />\n", config->version);
			break;
		default:
			printf("header: {version: 0x%x}\n", config->version);
			break;
	}
}

static void print_foot() {
	switch (output_format) {
		default:
			;
	}
}

static void print_stamp(u_int32_t t) {
	time_t tt;
	char *c, *c2;	
	tt = (time_t) t;
	c = ctime(&tt);
	for (c2 = c+strlen(c); c2 >= c && *c2 <= ' ' ; c2--) {
		*c2 = '\0';
	}
	switch (output_format) {

		case fmtXml:
			printf("tick=\"%u\" time=\"%s\"", t, c);
			break;
		case fmtTab:
			printf("%u\t%s", t, c);
			break;
		default:
			printf("{tick: %u, time: %s}", t, c);
			break;
	}
}

static void print_info(struct xt_rtpproxy_sockopt_out_config *info) {
	static char *cmd_s[XT_RTPPROXY_MAX_CMD] = {"delete", "replicate", "ha_state",  "config", "alloc", "update", "retrieve"};
	int i;

	switch (output_format) {

		case fmtXml:
			printf("<started "); print_stamp(info->started); printf(" />\n");
			printf("<total switchboards=\"%u\" sessions=\"%u\" />\n", info->total.num_switchboard, info->total.num_pair_session); 
			printf("<stat ok=\"%u\" fail=\"%u\" perm=\"%u\" unknown=\"%u\"", 
				info->stat.ok, 
				info->stat.fail, 
				info->stat.perm, 
				info->stat.unknown
			);
			for (i=0; i<XT_RTPPROXY_MAX_CMD; i++) {
				printf(" %s=\"%u\"", cmd_s[i], info->stat.cmd[i]); 
			}
			printf(" />\n");
			break;
		case fmtTab:
			printf("started/tick\tstarted/datetime");
			printf("\ttotal/switchboards\ttotal/sessions\tstat/ok\tstat/fail\tstat/perm\tstat/unknown");
			for (i=0; i<XT_RTPPROXY_MAX_CMD; i++) {
				printf("\tstat/%s", cmd_s[i]); 
			}
			printf("\n");
			print_stamp(info->started);
			printf("\t%u\t%u", info->total.num_switchboard, info->total.num_pair_session);
			printf("\t%u\t%u\t%u\t%u", 
				info->stat.ok, 
				info->stat.fail, 
				info->stat.perm, 
				info->stat.unknown
			);
			for (i=0; i<XT_RTPPROXY_MAX_CMD; i++) {
				printf("\t%u", info->stat.cmd[i]); 
			}
			printf("\n");
			break;
		default:
			printf("started: "); print_stamp(info->started); printf("\n");
			printf("total: {switchboards: %u, sessions: %u}\n", info->total.num_switchboard, info->total.num_pair_session); 
			printf("stat: {ok: %u, fail: %u, perm: %u, unknown: %u", 
				info->stat.ok, 
				info->stat.fail, 
				info->stat.perm, 
				info->stat.unknown
			);
			for (i=0; i<XT_RTPPROXY_MAX_CMD; i++) {
				printf(", %s: %u", cmd_s[i], info->stat.cmd[i]); 
			}
			printf("}\n");
			break;
	}
}

static void print_stat_title(char *gate) {
	switch (output_format) {
		case fmtTab:
			printf("%sstat/packets/in\t%sstat/packets/out\t%sstat/bytes/in\t%sstat/bytes/out", 
				gate, gate, gate, gate);
			break;
		default:;
	}
}

static void print_stat(struct xt_rtpproxy_stat *stat) {
	switch (output_format) {
		case fmtXml:
			printf("<stat><packets in=\"%u\" out=\"%u\" /><bytes in=\"%u\" out=\"%u\" /></stat>\n",
				stat->packets_in,
				stat->packets_out,
				stat->bytes_in,
				stat->bytes_out
			);
			break;
		case fmtTab:
			printf("%u\t%u\t%u\t%u",
				stat->packets_in,
				stat->packets_out,
				stat->bytes_in,
				stat->bytes_out
			);
			break;
		default:
			printf("packets: {in: %u, out: %u}, bytes: {in: %u, out: %u}",
				stat->packets_in,
				stat->packets_out,
				stat->bytes_in,
				stat->bytes_out
			);
	};
}

static void print_switchboard(struct xt_rtpproxy_switchboard_id *s) {
	fprintf(stderr, "{ip: %s, port: %d}", 
		ip2s(s->ip), 
		s->port
	); /* ip2s ... result in static buffer */
}

static void print_in_switchboards(char *suffix, xt_rtpproxy_sockopt_count count, struct xt_rtpproxy_switchboard_id (*s)[]) {
	int i;
	fprintf(stderr, "  switchboard%s: ", suffix);
	if (count) {
		for (i=0; i<count; i++) {
			if (i > 0) fprintf(stderr, ", ");
			print_switchboard(&(*s)[i]);
		}
	}
	else {
		fprintf(stderr, "NULL");
	}
}

static void print_in_sessions(xt_rtpproxy_sockopt_count count, struct xt_rtpproxy_sockopt_session (*s)[], int print_switchboards) {
	int i, j, k;
	fprintf(stderr, "  session: ");
	if (count) {
		for (k = 0; k<count; k++) {		
			if (k > 0) fprintf(stderr, ",\n");
			
			fprintf(stderr, "{\n");
			if (print_switchboards) {
				for (i=0; i<2; i++) {
					fprintf(stderr, "    switchboard-%c: ", 'a'+i); print_switchboard(&(*s)[k].dir[i].switchboard.addr); fprintf(stderr, "\n");
				}
			}
			if ((*s)[k].sh.flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_EXPIRES) {
					fprintf(stderr, "    expires-timeout: %d,\n", (*s)[k].sh.expires_timeout);
			}
			if ((*s)[k].sh.flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_TTL) {
					fprintf(stderr, "    ttl: %d,\n", (*s)[k].sh.ttl);
			}
			for (i=0; i<2; i++) {
				for (j=0; j<2; j++) {
					if (((*s)[k].dir[i].stream[j].flags & ~XT_RTPPROXY_SOCKOPT_FLAG_LEARN_PORT_ONLY) != 0) {
						fprintf(stderr, "      %s-%c: {", j?"rtcp":"rtp", 'a'+i);
						if ((*s)[k].dir[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR) {
							fprintf(stderr, "addr: {ip: %s, port: %d}, learn-port-only: %d", 
									ip2s((*s)[k].dir[i].stream[j].source.ip), 
									(*s)[k].dir[i].stream[j].source.port,
									((*s)[k].dir[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_LEARN_PORT_ONLY) != 0
								);
						}
						if ((*s)[k].dir[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_LEARNING_TIMEOUT) {
							if ((*s)[k].dir[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR)
								fprintf(stderr, ", ");
							fprintf(stderr, "learn-timeout: %d", (*s)[k].dir[i].stream[j].learning_timeout);
						}
						if ((*s)[k].dir[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_THROTTLE_MARK) {
							if ((*s)[k].dir[i].stream[j].flags & (XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR|XT_RTPPROXY_SOCKOPT_FLAG_SESSION_LEARNING_TIMEOUT))
								fprintf(stderr, ", ");
							fprintf(stderr, "throttle-mark: %d", (*s)[k].dir[i].stream[j].throttle.mark);
						}
						if ((*s)[k].dir[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_THROTTLE_BANDWIDTH) {
							if ((*s)[k].dir[i].stream[j].flags & (XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR|XT_RTPPROXY_SOCKOPT_FLAG_SESSION_LEARNING_TIMEOUT|XT_RTPPROXY_SOCKOPT_FLAG_THROTTLE_MARK))
								fprintf(stderr, ", ");
							fprintf(stderr, "throttle-bandwidth: {bytes:%d, packets: %d}", 
								(*s)[k].dir[i].stream[j].throttle.max_bandwidth.bytes,
								(*s)[k].dir[i].stream[j].throttle.max_bandwidth.packets
							);
						}
						fprintf(stderr, "},\n");
					}
				}
				if ((*s)[k].dir[i].flags & XT_RTPPROXY_SOCKOPT_FLAG_ALWAYS_LEARN) {
					fprintf(stderr, "    always-learn-%c: %d\n", 'a'+i, (*s)[k].dir[i].always_learn);
				}
			}
			fprintf(stderr, "    destroy: %d,\n", ((*s)[k].sh.flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_DESTROY) !=0);
		}
		fprintf(stderr, "  }");
	}
	else {
		fprintf(stderr, "NULL");
	}
}

enum {posFull = 0x01, posAlloc = 0x02};

static void print_out_sessions_header(int full, char *indent) {
	int g;
	switch (output_format) {
		case fmtXml:
			printf("%s<sessions>\n", indent);
			break;
		case fmtTab:
			printf("switchboard-a/ip\tswitchboard-a/port\tswitchboard-b/ip\tswitchboard-b/port\t"
				"sess-id-a\tsess-id-b\tcreated/tick\tcreated/datetime");
			if (full & posAlloc) {
				printf("\tswitchboard-a/free\tswitchboard-a/alloc\tswitchboard-b/free\tswitchboard-b/alloc");
			}
			if (full & posFull) {
				printf("\texpires/timeout\texpires/tick\texpires/datetime\tttl");
			}
			for (g=0; g<2; g++) {
				int st;
				if (full & posFull) {
					printf("\talways-learn-%c", 'a'+g);
				}
				for (st=0; st<2; st++) {
					char c[15];
					sprintf(c, "%s-%c/", st?"rtcp":"rtp", 'a'+g);
					printf("\t%sport", c);
					if (full & posFull) {
						printf("\t%ssource/ip\t%ssource/port\t"
							"%ssource-learnt/ip\t%ssource-learnt/port\t"
							"%slearning/tick\t%slearning/datetime\t"
							"%slast-packet/tick\t%slast-packet/datetime\t"
							"%sthrottle/mark\t%sthrottle/max-bandwidth/bytes\t%sthrottle/max-bandwidth/packets\t"
							"%sthrottle/bandwidth/bytes\t%sthrottle/bandwidth/packets\t"
							"%sthrottle/bandwidth/stamp/tick\t%sthrottle/bandwidth/stamp/datetime\t"
							"%sthrottle/mark-counter/bytes\t%sthrottle/mark-counter/packets\t",
							c, c, c, c, c, c, c, c, c, c, c, c, c, c, c, c, c);
					}
				}
				if (full & posFull) {
					char c[15];
					sprintf(c, "%s/", st?"rtcp":"rtp");
					print_stat_title(c);
				}
			}
			printf("\n");
			break;

		default:
			printf("%ssessions: {\n", indent);
			break;
	}
}
			
static void print_out_session(struct xt_rtpproxy_sockopt_session *out_session, int full, char *indent) {
	int g;
	switch (output_format) {
		case fmtXml:
			printf("%s<session", indent);
			if (full & posFull) {
				printf(" state=\"%s\"  expires-timeout=\"%d\" ttl=\"%d\"", 
					state_s[out_session->sh.state], out_session->sh.expires_timeout, out_session->sh.ttl);
			}
			printf(">\n");

			printf("%s  <created ", indent); print_stamp(out_session->sh.created); printf(" />\n");
			if (full & posFull) {
				printf("%s  <expires ", indent); print_stamp(out_session->sh.expires); printf(" />\n");
			}
			for (g=0; g<2; g++) {
				int st;
				printf("%s  <switchboard id=\"%c\" addr=\"%s\" port=\"%d\" sess-id=\"%d\"", 
						indent, 
						'a'+g, 
						ip2s(out_session->dir[g].switchboard.addr.ip), 
						out_session->dir[g].switchboard.addr.port, 
						out_session->dir[g].sess_id
				);
				if (full & posAlloc) {
					printf(" free=\"%d\" alloc=\"%d\"", 
						out_session->dir[g].switchboard.free,
						out_session->dir[g].switchboard.alloc
					);
				}
				if (full & posFull) {
					printf(" always-learn=\"%d\"", out_session->dir[g].always_learn);
				}
				printf(">\n");
				for (st=0; st<2; st++) {
					if (full & posFull) {
						printf("%s    <port proto=\"%s\" port=\"%d\"><source ip=\"%s\" port=\"%d\" /><source_learnt ip=\"%s\" port=\"%d\" />\n",
							indent,
							st==0?"rtp":"rtcp",
							out_session->dir[g].stream[st].port,
							ip2s(out_session->dir[g].stream[st].source.ip),
							out_session->dir[g].stream[st].source.port,
							ip2s(out_session->dir[g].stream[st].source_learnt.ip),
							out_session->dir[g].stream[st].source_learnt.port
						);
						printf("%s      <learning ", indent);
						print_stamp(out_session->dir[g].stream[st].learning),
						printf(" />\n");
						printf("%s      <last_packet ", indent);
						print_stamp(out_session->dir[g].stream[st].last_packet_stamp);
						printf(" />\n");
						printf("%s      <throttle mark=\"%d\">\n", indent, out_session->dir[g].stream[st].throttle.mark);
						printf("%s        <max_bandwidth bytes=\"%d\" packets=\"%d\" />\n", 
							indent,
							out_session->dir[g].stream[st].throttle.max_bandwidth.bytes,
							out_session->dir[g].stream[st].throttle.max_bandwidth.packets
						);
						printf("%s        <bandwidth bytes=\"%d\" packets=\"%d\" ", 
							indent,
							out_session->dir[g].stream[st].throttle.bandwidth.bytes,
							out_session->dir[g].stream[st].throttle.bandwidth.packets
						);
						print_stamp(out_session->dir[g].stream[st].throttle.stamp);
						printf(" />\n");
						printf("%s        <mark_counter bytes=\"%d\" packets=\"%d\" />\n", 
							indent,
							out_session->dir[g].stream[st].throttle.mark_counter.bytes,
							out_session->dir[g].stream[st].throttle.mark_counter.packets
						);
						printf("%s      </throttle>\n", indent);
						printf("%s    </port>\n", indent);
					}
					else {
						printf("%s    <port proto=\"%s\" port=\"%d\" />\n",
							indent,
							st==0?"rtp":"rtcp",
							out_session->dir[g].stream[st].port
						);
					}
				}
				if (full & posFull) {
					printf("%s    ", indent); print_stat(&out_session->dir[g].stat); 
				}
				printf("%s  </switchboard>\n", indent);
			}
			printf("%s</session>\n", indent); 

			break;

		case fmtTab:

			printf("%s\t%d", 
				ip2s(out_session->dir[0].switchboard.addr.ip),
				out_session->dir[0].switchboard.addr.port
			); /* ip2s ... result in static buffer */
			printf("\t%s\t%d", 
				ip2s(out_session->dir[1].switchboard.addr.ip),
				out_session->dir[1].switchboard.addr.port
			); /* ip2s ... result in static buffer */
			printf("\t%d\t%d\t", 
				out_session->dir[0].sess_id,
				out_session->dir[1].sess_id 
			);
			print_stamp(out_session->sh.created);
			if (full & posAlloc) {
				printf("\t%d\t%d\t%d\t%d", 
					out_session->dir[0].switchboard.free,
					out_session->dir[0].switchboard.alloc,
					out_session->dir[1].switchboard.free,
					out_session->dir[1].switchboard.alloc
				);
			}
			if (full & posFull) {
				printf("\t%d\t", out_session->sh.expires_timeout); print_stamp(out_session->sh.expires); printf("%d\t", out_session->sh.ttl);
			}
			for (g=0; g<2; g++) {
				int st;
				if (full & posFull) {
					printf("\t%d", out_session->dir[g].always_learn);
				}
				for (st=0; st<2; st++) {
					printf("\t%d", out_session->dir[g].stream[st].port);
					if (full & posFull) {
						printf("\t%s\t%d\t%s\t%d\t",
							ip2s(out_session->dir[g].stream[st].source.ip),
							out_session->dir[g].stream[st].source.port,
							ip2s(out_session->dir[g].stream[st].source_learnt.ip),
							out_session->dir[g].stream[st].source_learnt.port
						);
						print_stamp(out_session->dir[g].stream[st].learning); printf("\t");
						print_stamp(out_session->dir[g].stream[st].last_packet_stamp); printf("\t");
						printf("%d\t%d\t%d\t%d\t%d\t",
							out_session->dir[g].stream[st].throttle.mark,
							out_session->dir[g].stream[st].throttle.max_bandwidth.bytes,
							out_session->dir[g].stream[st].throttle.max_bandwidth.packets,
							out_session->dir[g].stream[st].throttle.bandwidth.bytes,
							out_session->dir[g].stream[st].throttle.bandwidth.packets
						);
						print_stamp(out_session->dir[g].stream[st].throttle.stamp); 
						printf("\t%d\t%d\t",
							out_session->dir[g].stream[st].throttle.mark_counter.bytes,
							out_session->dir[g].stream[st].throttle.mark_counter.packets
						);
					}
				}
				if (full & posFull) {
					print_stat(&out_session->dir[g].stat); 
				}
			}
			printf("\n");
			break;

		default:
			printf("%s{\n", indent);
			printf("%s  ", indent);
			if (full & posFull) {
				printf("state: %s, ", state_s[out_session->sh.state]);
			}
			printf("created: ");
			print_stamp(out_session->sh.created);
			if (full & posFull) {
				printf(", expires-timeout: %d, expires: ", out_session->sh.expires_timeout);
				print_stamp(out_session->sh.expires);
				printf(", ttl: %d", out_session->sh.ttl);
			}
			printf("\n");
			printf("%s  switchboard: {\n", indent);
			for (g=0; g<2; g++) {
				int st;
				printf("%s    %c: {\n", indent, 'a'+g);
				printf("%s      addr: %s, port: %d, sess-id: %d\n", 
						indent, 
						ip2s(out_session->dir[g].switchboard.addr.ip), 
						out_session->dir[g].switchboard.addr.port, 
						out_session->dir[g].sess_id
				);
				if (full & posAlloc) {
					printf(", free: %d, alloc: %d\n", 
						out_session->dir[g].switchboard.free,
						out_session->dir[g].switchboard.alloc
					);
				}
				for (st=0; st<2; st++) {
					if (full & posFull) {
						printf("%s      %s: {port: %d, source: {ip: %s, port: %d}, source-learnt: {ip: %s, port: %d}\n", 
							indent,
							st==0?"rtp":"rtcp",
							out_session->dir[g].stream[st].port,
							ip2s(out_session->dir[g].stream[st].source.ip),
							out_session->dir[g].stream[st].source.port,
							ip2s(out_session->dir[g].stream[st].source_learnt.ip),
							out_session->dir[g].stream[st].source_learnt.port
						);
						printf("%s        learning: ", indent);
						print_stamp(out_session->dir[g].stream[st].learning),
						printf("\n");
						printf("%s        last-packet: ", indent);
						print_stamp(out_session->dir[g].stream[st].last_packet_stamp);
						printf("}\n");
						printf("%s        throttle: {mark: %d, max-bandwidth: {bytes: %d, packets: %d}, bandwidth: {bytes: %d, packets: %d, ", 
							indent,
							out_session->dir[g].stream[st].throttle.mark,
							out_session->dir[g].stream[st].throttle.max_bandwidth.bytes,
							out_session->dir[g].stream[st].throttle.max_bandwidth.packets,
							out_session->dir[g].stream[st].throttle.bandwidth.bytes,
							out_session->dir[g].stream[st].throttle.bandwidth.packets
						);
						print_stamp(out_session->dir[g].stream[st].throttle.stamp);
						printf("}, mark-counter: {bytes: %d, packets: %d}}\n",
							out_session->dir[g].stream[st].throttle.mark_counter.bytes,
							out_session->dir[g].stream[st].throttle.mark_counter.packets
						);
					}
					else {
						printf("%s      %s: {port: %d}\n", indent, st==0?"rtp":"rtcp", out_session->dir[g].stream[st].port);
					}
				}
				if (full & posFull) {
					printf("%s      always-learn: %d\n", indent, out_session->dir[g].always_learn);
					printf("%s      stat: ", indent); print_stat(&out_session->dir[g].stat); printf("\n");
				}		
				printf("%s    }\n", indent);
			}
			printf("%s  }\n", indent);
			printf("%s}\n", indent);

			break;
	}
}

static void print_out_sessions_foot(char *indent) {
	switch (output_format) {
		case fmtXml:
			printf("%s</sessions>\n", indent);
			break;
		case fmtTab:
			break;
		default:
			printf("%s}\n", indent);
			break;
	}
}

static void print_out_sessions(xt_rtpproxy_sockopt_count count, struct xt_rtpproxy_sockopt_session (*sessions)[], int full) {
	int i;
	print_out_sessions_header(full, "");
	for (i=0; i<count; i++) {
		if (((*sessions)[i].sh.flags & XT_RTPPROXY_SOCKOPT_FLAG_NOT_FOUND) == 0) {
			print_out_session(&(*sessions)[i], full, "  ");
		}
	}
	print_out_sessions_foot("");
}

static void print_replication_package_info(struct xt_rtpproxy_replication_package_info *info) {
	switch (output_format) {
		case fmtXml:
			printf("  <repl record-count=\"%u\">\n", info->record_count);
			printf("    <interval>\n");
			printf("      <from "); print_stamp(info->interval.from_stamp); printf(" />\n");
			printf("      <to "); print_stamp(info->interval.to_stamp); printf(" />\n");
			printf("    </interval>\n");
			printf("  </repl>\n"); 
			break;
		case fmtTab:
			printf("repl/interval/from/tick\trepl/interval/from/datetime\trepl/interval/to/tick\trepl/interval/to/datetime\trepl/record-count\n");
			print_stamp(info->interval.from_stamp); printf("\t"); print_stamp(info->interval.to_stamp); printf("\t"); 
			printf("%u", info->record_count);
			printf("\n");
			break;
		default:
			printf("repl: {\n");
			printf("  record-count: %u\n", info->record_count);
			printf("  interval: {\n");
			printf("    from: "); print_stamp(info->interval.from_stamp); printf("\n");
			printf("    to: "); print_stamp(info->interval.to_stamp); printf("\n");
			printf("  }\n");
			printf("}\n");
			break;
	}
}

static void print_format() {
	struct enum_entry *e = enum_by_id(output_format, enum_format_option);
	if (e) fprintf(stderr, "format: %s\n", e->name); 
}

static char stop_flag = 0;
static void sig_term(int signr) {
    signal(SIGINT, SIG_IGN);
    signal(SIGQUIT, SIG_IGN);
    signal(SIGTERM, SIG_IGN);
    stop_flag = 1;
}
#define SIGNAL(sig) \
	if (signal(sig, sig_term) == SIG_ERR) {\
		printf("can't install signal handler for %s\n", #sig);\
	exit (-1);\
}
 
static u_int32_t s2ip(char *optarg) {
	u_int32_t res = 0;
	if (!inet_aton(optarg, (struct in_addr*) &res)) {
		fprintf(stderr, "Invalid IP address %s\n", optarg);
		exit(-1);
	}
	return res;
}

static u_int32_t s2u(char *optarg) {
	long res = atol(optarg);
	if (res < 0) {
		fprintf(stderr, "Invalid number %s\n", optarg);
		exit(-1);
	}
	return res;
}

int main(int argc, char** argv) { 

	int cmd = cmdNone, help_flag = 0;
	static char short_opts[100] = "+HvV";
	struct xt_rtpproxy_switchboard_id (*switchboards_a)[], (*switchboards_b)[], cur_switchboard;
	struct xt_rtpproxy_sockopt_session (*sessions)[], cur_session;
	xt_rtpproxy_sockopt_count switchboard_a_count, switchboard_b_count, session_count;
	struct xt_rtpproxy_handle handle;
	int pending_switchboard; /*0..none, 1..A, 2..B*/
	int pending_session;
	int i, j;
	u_int32_t in_flags, list_verbose, ha_state;
	char *package_file = NULL;
	struct xt_rtpproxy_sockopt_in_interval interval;
	struct xt_rtpproxy_connection_rpc_params rpc_params;
	int rpc_params_flag = 0;

	
	#define PUSH_ARRAY(arr, count, item) {\
		void *p; \
		p = realloc(arr, (count+1)*sizeof((*arr)[0])); \
		if (!p) { \
			fprintf(stderr, "ERROR: Not enough memory\n"); \
			goto err; \
		} \
		arr = p; \
		(*arr)[count] = item; \
		count++; \
	}
	
	#define PUSH_SWITCHBOARD \
		if (pending_switchboard==1) { PUSH_ARRAY(switchboards_a, switchboard_a_count, cur_switchboard); } \
		else if (pending_switchboard==2) { PUSH_ARRAY(switchboards_b, switchboard_b_count, cur_switchboard); }
	
	#define RETURN(c) { \
		if (switchboards_a) free(switchboards_a); \
		if (switchboards_b) free(switchboards_b); \
		if (sessions) free(sessions); \
		return c; \
	}

	#define CLEAR_SESSION {\
		if (pending_session != 1) {\
			memset(&cur_session, 0, sizeof(cur_session)); \
			pending_session = 1; \
		}\
	}
	
	program_name = basename(argv[0]);
	switchboard_a_count = switchboard_b_count = session_count = 0;
	switchboards_a = switchboards_b = NULL;
	memset(&interval, 0, sizeof(interval));
	memset(&rpc_params, 0, sizeof(rpc_params));
	sessions = NULL;
	if (argc <= 1) {
		help(cmd);
		RETURN(2);
	}
	memset(&cur_switchboard, 0, sizeof(cur_switchboard));
	pending_session = 0;
	CLEAR_SESSION;
	pending_switchboard = 0;

	j = strlen(short_opts);
	for (i=0; long_options[i].name; i++) {
		if (long_options[i].val > ' ' && long_options[i].val < 128) {
			short_opts[j++] = (char) long_options[i].val;
			if (long_options[i].has_arg != no_argument) 
				short_opts[j++] = ':';
			if (long_options[i].has_arg == optional_argument) 
				short_opts[j++] = ':';
			short_opts[j] = '\0';
		}
	}
	in_flags = 0;
	list_verbose = 0;
	ha_state = xt_rtpproxy_ACTIVE;
	while (optind < argc) {
		int option_index;
		int c;
		option_index = -1;
		c = getopt_long_only (argc, argv, short_opts, long_options, &option_index);
		if (c == -1) {
			if (optind >= argc) 
				break;

			if (!cmd) {
				struct enum_entry *e;
				e = enum_by_name(argv[optind], enum_command);
				if (!e) {
					fprintf(stderr, "ERROR: Unknown command: %s\n", argv[optind]);
					RETURN(2);
				}
				cmd = e->id;
				optind++;
			} 
			else {
				fprintf(stderr, "ERROR: Multiple commands not allowed (%s)\n", argv[optind]);
				RETURN(2);
			}
		}
		else {
			if (cmd == cmdDelete || cmd==cmdList || cmd==cmdHaState || cmd==cmdRetrieve) {
				/* addr-a/b, port-a/b are aliases */
				if ((c & CMDVAL_GRP_MASK) == CMDVAL_GRP_ADDR) {
					c &= ~CMDVAL_A_B;
				}				
			}
			if ((pending_switchboard == 1 && c != (CMDVAL_GRP_ADDR|CMDVAL_ADDR_PORT)) || 
			    (pending_switchboard == 2 && c != (CMDVAL_GRP_ADDR|CMDVAL_ADDR_PORT|CMDVAL_A_B))) {
				PUSH_SWITCHBOARD;
				pending_switchboard = 0;
			}

			switch ((c & CMDVAL_GRP_MASK)?c & CMDVAL_GRP_MASK:c) {
				case CMDVAL_GRP_ADDR:
					if (cmd == cmdNone) {
						fprintf(stderr, "ERROR: Command not yet specified (%s)\n", argv[optind]);
						RETURN(2);
					}
					switch (cmd) {
						case cmdDelete:
						case cmdList:
						case cmdAllocGroup:
						case cmdHaState:
						case cmdRetrieve:
							if ((c & CMDVAL_ADDR_PORT) && !cur_switchboard.ip) {
								fprintf(stderr, "ERROR: Port does not follow ip (%s)\n", argv[optind]);
								RETURN(2);
							}
							pending_switchboard = 1;
							if (cmd == cmdAllocGroup && (c & CMDVAL_A_B) != 0) {
								pending_switchboard++;
							}
							if ((c & CMDVAL_ADDR_PORT) != 0) {
								cur_switchboard.port = s2u(optarg);
								PUSH_SWITCHBOARD;
								pending_switchboard = 0;
							}
							else {
								cur_switchboard.ip = s2ip(optarg);
							}
							break;
						default:
							CLEAR_SESSION;
							if ((c & CMDVAL_ADDR_PORT) != 0)
								cur_session.dir[(c & CMDVAL_A_B) != 0].switchboard.addr.port = s2u(optarg);
							else
								cur_session.dir[(c & CMDVAL_A_B) != 0].switchboard.addr.ip = s2ip(optarg);							
							break;							
					}
					break;
				case CMDVAL_GRP_RADDR:
					CLEAR_SESSION;
					if ((c & CMDVAL_LEARN_PORT_ONLY) != 0)
						cur_session.dir[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].flags |= XT_RTPPROXY_SOCKOPT_FLAG_LEARN_PORT_ONLY;
					else {
						if ((c & CMDVAL_ADDR_PORT) != 0)
							cur_session.dir[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].source.port = s2u(optarg);
						else
							cur_session.dir[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].source.ip = s2ip(optarg);
						cur_session.dir[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].flags |= XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR;
					}
					break;
				case CMDVAL_GRP_LEARNING_TIMEOUT:
					CLEAR_SESSION;
					cur_session.dir[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].learning_timeout = s2u(optarg);
					cur_session.dir[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].flags |= XT_RTPPROXY_SOCKOPT_FLAG_SESSION_LEARNING_TIMEOUT;
					break;
				case CMDVAL_GRP_ALWAYS_LEARN:
					CLEAR_SESSION;
					cur_session.dir[(c & CMDVAL_A_B) != 0].always_learn = s2u(optarg);
					cur_session.dir[(c & CMDVAL_A_B) != 0].flags |= XT_RTPPROXY_SOCKOPT_FLAG_ALWAYS_LEARN;
					break;
				case CMDVAL_GRP_THROTTLE_MARK:
					CLEAR_SESSION;
					cur_session.dir[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].throttle.mark = s2u(optarg);
					cur_session.dir[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].flags |= XT_RTPPROXY_SOCKOPT_FLAG_THROTTLE_MARK;
					break;
				case CMDVAL_GRP_THROTTLE_BANDWIDTH:
					CLEAR_SESSION;
					if (c & CMDVAL_BYTES_PACKETS) {
						cur_session.dir[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].throttle.max_bandwidth.packets = s2u(optarg);
					} else {
						cur_session.dir[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].throttle.max_bandwidth.bytes = s2u(optarg);
					}
					cur_session.dir[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].flags |= XT_RTPPROXY_SOCKOPT_FLAG_THROTTLE_BANDWIDTH;
					break;
				case CMDVAL_GRP_SESSION_ID:
					CLEAR_SESSION;
					cur_session.dir[(c & CMDVAL_A_B) != 0].sess_id = s2u(optarg);
					break;
				case CMDVAL_GRP_SESSION_CREATED:
					CLEAR_SESSION;
					cur_session.sh.created = s2u(optarg);
					break;
				case CMDVAL_GRP_EXPIRES:
					CLEAR_SESSION;
					cur_session.sh.expires_timeout = s2u(optarg);
					cur_session.sh.flags |= XT_RTPPROXY_SOCKOPT_FLAG_SESSION_EXPIRES;
					break;
				case CMDVAL_GRP_TTL:
					CLEAR_SESSION;
					cur_session.sh.ttl = s2u(optarg);
					cur_session.sh.flags |= XT_RTPPROXY_SOCKOPT_FLAG_SESSION_TTL;
					break;
				case CMDVAL_GRP_HA_STATE:
					ha_state = s2u(optarg);
					if (ha_state > xt_rtpproxy_STANDBY) {
						ha_state = xt_rtpproxy_STANDBY;
					}
					break;
				case CMDVAL_GRP_PACKAGE_FILE:
					package_file = optarg;
					break;
				case CMDVAL_GRP_INTERVAL:
					switch (c & 0x0F) {
						case 0:
							interval.from_stamp = s2u(optarg);
							break;
						case 1:
							interval.to_stamp = s2u(optarg);
							break;
						default:
							;
					}
					break;
				case CMDVAL_GRP_DESTROY:
					CLEAR_SESSION;
					cur_session.sh.flags |= XT_RTPPROXY_SOCKOPT_FLAG_SESSION_DESTROY;
					break;
				case CMDVAL_GRP_LIST:
					switch (c & 0x0F) {
						case 0:
							list_verbose |= XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS;
							break;
						default: ;
					}
					break;
				case CMDVAL_GRP_RESET:
					switch (c & 0x0F) {
						case 0:
							in_flags |= XT_RTPPROXY_SOCKOPT_FLAG_RESET_GLOBAL_STAT;
							break;
						case 1:
							in_flags |= XT_RTPPROXY_SOCKOPT_FLAG_RESET_SWITCHBOARD_STAT;
							break;
						case 2:
							in_flags |= XT_RTPPROXY_SOCKOPT_FLAG_RESET_SWITCHBOARD_PACKET_STAT;
							break;
						default: ;
					}
					break;
				case CMDVAL_GRP_FORMAT: {
					struct enum_entry *e;
					e = enum_by_name(optarg, enum_format_option);
					if (!e) {
						fprintf(stderr, "ERROR: Unknown format option: %s\n", optarg);
						RETURN(2);
					}
					output_format = e->id;
					break;
				}
				case CMDVAL_GRP_NEW_SESSION:
					if (pending_session == 1 || (pending_session == 2 && (cmd==cmdAlloc || cmd==cmdAllocGroup)))  {
						/* we can put multiple new sessions, may be handy when allocating autolearning sessions */
						PUSH_ARRAY(sessions, session_count, cur_session);
						pending_session = 2;
					}
					break;
				case CMDVAL_GRP_RPC:
					switch (c & 0x0F) {
						case CMDVAL_RPC_PROTO:
							if (strcasecmp(optarg, "udp") == 0) {
								rpc_params.proto = xt_rtpproxy_connection_UDP;
							} 
							else if (strcasecmp(optarg, "tcp") == 0) {
								rpc_params.proto = xt_rtpproxy_connection_TCP;
							} 
							/* else if (strcasecmp(optarg, "raw") == 0) {
								rpc_params.proto = xt_rtpproxy_connection_RAW;
							} */
							else {
								fprintf(stderr, "ERROR: Unknown protocol: %s\n", optarg);
								RETURN(2);
							}
							rpc_params_flag = 1;
							break;
						case CMDVAL_RPC_ADDR:
							rpc_params.addr = s2ip(optarg);
							rpc_params_flag = 1;
							break;
						case CMDVAL_RPC_PORT:
							rpc_params.port = s2u(optarg);
							rpc_params_flag = 1;
							break;
						case CMDVAL_RPC_REQUEST_SIZE:
							rpc_params.request_size = s2u(optarg);
							break;
						case CMDVAL_RPC_REPLY_SIZE:
							rpc_params.reply_size = s2u(optarg);
							break;
						case CMDVAL_RPC_TOTAL_TIMEOUT:
							rpc_params.total_timeout = s2u(optarg);
							break;
						case CMDVAL_RPC_UDP_RETRY_TIMEOUT:
							rpc_params.udp_retry_timeout = s2u(optarg);
							break;
						default:
							;
					}

					break;
				case 'v':
					verbose++;
					break;
				case 'V':
					print_version();
					printf("Userspace size of pointer: %d Bytes, %dbits\n", (int) sizeof(void*), 8*(int) sizeof(void*));
					RETURN(0);
				case 'h':
				case 'H':
					help_flag = 1;
					break;
				case '?':
					break;
				default:
					fprintf(stderr, "getopt_long_only returned character code 0x%x\n", c);
					RETURN(2);
			}
		}
		if (cmd && help_flag)
			break;
	}
	if (!cmd || help_flag) {
		help(cmd);
		RETURN(help_flag?0:2);
	}
	if (pending_switchboard) {
		PUSH_SWITCHBOARD;
	}
	if (pending_session == 1 || (pending_session == 2 && (cmd==cmdAlloc || cmd==cmdAllocGroup))) {
		PUSH_ARRAY(sessions, session_count, cur_session);
	}

	SIGNAL(SIGTERM);
	SIGNAL(SIGINT);
	SIGNAL(SIGQUIT);

	if ((switchboard_a_count + switchboard_b_count == 0) && cmd == cmdAllocGroup) {
		fprintf(stderr, "ERROR: switchboard definition is missing\n");		
		RETURN(2);
	}
	switch (cmd) {
		case cmdAlloc:
		case cmdAllocGroup:
		case cmdUpdate:
		case cmdSessionInfo:
			if (!session_count) {
				fprintf(stderr, "ERROR: session definition is missing\n");
				RETURN(2);
			}
			if (cmd != cmdAllocGroup) {
				for (i=0; i<session_count; i++) {
					for (j=0; j<2; j++) {
						if ((*sessions)[i].dir[j].switchboard.addr.ip && (*sessions)[i].dir[j].switchboard.addr.port) {
							goto sess_ok;
						}
					}
					fprintf(stderr, "ERROR: uncomplete session switchboard identification (#%d)\n", i);
					RETURN(2);
				sess_ok:
					;
				}
			}
			break;
		default:
			;
	}

	/* switchboard may be optional but when at least one port or ip is defined then complete definition is mandatory */
	for (i=0; i<switchboard_a_count; i++) {
		if (!(*switchboards_a)[i].ip || !(*switchboards_a)[i].port) {
			fprintf(stderr, "ERROR: uncomplete switchboard-A identification (#%d)\n", i);
			RETURN(2);
		}
	}
	for (i=0; i<switchboard_b_count; i++) {
		if (!(*switchboards_b)[i].ip || !(*switchboards_b)[i].port) {
			fprintf(stderr, "ERROR: uncomplete switchboard-B identification (#%d)\n", i);
			RETURN(2);
		}
	}
	if (interval.from_stamp && interval.to_stamp) {
		if (interval.from_stamp > interval.to_stamp) {
			u_int32_t x;
			x = interval.to_stamp;
			interval.to_stamp = interval.from_stamp;
			interval.from_stamp = x;
		}
	}
	
	if (cmd == cmdRetrieve || cmd == cmdReplicate) {
		if (!package_file || !strlen(package_file)) {
			fprintf(stderr, "ERROR: package file not specified\n");
			RETURN(2);
		}		
	}

	if (cmd == cmdRpcServer) {
		if (verbose) fprintf(stderr, "xt_RTPPROXY_rpc_svc_run({proto:%d,addr:%s,port:%d,request_size:%d,reply_size:%d})\n",
			rpc_params.proto, ip2s(rpc_params.addr), rpc_params.port, 
			rpc_params.request_size, rpc_params.reply_size
		);
		if (xt_RTPPROXY_rpc_svc_run(&handle, &rpc_params) < 0) {
			goto err; /* if OK never returns unless raw */
		}
	}
	if (rpc_params_flag) {
		if (verbose) fprintf(stderr, "xt_RTPPROXY_open(REMOTE, {proto:%d,addr:%s,ip:%d,request_size:%d,reply_size:%d,total_timeout:%d,udp_retry_timeout:%d})\n",
			rpc_params.proto, ip2s(rpc_params.addr), rpc_params.port, 
			rpc_params.request_size, rpc_params.reply_size,
			rpc_params.total_timeout, rpc_params.udp_retry_timeout

		);
		if (xt_RTPPROXY_open(&handle, xt_rtpproxy_REMOTE, &rpc_params) < 0) goto err;
	}
	else {
		if (verbose) fprintf(stderr, "xt_RTPPROXY_open(LOCAL)\n");
		if (xt_RTPPROXY_open(&handle, xt_rtpproxy_LOCAL, NULL) < 0) goto err;
	}
	
	if (output_format == fmtXml) printf("<iptrtpproxy>\n");
	switch (cmd) {
		case cmdInfo: {
			struct xt_rtpproxy_sockopt_out_config out_config;
			if (verbose) fprintf(stderr, "xt_RTPPROXY_get_config(flags:0x%x)\n", in_flags);

			if (xt_RTPPROXY_get_config(&handle, 0, NULL, xt_rtpproxy_CONFIG, in_flags, &out_config) < 0)
				goto err;
			if (verbose) print_format();
			if (verbose && output_format!=fmtTab) print_header(&out_config);
			print_info(&out_config);
			if (verbose && output_format!=fmtTab) print_foot();
			break;
		}
		case cmdList: {
			struct xt_rtpproxy_sockopt_out_config out_config;
			xt_rtpproxy_sockopt_count out_switchboard_count;

			if (!switchboard_a_count) {
				in_flags |= XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS;
			}
			in_flags |= list_verbose;

			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_get_config(\n");
				print_in_switchboards("", switchboard_a_count, switchboards_a); fprintf(stderr, ",\n");
				fprintf(stderr, "  flags: 0x%x\n", in_flags); 
				fprintf(stderr, ")\n");
			}
			if (xt_RTPPROXY_get_config(&handle, switchboard_a_count, switchboards_a, 
						(in_flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS)? xt_rtpproxy_SESSIONS:xt_rtpproxy_SWITCHBOARDS,
						in_flags, &out_config) < 0)
				goto err;
			if (xt_RTPPROXY_read_from_buf(&handle, sizeof(out_switchboard_count), &out_switchboard_count) < 0) 
				goto err;

			if (verbose) print_format(); 
			if (verbose && output_format!=fmtTab) print_header(&out_config);
			if (verbose && output_format!=fmtTab) print_info(&out_config);

			switch (output_format) {
				case fmtXml:
					printf("<switchboards>\n");
					break;
				case fmtTab:
					if (in_flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS) {
						print_out_sessions_header(posFull, "");
					}
					else {
						printf("ip\tport\tmax-sess\tresurrection-timeout\tha-state\talloc\t");
						print_stat_title("");
						for (j=0; j<xt_rtpproxy_MAX_STATES; j++) {
							printf("\tstat/%s", state_s[j]);
						}
						printf("\tcounter/alloc\tcounter/alloc-failed\tcounter/updated\tcounter/destroyed\tcounter/staled\tcounter/expired");
						printf("\tthrottle/mark/num-session\tthrottle/mark/counter/packets\tthrottle/mark/counter/bytes");
						printf("\n");
					}
					break;
				default:
					printf("{\n");
					break;
			}
			while (out_switchboard_count) {
				struct xt_rtpproxy_sockopt_switchboard out_switchboard;
				xt_rtpproxy_sockopt_count out_session_count;
				if (xt_RTPPROXY_read_from_buf(&handle, sizeof(out_switchboard), &out_switchboard) < 0) 
					goto err;
				if (xt_RTPPROXY_read_from_buf(&handle, sizeof(out_session_count), &out_session_count) < 0) 
					goto err;
								
				switch (output_format) {
					case fmtXml:
						printf("  <switchboard ip=\"%s\" port=\"%d\" max-sess=\"%d\" resurrection-timeout=\"%d\" ha-state=\"%d\" alloc=\"%d\">\n", 
               				ip2s(out_switchboard.addr.ip),
			                out_switchboard.addr.port,
							out_switchboard.max_sess, 
							out_switchboard.resurrection_timeout,
							out_switchboard.ha_state,
							out_switchboard.stat.num_session
					    ); /* ip2s ... result in static buffer */

						printf("    "); print_stat(&out_switchboard.stat.traffic);
						printf("    <state");
							for (j=0; j<xt_rtpproxy_MAX_STATES; j++) {
								printf(" %s=\"%u\"", state_s[j], out_switchboard.stat.state[j]);
							}
						printf(" />\n");
						printf("    <counter alloc=\"%u\" alloc-failed=\"%u\" updated=\"%u\" destroyed=\"%u\" staled=\"%u\" expired=\"%u\" />\n",
							out_switchboard.stat.counter.alloc,
							out_switchboard.stat.counter.alloc_failed,
							out_switchboard.stat.counter.updated,
							out_switchboard.stat.counter.destroyed,
							out_switchboard.stat.counter.staled,
							out_switchboard.stat.counter.expired 
						);
						printf("    <throttle_mark num-session=\"%u\" packets=\"%u\" bytes=\"%u\" />\n",
							out_switchboard.stat.throttle.num_mark_session,
							out_switchboard.stat.throttle.mark_counter.packets,
							out_switchboard.stat.throttle.mark_counter.bytes
						);
						if (out_session_count) {
							print_out_sessions_header(posFull, "    ");
							
							while (out_session_count) {
								struct xt_rtpproxy_sockopt_session out_session;
								if (xt_RTPPROXY_read_from_buf(&handle, sizeof(out_session), &out_session) < 0) 
									goto err;
								print_out_session(&out_session, posFull, "      ");
								
								out_session_count--;
							}
							print_out_sessions_foot("    ");
						}
							
						printf("  </switchboard>\n");
						break;

					case fmtTab:
						if (in_flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS) {
							while (out_session_count) {
								struct xt_rtpproxy_sockopt_session out_session;
								if (xt_RTPPROXY_read_from_buf(&handle, sizeof(out_session), &out_session) < 0) 
									goto err;

								print_out_session(&out_session, posFull, "");	
								out_session_count--;
							}

						}
						else {

							printf("%s\t%d\t", 
								ip2s(out_switchboard.addr.ip),
								out_switchboard.addr.port
							); /* ip2s ... result in static buffer */
							printf("%d\t%d\t%d\t%d\t", 
								out_switchboard.max_sess, 
								out_switchboard.resurrection_timeout,
								out_switchboard.ha_state,
								out_switchboard.stat.num_session
							);
							print_stat(&out_switchboard.stat.traffic); printf("\t");

							for (j=0; j<xt_rtpproxy_MAX_STATES; j++) {
								printf("%u\t", out_switchboard.stat.state[j]);
							}
							printf("%u\t%u\t%u\t%u\t%u\t%u\t%u\t%u\t%u",
								out_switchboard.stat.counter.alloc,
								out_switchboard.stat.counter.alloc_failed,
								out_switchboard.stat.counter.updated,
								out_switchboard.stat.counter.destroyed,
								out_switchboard.stat.counter.staled,
								out_switchboard.stat.counter.expired,
								out_switchboard.stat.throttle.num_mark_session,
								out_switchboard.stat.throttle.mark_counter.packets,
								out_switchboard.stat.throttle.mark_counter.bytes
							);
							printf("\n");

							while (out_session_count) {
								struct xt_rtpproxy_sockopt_session out_session;
								if (xt_RTPPROXY_read_from_buf(&handle, sizeof(out_session), &out_session) < 0) 
									goto err;
								out_session_count--;
							}
						}
						break;

					default:
						printf("  {\n");
						printf("    switchboard: {\n");
						printf("      ip: %s, port: %d, max-sess: %d, resurrection-timeout: %d, ha-state: %d, alloc: %d\n",
							ip2s(out_switchboard.addr.ip), 
							out_switchboard.addr.port,
							out_switchboard.max_sess, 
							out_switchboard.resurrection_timeout,
							out_switchboard.ha_state,
							out_switchboard.stat.num_session
						); /* ip2s ... result in static buffer */
						printf("      traffic: {"); print_stat(&out_switchboard.stat.traffic); printf("}\n");
						printf("      state: {");
						for (j=0; j<xt_rtpproxy_MAX_STATES; j++) {
							if (j) printf(", ");
							printf("%s: %u", state_s[j], out_switchboard.stat.state[j]);
						}
						printf("}\n");
						printf("      counter: {alloc: %u, alloc-failed: %u, updated: %u, destroyed: %u, staled: %u, expired: %u}\n",
							out_switchboard.stat.counter.alloc,
							out_switchboard.stat.counter.alloc_failed,
							out_switchboard.stat.counter.updated,
							out_switchboard.stat.counter.destroyed,
							out_switchboard.stat.counter.staled,
							out_switchboard.stat.counter.expired 
						);
						printf("      throttle-mark: {num-session: %u, packets: %u, bytes: %u}\n",
							out_switchboard.stat.throttle.num_mark_session,
							out_switchboard.stat.throttle.mark_counter.packets,
							out_switchboard.stat.throttle.mark_counter.bytes
						);
						printf("    }\n");
						if (out_session_count) {
							print_out_sessions_header(posFull, "    ");

							while (out_session_count) {
								struct xt_rtpproxy_sockopt_session out_session;
								if (xt_RTPPROXY_read_from_buf(&handle, sizeof(out_session), &out_session) < 0) 
									goto err;
							
								print_out_session(&out_session, posFull, "      ");

								out_session_count--;
							}
							print_out_sessions_foot("    ");
						}
						printf("  }\n");
						break;
				}
				
				out_switchboard_count--;
			}	
			switch (output_format) {
				case fmtXml:
					printf("</switchboards>\n");
					break;
				case fmtTab:
					break;
				default:
					printf("}\n");
					break;
			}

			if (verbose && output_format!=fmtTab) print_foot();
			break;
		}
		case cmdSessionInfo: {
			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_update_sessions(\n");
				print_in_sessions(session_count, sessions, 1); fprintf(stderr, "\n");
				fprintf(stderr, ")\n");
			}
			for (i = 0; i < session_count; i++) {
				(*sessions)[i].sh.flags |= XT_RTPPROXY_SOCKOPT_FLAG_SESSION_INFO;
			}
			if (xt_RTPPROXY_update_sessions(&handle, session_count, sessions
				) < 0)
				goto err; 
			if (verbose) print_format(); 
			print_out_sessions(session_count, sessions, posFull || posAlloc);
			if (verbose && output_format!=fmtTab) print_foot();
			break;
		}
		case cmdAllocGroup:
		case cmdAlloc: {

			for (i=0; i<session_count; i++) {
				for (j=0; j<2; j++) {
					(*sessions)[i].dir[j].flags |= XT_RTPPROXY_SOCKOPT_FLAG_ALWAYS_LEARN;
				}
			}
			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_alloc_sessions(\n");
				if (switchboard_a_count) {
					print_in_switchboards("-a", switchboard_a_count, switchboards_a); fprintf(stderr, ",\n");
				}
				if (switchboard_b_count) {
					print_in_switchboards("-b", switchboard_b_count, switchboards_b); fprintf(stderr, ",\n");
				}
				print_in_sessions(session_count, sessions, cmd==cmdAlloc); fprintf(stderr, "\n");
				fprintf(stderr, ")\n");
			}
			if (xt_RTPPROXY_alloc_sessions(&handle, 
						switchboard_a_count, switchboards_a, 
						switchboard_b_count, switchboards_b, 
						session_count, sessions
				) < 0)
				goto err; 
			if (verbose) print_format(); 
			print_out_sessions(session_count, sessions, posAlloc);
			break;
		}
		case cmdUpdate:
			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_update_sessions\n");
				print_in_sessions(session_count, sessions, 1); fprintf(stderr, "\n");
				fprintf(stderr, ")\n");
			}
			if (xt_RTPPROXY_update_sessions(&handle, session_count, sessions) < 0)
				goto err;
			if (verbose) print_format(); 
			print_out_sessions(session_count, sessions, posFull || posAlloc);
			if (verbose && output_format!=fmtTab) print_foot();
			break;

		case cmdDelete:
			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_delete_switchboard_sessions(\n");
				print_in_switchboards("", switchboard_a_count, switchboards_a); fprintf(stderr, "\n");
				fprintf(stderr, ")\n");
			}
			if (xt_RTPPROXY_delete_switchboard_sessions(&handle, switchboard_a_count, switchboards_a) < 0)
				goto err;
			break;

		case cmdHaState:
			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_set_ha_state(\n");
				fprintf(stderr, "  ha-state: %d,\n", ha_state);
				print_in_switchboards("", switchboard_a_count, switchboards_a); fprintf(stderr, "\n");
				fprintf(stderr, ")\n");
			}
			if (xt_RTPPROXY_set_ha_state(&handle, ha_state, switchboard_a_count, switchboards_a) < 0)
				goto err;
			break;

		case cmdRetrieve: {
			u_int32_t buf_size;
			void *buf;
			FILE *f;
			struct xt_rtpproxy_replication_package_info info;
			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_retrieve(\n");
				fprintf(stderr, "  package-file: '%s',\n", package_file);
				fprintf(stderr, "  stamp: [%u, %u]\n", interval.from_stamp, interval.to_stamp);
				print_in_switchboards("", switchboard_a_count, switchboards_a); fprintf(stderr, "\n");
				fprintf(stderr, ")\n");
			}
			if (xt_RTPPPROXY_retrieve_sessions(&handle, &interval, switchboard_a_count, switchboards_a, &buf_size, &buf) < 0)
				goto err;
			if (xt_RTPPPROXY_get_replication_package_info(&handle, buf_size, buf, &info) < 0)
				goto err;
			print_replication_package_info(&info);
			f = fopen(package_file, "w");
			if (!f) {
				fprintf(stderr, "ERROR: file '%s' open error (%d)\n", package_file, errno);
				goto err2;
			}
			fwrite(buf, 1, buf_size, f);
			fclose(f);
			break;
		}
		case cmdReplicate: {
			u_int32_t buf_size;
			void *buf;
			FILE *f;
			struct xt_rtpproxy_replication_package_info info;
			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_replicate(package-file: '%s')\n", package_file);
			}
			f = fopen(package_file, "r");
			if (!f) {
				fprintf(stderr, "ERROR: file '%s' open error (%d)\n", package_file, errno);
				goto err2;
			}
			if (fseek(f, 0, SEEK_END) == -1) {
				fprintf(stderr, "ERROR: file '%s' seek error (%d)\n", package_file, errno);
				fclose(f);
				goto err2;
			};
			buf_size = ftell(f);
			buf = malloc(buf_size);
			if (!buf) {
				fprintf(stderr, "ERROR: cannot allocate %d bytes\n", buf_size);
				fclose(f);
				goto err2;
			}
			rewind(f);
			fread(buf, 1, buf_size, f);
			fclose(f);
			if (xt_RTPPPROXY_replicate_sessions(&handle, buf_size, buf) < 0) {
				free(buf);
				goto err;
			}
			if (xt_RTPPPROXY_get_replication_package_info(&handle, buf_size, buf, &info) < 0)
				goto err;
			print_replication_package_info(&info);
			free(buf);
			break;
		}
	}
	if (output_format == fmtXml) printf("</iptrtpproxy>\n");

/* destroy: */
	if (verbose) fprintf(stderr, "xt_RTPPROXY_close()\n");
	xt_RTPPROXY_close(&handle);
	if (verbose) fprintf(stderr, "Done\n");
	RETURN(0);

err:
	fprintf(stderr, "ERROR: %s (%d)\n", handle.err_str, handle.err_no);
err2:
	if (verbose) fprintf(stderr, "xt_RTPPROXY_close()\n");
	xt_RTPPROXY_close(&handle);
	if (verbose) fprintf(stderr, "Done\n");
	RETURN(1);
}

#if !defined(NO_SHARED_LIBS) || NO_SHARED_LIBS==0
/* make compiler happy and give it missing symbols */
#ifdef IPT_RTPPROXY_IPTABLES_API
#include <iptables.h>
#undef IPT_RTPPROXY_IPTABLES_API
#else
#include <xtables.h>
#endif

#ifdef _IPTABLES_COMMON_H
/* old iptables API, it uses iptables_common.h (instead of xtables.h) included from iptables.h */
/* #ifndef XTABLES_VERSION ... optional test */
#define IPT_RTPPROXY_IPTABLES_API 1
#endif

#ifdef IPT_RTPPROXY_IPTABLES_API
void register_target(struct iptables_target *me) {
}
#else
void xtables_register_target(struct xtables_target *me) {
}
#endif

#if IPT_RTPPROXY_IPTABLES_API
void exit_error(enum exittype status, char *msg, ...)
#else
void exit_error(enum exittype status, const char *msg, ...)
#endif
{
	va_list args;
	
	va_start(args, msg);
	fprintf(stderr, "%s v%s: ", program_name, XT_RTPPROXY_VERSION);
	vfprintf(stderr, msg, args);
	va_end(args);
	fprintf(stderr, "\n");
	exit(status);
}
																		
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

int check_inverse(const char option[], int *invert, int *optind, int argc)
{
	if (option && strcmp(option, "!") == 0) {
		if (*invert)
			exit_error(PARAMETER_PROBLEM, "Multiple `!' flags not allowed");
		*invert = TRUE;
		if (optind) {
			*optind = *optind+1;
			if (argc && *optind > argc)
				exit_error(PARAMETER_PROBLEM, "no argument following `!'");
		}
		return TRUE;
	}
	return FALSE;
}

#endif

