/***************************************************************************
 *   (C) 2007 by Tomas Mandys <tomas.mandys at 2p dot cz>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License           *
 *   version 2.1 as published by the Free Software Foundation;             *
 *                                                                         *
 ***************************************************************************/

/* Header file for iptables xt_RTPPROXY target */

#ifndef _XT_RTPPROXY_H_target
#define _XT_RTPPROXY_H_target

#ifdef __KERNEL__
#include <linux/spinlock_types.h>
#endif

#define PACKED __attribute__ ((packed));

/* version updated by Makefile */
#define XT_RTPPROXY_VERSION "0.7"
#define XT_RTPPROXY_VERSION_HEX 0x07

enum xt_rtpproxy_session_state {
	xt_rtpproxy_NONE,	/* ready for allocation */
	xt_rtpproxy_INIT1,	/* half initialized session, we know caller IP|learning */
	xt_rtpproxy_INIT2,	/* full initialized session, we know caller&callee IP|learning, ready to forward */
	xt_rtpproxy_FORWARD1,	/* at least one packet forwarded caller->callee or callee->caller */
	xt_rtpproxy_FORWARD2,	/* packets forwarded in both directions */
	xt_rtpproxy_DESTROYED,	/* destroyed by sockopt command */
	xt_rtpproxy_STALED,		/* TTL exceeded */
	xt_rtpproxy_EXPIRED 	/* expired, no packet for long time (expires_timeout) */
};

enum xt_rtpproxy_ha_state {
	xt_rtpproxy_ACTIVE = 0,
	xt_rtpproxy_STANDBY = 1
};

#define xt_rtpproxy_MAX_STATES (xt_rtpproxy_EXPIRED+1)

struct xt_rtpproxy_addr {
	/* protocol: UDP only */
	u_int32_t ip;
	u_int16_t port;
} PACKED;

#define xt_rtpproxy_switchboard_id xt_rtpproxy_addr

struct xt_rtpproxy_stat {
	u_int32_t packets_in;
	u_int32_t packets_out;
	u_int32_t bytes_in;
	u_int32_t bytes_out;
} PACKED;

struct xt_rtpproxy_throttle_stat {
	u_int32_t packets;
	u_int32_t bytes;
} PACKED;

struct xt_rtpproxy_switchboard;

struct xt_rtpproxy_sockopt_session_sh {
	u_int8_t/*enum xt_rtpproxy_session_state*/ state;

	u_int32_t created; /* search key sess_id+created, created used to randomize key to avoid malicious manipulation */
	u_int16_t expires_timeout;
	u_int32_t expires;
	u_int16_t ttl;
	
	u_int32_t last_change_stamp;  /* for replication */

	u_int32_t flags;
} PACKED;

/* needed defined as extra struct because if defined as array inside xt_rtpproxy_sockopt_session_dir then array was not packed */
struct xt_rtpproxy_sockopt_session_stream {
	/* we need keep both RTP&RTCP ports because two consecutive port may be translated by NAT to nonconsecutive ones */
	u_int16_t port;  /* port allocated at gate (for sockopt) */
	struct xt_rtpproxy_addr source;
	struct xt_rtpproxy_addr source_learnt;
	u_int32_t learning;
	u_int32_t last_packet_stamp;
	u_int16_t learning_timeout;
	struct {	
		struct xt_rtpproxy_throttle_stat max_bandwidth;
		u_int32_t mark;
		u_int32_t stamp;
		struct xt_rtpproxy_throttle_stat bandwidth;
		struct xt_rtpproxy_throttle_stat counter;
		struct xt_rtpproxy_throttle_stat mark_counter;
	} throttle;
	u_int32_t flags;
} PACKED;

struct xt_rtpproxy_sockopt_session_dir {
	u_int16_t sess_id;
	struct {	
		struct xt_rtpproxy_switchboard_id addr;	
		u_int16_t free;    /* these fields may be used for load balancing, when allocated number of sessions is retrieved to balance next allocation */
		u_int16_t alloc;
	} switchboard;
	struct xt_rtpproxy_sockopt_session_stream stream[2];
	
	u_int8_t always_learn;

	struct xt_rtpproxy_stat stat;
	
	u_int32_t flags;
} PACKED;

#ifdef __KERNEL__
struct xt_rtpproxy_session {
	struct xt_rtpproxy_sockopt_session_sh so;
	struct xt_rtpproxy_sockopt_session_dir so_dir;
	struct xt_rtpproxy_session *session2;
	
	u_int8_t dir;  /* 1..gate A, 0..gate B,  needed to ge proper session lock, lock is acquired at the gateA side only, xt_rtpproxy_sockopt_session_sh are maintained at A side only */
	spinlock_t lock;
	u_int32_t retrieve_counter; /* used to avoid double retrieving sessions */

	struct xt_rtpproxy_switchboard *switchboard;
};
#endif

struct xt_rtpproxy_sockopt_switchboard {
	struct {
		u_int16_t state[xt_rtpproxy_MAX_STATES]; /* gather on request */
		struct xt_rtpproxy_stat traffic;  /* traffic is gathered from switchboards on demand in sockopt to keep switchboard lockless as much as possible */
		struct {
			struct xt_rtpproxy_throttle_stat mark_counter;
			u_int16_t num_mark_session;
		} throttle;
		u_int16_t num_session;   /* num of allocated session */
		struct {
			u_int32_t alloc;
			u_int32_t alloc_failed;
			u_int32_t updated;
			u_int32_t destroyed;
			u_int32_t staled;
			u_int32_t expired;
		} counter;
	} stat;
	struct xt_rtpproxy_switchboard_id addr;
	u_int16_t max_sess;	/* max number of handled sessions */
	u_int16_t resurrection_timeout; /* time that the session is not possible activate after death */
	u_int8_t /* enum xt_rtpproxy_ha_state */ ha_state;
} PACKED;

struct xt_rtpproxy_switchboard {
	struct xt_rtpproxy_sockopt_switchboard so; /* we need it for both identification and sockopt */
	struct xt_rtpproxy_session *sessions;
	/* session alloc list maintanance */
	u_int16_t usage;
	u_int16_t next_sess_id;
	struct {
		u_int16_t rate;
		u_int16_t counter;
		u_int32_t stamp;
	} cps;  /* calls per sec to predict next_sess_id after HA switchover */
	struct xt_rtpproxy_switchboard *prev, *next;
};

struct xt_rtpproxy_target_info {
	struct xt_rtpproxy_switchboard_id addr;
	u_int16_t max_sess;	/* max number of handled sessions */
	/* end of part for userspace comparision, if changes adjust userspacesize in libipt_RTPPROXY.c */
	u_int16_t resurrection_timeout; /* time that the session is not possible activate after death */
	u_int8_t /* enum xt_rtpproxy_ha_state */ ha_state;

	u_int32_t libipt_flags; /* aux flags for libipt */
#ifdef KERNEL_64_USERSPACE_32
	/* placeholder */
	u_int64_t switchboard;
#else
	/* Used internally by the kernel */
	#ifdef __KERNEL__
	struct xt_rtpproxy_switchboard *switchboard;
	#else
	void *switchboard;
	#endif
#endif
};

#ifdef KERNEL_64_USERSPACE_32
	/* placeholder */
	typedef u_int64_t xt_rtpproxy_handle;
#else
	/* Used internally by the kernel */
	#ifdef __KERNEL__
	typedef struct xt_rtpproxy_switchboard* xt_rtpproxy_handle;
	#else
	typedef void* xt_rtpproxy_handle;
	#endif
#endif

#define XT_RTPPROXY_MAX_GET_SWITCHBOARD 20 /* max.number of switchboards to get info in one request */
#define XT_RTPPROXY_MAX_ALLOC_SESSION 20 /* max.number of sessions to allocate in one request */

#define XT_RTPPROXY_SO_BASE_CTL (64+1024+128)	/* unique socket command range */

/* data passed via sockopt interface are packed because may be originated at remote machine.
 * Big/low endian correctness is checked by magic. If endian differs then
 * error is returned.
 */

#define XT_RTPPROXY_SO_SET_DELETE_SWITCHBOARD_SESSIONS	(XT_RTPPROXY_SO_BASE_CTL + 0)
/* in:
 *   struct xt_rtpproxy_sockopt_header header;
 *   if (!header.flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS) {
 *     xt_rtpproxy_sockopt_count count;
 *     struct xt_rtpproxy_sockopt_switchboard switchboards[count]:
 *   }
 */
#define XT_RTPPROXY_SO_SET_REPLICATE_SESSIONS	(XT_RTPPROXY_SO_BASE_CTL + 1)
/* in:
 *   u_int8_t buf[];    raw data returned by XT_RTPPROXY_SO_GET_RETRIEVE_SESSIONS
 */
	
#define XT_RTPPROXY_SO_SET_HA_STATE	(XT_RTPPROXY_SO_BASE_CTL + 2)
/* in:
 *   struct xt_rtpproxy_sockopt_header header;
 *   u_int8_t ha_state;
 *   if (!header.flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS) {
 *     xt_rtpproxy_sockopt_count count;
 *     struct xt_rtpproxy_sockopt_switchboard switchboards[count]:
 *   }
 */
#define XT_RTPPROXY_SO_SET_MAX         	XT_RTPPROXY_SO_SET_HA_STATE

#define XT_RTPPROXY_SO_GET_CONFIG		(XT_RTPPROXY_SO_BASE_CTL + 0)
/* in: 
 *   struct xt_rtpproxy_sockopt_header header;
 *   xt_rtpproxy_sockopt_count out_switchboard_count;  0..no switchboards 
 *   xt_rtpproxy_sockopt_count out_session_count;      0..no sessions 
 *   if (!header.flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS) {
 *     xt_rtpproxy_sockopt_count switchboard_count;
 *     xt_rtpproxy_sockopt_switchboard switchboard1[switchboard_count];
 *   }
 * out:
 *   struct xt_rtpproxy_sockopt_out_config config;
 *   xt_rtpproxy_sockopt_count switchboard_count;
 *   struct {
 *     xt_rtpproxy_sockopt_switchboard switchboard;
 *     xt_rtpproxy_sockopt_count session_count;
 *     struct xt_rtpproxy_sockopt_session sessions[session_count]:
 *   } [switchboard_count];
*/
#define XT_RTPPROXY_SO_GET_ALLOC_SESSIONS	(XT_RTPPROXY_SO_BASE_CTL + 1)
/* in: 
 *   struct xt_rtpproxy_sockopt_header header;
 *   xt_rtpproxy_sockopt_count session_count;
 *   if (header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD_GROUPS) {
 *     xt_rtpproxy_sockopt_count switchboard_count1;
 *     xt_rtpproxy_sockopt_switchboard switchboard1[switchboard_count1];
 *     xt_rtpproxy_sockopt_count switchboard_count2;
 *     xt_rtpproxy_sockopt_switchboard switchboard2[switchboard_count2];
 *   } 
 *   xt_rtpproxy_sockopt_session sessions[session_count];
 * out:
 *   xt_rtpproxy_sockopt_count session_count;
 *   xt_rtpproxy_sockopt_session sessions[session_count];
 */

#define XT_RTPPROXY_SO_GET_UPDATE_SESSIONS	(XT_RTPPROXY_SO_BASE_CTL + 2)
/* in:
 *   struct xt_rtpproxy_sockopt_header header;
 *   xt_rtpproxy_sockopt_count count;
 *   struct xt_rtpproxy_sockopt_session sessions[count]:
 * out:
 *   xt_rtpproxy_sockopt_count count;
 *   struct xt_rtpproxy_sockopt_session sessions[count]:
*/

#define XT_RTPPROXY_SO_GET_RETRIEVE_SESSIONS	(XT_RTPPROXY_SO_BASE_CTL + 3)
/* in:
 *   struct xt_rtpproxy_sockopt_header header;
 *   if (!header.flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS) {
 *     xt_rtpproxy_sockopt_count count;
 *     struct xt_rtpproxy_sockopt_switchboard switchboards[count]:
 *   }
 *   if (!header.flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS) {
 *     struct xt_rtpproxy_sockopt_in_interval interval;
 *   }
 * out:
 *   u_int8_t buf[buf_size];
*/
#define XT_RTPPROXY_SO_GET_MAX			XT_RTPPROXY_SO_GET_RETRIEVE_SESSIONS

/* userspace communication */
#define XT_RTPPROXY_SOCKOPT_MAGIC 0xC53A
#define XT_RTPPROXY_SOCKOPT_VERSION 0x03

#define XT_RTPPROXY_SOCKOPT_FLAG_RESET_GLOBAL_STAT		0x0001
#define XT_RTPPROXY_SOCKOPT_FLAG_RESET_SWITCHBOARD_STAT	0x0002
#define XT_RTPPROXY_SOCKOPT_FLAG_RESET_SWITCHBOARD_PACKET_STAT	0x0004
#define XT_RTPPROXY_SOCKOPT_FLAG_SWITCHBOARD_GROUPS 0x0008
#define XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS 0x0010
#define XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS 0x0020
	
#define XT_RTPPROXY_MAX_CMD (XT_RTPPROXY_SO_SET_MAX-XT_RTPPROXY_SO_BASE_CTL+1+XT_RTPPROXY_SO_GET_MAX-XT_RTPPROXY_SO_BASE_CTL+1)

/* error contants for xt_rtpproxy_handle.err_no */
#define XT_RTPPROXY_ERR_CANNOT_OPEN_SOCKET	(-100)
#define XT_RTPPROXY_ERR_BAD_VERSION		(-101)
#define XT_RTPPROXY_ERR_UNEXPECTED		(-102)
#define XT_RTPPROXY_ERR_OUT_OF_MEMORY		(-103)
#define XT_RTPPROXY_ERR_PARAM			(-104)
#define XT_RTPPROXY_ERR_RPC			(-105)
#define XT_RTPPROXY_ERR_BAD_MAGIC		(-106)

struct xt_rtpproxy_sockopt_header {
	u_int16_t magic; /*  catches also little <-> big endian uncompatability when replicating between active <-> standby machines */
	u_int16_t version;
	u_int32_t flags;
} PACKED;

typedef u_int16_t xt_rtpproxy_sockopt_count;

struct xt_rtpproxy_sockopt_num_session {
	u_int32_t num_switchboard;
	u_int32_t num_pair_session;
} PACKED;

struct xt_rtpproxy_sockopt_out_config {
	u_int16_t version;
	u_int32_t started;
	struct {
		u_int32_t ok;
		u_int32_t fail;
		u_int32_t perm;
		u_int32_t unknown;
		u_int32_t cmd[XT_RTPPROXY_MAX_CMD];
	} stat;
	struct xt_rtpproxy_sockopt_num_session total;
} PACKED;

struct xt_rtpproxy_sockopt_in_interval {
	u_int32_t from_stamp;
	u_int32_t to_stamp;
} PACKED;

struct xt_rtpproxy_replication_package_info {
	struct xt_rtpproxy_sockopt_header header;
	struct xt_rtpproxy_sockopt_in_interval interval;
	xt_rtpproxy_sockopt_count record_count;
} PACKED;

#define XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR 0x0001
#define XT_RTPPROXY_SOCKOPT_FLAG_SESSION_LEARNING_TIMEOUT 0x0002
#define XT_RTPPROXY_SOCKOPT_FLAG_SESSION_DESTROY 0x0004
#define XT_RTPPROXY_SOCKOPT_FLAG_SESSION_EXPIRES 0x0008
#define XT_RTPPROXY_SOCKOPT_FLAG_ALWAYS_LEARN 0x0010
#define XT_RTPPROXY_SOCKOPT_FLAG_NOT_FOUND 0x0020
#define XT_RTPPROXY_SOCKOPT_FLAG_SESSION_INFO 0x0040
#define XT_RTPPROXY_SOCKOPT_FLAG_LEARN_PORT_ONLY 0x0080
#define XT_RTPPROXY_SOCKOPT_FLAG_THROTTLE_MARK 0x0100
#define XT_RTPPROXY_SOCKOPT_FLAG_THROTTLE_BANDWIDTH 0x0200
#define XT_RTPPROXY_SOCKOPT_FLAG_SESSION_TTL 0x0400

struct xt_rtpproxy_sockopt_session {
	struct xt_rtpproxy_sockopt_session_sh sh;
	struct xt_rtpproxy_sockopt_session_dir dir[2];
} PACKED;

#ifndef __KERNEL__

#include <rpc/rpc.h>

#define XT_RTPPROXY_RPC_DEFAULT_UDP_REPLY_TIMEOUT 400
#define XT_RTPPROXY_RPC_DEFAULT_TOTAL_TIMEOUT 1500
#define XT_RTPPROXY_RPC_DEFAULT_REQUEST_SIZE 1000
#define XT_RTPPROXY_RPC_DEFAULT_REPLY_SIZE 65000


enum xt_rtpproxy_connection {
	xt_rtpproxy_LOCAL = 0,
	xt_rtpproxy_REMOTE
};

struct xt_rtpproxy_connection_rpc_params {
	enum {xt_rtpproxy_connection_UDP, xt_rtpproxy_connection_TCP, xt_rtpproxy_connection_RAW} proto;
	u_int32_t addr;
	u_int16_t port;
	unsigned int udp_retry_timeout;
	unsigned int total_timeout;
	unsigned int request_size;
	unsigned int reply_size;
};

/* libipt function prototypes */
struct xt_rtpproxy_handle {
	enum xt_rtpproxy_connection type;
	int sockfd;
	union {
		struct {
			CLIENT *client;
			unsigned int total_timeout;
		} rpc;
	} u;
	int err_no; /* see XT_RTPPROXY_ERR_xxx */
	char *err_str;
	unsigned int buf_size;
	void *buf, *buf_ptr;
	int buf_iter;
};


#ifdef __cplusplus
extern "C" {
#endif

int xt_RTPPROXY_open(struct xt_rtpproxy_handle *handle, enum xt_rtpproxy_connection type, void* params);
void xt_RTPPROXY_close(struct xt_rtpproxy_handle *handle);

enum xt_rtpproxy_out_level {
	xt_rtpproxy_CONFIG = 0,
	xt_rtpproxy_SWITCHBOARDS,
	xt_rtpproxy_SESSIONS
};

int xt_RTPPROXY_get_config(
	struct xt_rtpproxy_handle *handle,
	xt_rtpproxy_sockopt_count switchboard_count,
	struct xt_rtpproxy_switchboard_id (*switchboards)[],

	enum xt_rtpproxy_out_level out_level,  /* switchboard/sessions are available via xt_RTPPROXY_read_from_buf */
	unsigned int flags,
	struct xt_rtpproxy_sockopt_out_config *config
);

int xt_RTPPROXY_read_from_buf(struct xt_rtpproxy_handle *handle, unsigned int sz, void *val);

int xt_RTPPROXY_alloc_sessions(
	struct xt_rtpproxy_handle *handle,
	xt_rtpproxy_sockopt_count switcboard_count1,  /* if >0 allocate all sessions at one of provided switchboard, otherwise alloc at switchboard provided in sessions */
	struct xt_rtpproxy_switchboard_id (*switchboards1)[],
	xt_rtpproxy_sockopt_count switcboard_count2,
	struct xt_rtpproxy_switchboard_id (*switchboards2)[],
	xt_rtpproxy_sockopt_count session_count,
	struct xt_rtpproxy_sockopt_session (*sessions)[]    /* in-out param */
);

int xt_RTPPROXY_update_sessions( /* also xt_RTPPROXY_get_sessions */
	struct xt_rtpproxy_handle *handle,
	xt_rtpproxy_sockopt_count session_count,
	struct xt_rtpproxy_sockopt_session (*sessions)[]   /* in-out param */
);

int xt_RTPPROXY_delete_switchboard_sessions(
	struct xt_rtpproxy_handle *handle,
	xt_rtpproxy_sockopt_count switchboard_count,   /* 0 to delete all switchboards */
	struct xt_rtpproxy_switchboard_id (*switchboards)[]
);

int xt_RTPPPROXY_retrieve_sessions(
	struct xt_rtpproxy_handle *handle,
	struct xt_rtpproxy_sockopt_in_interval *interval,  /* NULL .. all sessions */
	xt_rtpproxy_sockopt_count switchboard_count,   /* 0 .. all switchboards */
	struct xt_rtpproxy_switchboard_id (*switchboards)[],
	u_int32_t *buf_size,                                 /* out param */
	void **buf
);

int xt_RTPPPROXY_replicate_sessions(
	struct xt_rtpproxy_handle *handle,
	u_int32_t buf_size,
	void *buf
);

int xt_RTPPPROXY_get_replication_package_info(
	struct xt_rtpproxy_handle *handle,
	u_int32_t buf_size,
	void *buf,
	struct xt_rtpproxy_replication_package_info *info
);

int xt_RTPPROXY_set_ha_state(
	struct xt_rtpproxy_handle *handle,
	u_int8_t ha_state,
	xt_rtpproxy_sockopt_count switchboard_count,   /* 0 .. all switchboards */
	struct xt_rtpproxy_switchboard_id (*switchboards)[]
);

int xt_RTPPROXY_rpc_svc_run(
	struct xt_rtpproxy_handle *rpc_svs_handle,  /* return handle if fails */
	struct xt_rtpproxy_connection_rpc_params *params
);

#ifdef __cplusplus
}
#endif

#endif /* __KERNEL__ */

#endif /*_XT_RTPPROXY_H_target*/

