/***************************************************************************
 *   (C) 2007 by Tomas Mandys <tomas.mandys at 2p dot cz>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License           *
 *   version 2.1 as published by the Free Software Foundation;             *
 *                                                                         *
 ***************************************************************************/

/*
 * This implements the RTPPROXY target, which enables you to setup
 * RTP proxy - handle RTP+RTCP streams through proxy to traverse NATs
 *
 */

#include <linux/autoconf.h>

#if !defined(CONFIG_NETFILTER_XT_TARGET_RTPPROXY) && !defined(CONFIG_NETFILTER_XT_TARGET_RTPPROXY_MODULE)
/* if compiled from linux devel then kernel is not configured to use this module, i.e.
   use conntrack, which is neutral because does not require any kernel patches */
#  define CONFIG_NETFILTER_XT_TARGET_RTPPROXY_USE_CONNTRACK Y
#endif

#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/version.h>
#include <linux/netfilter.h>
#include <linux/netfilter/x_tables.h>
#if defined(CONFIG_NETFILTER_XT_TARGET_RTPPROXY_USE_CONNTRACK)
#  if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
#include <net/netfilter/nf_conntrack_compat.h>
#  else
#include <net/netfilter/nf_conntrack.h>
#  endif
#endif
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter/xt_RTPPROXY.h>
#include <linux/netdevice.h>
#include <net/ip.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif
#include <linux/spinlock.h>
#include <linux/kernel.h>

#define PRINTK(level, format, args...) printk(level "xt_RTPPROXY: " format , ## args)

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,24)
#  ifdef __KERNEL__
#define NF_IP_PRE_ROUTING NF_INET_PRE_ROUTING
#define NF_IP_LOCAL_IN NF_INET_LOCAL_IN
#define NF_IP_FORWARD NF_INET_FORWARD
#define NF_IP_LOCAL_OUT NF_INET_LOCAL_OUT
#define NF_IP_POST_ROUTING NF_INET_POST_ROUTING
#  endif
#endif

#if 1
/* override DEBUGP to printk, just for development purposes */
/* Note: __FILE__, __FUNCTION__ macro may be used */
#define DEBUGP(format, args...) printk(KERN_DEBUG "xt_RTPPROXY: " format, ## args)
#else
#define DEBUGP(format, args...)
#endif

#if 1
/* override DEBUGPT to printk in do_target, just for development purposes */
/* Note: __FILE__, __FUNCTION__ macro may be used */
#	define DEBUGPT(format, args...) printk(KERN_DEBUG "xt_RTPPROXY: " format, ## args)
#	define ASSERT(x) \
	do { \
		if (!(x)) { \
			printk(KERN_DEBUG "xt_RTPPROXY: ASSERT: %s:%i(%s)\n", __FILE__, __LINE__, __FUNCTION__); \
		} \
	} while(0)
#else
#	define DEBUGPT(format, args...)
#	define ASSERT(x)
#endif

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Tomas Mandys <tomas.mandys@2p.cz>");
MODULE_DESCRIPTION("x_tables RTPPROXY target module");
MODULE_ALIAS("ipt_RTPPROXY");

#define THROTTLE_INTERVAL 1  /* sec */
#define CPS_INTERVAL 1  /* sec */

/* serializes switchboard assignment 
 * write_lock: if switchboard chain or session->session2 is manipulated
 * read_lock: if session->session2 is being read
 */
static DEFINE_RWLOCK(rw_lock);

/* serializes user-context access
 * - switchboard allocations and deallocations + chain manipulation
 */
static DECLARE_MUTEX(xt_usercontext_mutex);

static struct xt_rtpproxy_switchboard *switchboard_reg = NULL;

static struct xt_rtpproxy_sockopt_out_config so_global;

static inline void set_state(struct xt_rtpproxy_session *sess, const u_int8_t state) {
//	atomic_dec(switchboard->so.stat.state + sess->so.state);
	sess->so.state = state;
//	atomic_inc(switchboard->so.stat.state + sess->so.state);
}

#if !defined(CONFIG_NETFILTER_XT_TARGET_RTPPROXY_USE_CONNTRACK)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
static u_int16_t ip_nat_cheat_check(u_int32_t oldvalinv, u_int32_t newval, u_int16_t oldcheck) {
	u_int32_t diffs[] = { oldvalinv, newval };
	return csum_fold(csum_partial((char *)diffs, sizeof(diffs), oldcheck^0xFFFF));
}
#endif
#endif

static inline struct xt_rtpproxy_session *get_sessA(struct xt_rtpproxy_session *sess) {
	if (!sess || !sess->session2)
		return NULL;
	else
		return sess->dir ? sess : sess->session2;
}

/* called as softirq, may run several do_target on multiple SMP simultaneously */
static unsigned int do_target(
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
				struct sk_buff **pskb,
#else
				struct sk_buff *skb,
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)
				const struct net_device *in,
				const struct net_device *out,
				unsigned int hooknum,
				const struct xt_target *target,
				const void *targinfo
#else
				const struct xt_target_param *params
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
				, void *userinfo
#endif
					 )
{
	const struct xt_rtpproxy_target_info *info;
	struct iphdr *piph;
	struct udphdr *pudph;
#ifdef CONFIG_NETFILTER_XT_TARGET_RTPPROXY_USE_CONNTRACK
#  if defined(CONFIG_IP_NF_CONNTRACK) || defined(CONFIG_IP_NF_CONNTRACK_MODULE)
	struct ip_conntrack *ct;
	u_int32_t ctinfo;
#  else
	struct nf_conn *ct;
	enum ip_conntrack_info ctinfo;
#  endif
#else

#endif
	u_int16_t port, *old_port, new_port;
	u_int32_t ip, *old_ip, new_ip;
	int sess_no, is_rtcp;
	struct xt_rtpproxy_session *sess, *sessA;
	time_t tstamp_off_sec;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)	
	struct sk_buff *skb = *pskb;
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)
	#define TG_PARAMS(f) f
#else
	#define TG_PARAMS(f) params->f
#endif
	info = TG_PARAMS(targinfo);

	/* may be called only for UDP packet, yes I'm sure, see target registration later */

	DEBUGPT("do_target(skb=%p, in=%p, out=%p, hooknum=%u, targinfo=%p)\n", skb, TG_PARAMS(in), TG_PARAMS(out), TG_PARAMS(hooknum), TG_PARAMS(targinfo));
	/* Now we take destination ip:port from tuple, we'll do it instead of skb->nh.iph because
	 * it's non manipulable part of tuple and it has correct value in both PRE_ROUTING/LOCAL_OUT
	 * and POST_ROUTING (where &skb->nh.iph->daddr is already rewritten and is useless
	 * for identification of RTP session). It cannot be a reply (stream in opposite direction)
	 * beacuse the replies have different tuple */
	ASSERT(TG_PARAMS(hooknum)==NF_IP_PRE_ROUTING || TG_PARAMS(hooknum)==NF_IP_LOCAL_OUT || TG_PARAMS(hooknum)==NF_IP_POST_ROUTING || TG_PARAMS(hooknum)==NF_IP_LOCAL_IN);
	/* get location to rewrite IP:port of UDP header, actually we need only first 8 bytes */
	piph = (void *)skb->data;
	pudph = (void *)(skb->data + piph->ihl*4);

#ifdef CONFIG_NETFILTER_XT_TARGET_RTPPROXY_USE_CONNTRACK
#  if defined(CONFIG_IP_NF_CONNTRACK) || defined(CONFIG_IP_NF_CONNTRACK_MODULE)
	DEBUGPT("do_target: ip_conntract_get()\n");
	ct = ip_conntrack_get(skb, &ctinfo);
	#define GET_TUPLE_IP(_dir_,_dest_) ct->tuplehash[_dir_].tuple._dest_.ip
#  else
	DEBUGPT("do_target: nf_ct_get()\n");
	ct = nf_ct_get(skb, &ctinfo);
	#define GET_TUPLE_IP(_dir_,_dest_) ct->tuplehash[_dir_].tuple._dest_.u3.ip
	if (!ct) {
		DEBUGPT("do_target: ct is null (bad checksum?), dropping\n");
		return NF_DROP;
	}
	if (ctinfo >= IP_CT_IS_REPLY) {
		DEBUGPT("do_target: ctinfo >= IP_CT_IS_REPLY, IPF_CONTINUE\n");
		return XT_CONTINUE;
	}
#  endif
	/* Connection must be valid and non reply - replies go in via opposite gate as originals, but allow CONTINUE not to block common communication not related to RTP, e.g. DNS */
	DEBUGPT("do_target: GET_TUPLE_IP\n");
	ip = GET_TUPLE_IP(IP_CT_DIR_ORIGINAL, dst);
	port = ntohs(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.u.udp.port);
#else
	/* save orig dest ip:port to our extra field into skb to avoid conntrack */
	if (TG_PARAMS(hooknum) == NF_IP_PRE_ROUTING || TG_PARAMS(hooknum) == NF_IP_LOCAL_OUT) {
		ip = piph->daddr;
		port = ntohs(pudph->dest);
		skb->orig_dst.ip4 = ip;
		skb->orig_dst.port = pudph->dest;
	}
	else {
		ip = skb->orig_dst.ip4;
		port = ntohs(skb->orig_dst.port);
	}
#endif	

	if (TG_PARAMS(hooknum) == NF_IP_PRE_ROUTING) {
		/* when all stuff is running on one box 2xRTP client+RTPPROXY (devel purposes only ?)
		 * then after POSTROUTING follows PREROUTING immediately and netfilter
		 * leaves skb->nfct untouched and RTPPROXY is mystified, source IP:port is address of opposite gate.
		 * conntrack should be invalidated and refreshed
		 */
		if (ip != piph->daddr ||
		    port != ntohs(pudph->dest)) {
			DEBUGPT("do_target: non-corresponding conntrack and dest address, local RTP routing? %d.%d.%d.%d:%u!=%d.%d.%d.%d:%u\n",
				NIPQUAD(ip), port, NIPQUAD(piph->daddr), ntohs(pudph->dest)
			);
			return NF_DROP;

		#if 0
			This dummy devel code is left because I expect the netfilter could behave
			correctly in future also for locally rerouted packets or if somebody
			would like to improve code not to repeat my unsucessfull attempts with kernel 2.6.17.8

			1) nf_conntrack_put causes kernel crash

			/* force conntrack refresh */
			DEBUGPT("do_target: refreshing conntrack\n");
			nf_conntrack_put(skb->nfct);
			nf_ct_untrack(skb);
			return XT_CONTINUE;		

			2) huge hack and ip_conntrack_in is not external symbol

			/* call early conntract callback */
			skb->nfct = NULL;
			it's nt possible to call ip_conntrack_in from module and I don't other way how to refresh conntrack
			if (ip_conntrack_in(hooknum, pskb, in, out, NULL) == NF_DROP) {
				return NF_DROP;
			}
			ct = ip_conntrack_get(skb, &ctinfo);
			DEBUGPT("do_target: ctinfo >= IP_CT_IS_REPLY, IPF_CONTINUE\n");
			if (ctinfo >= IP_CT_IS_REPLY) {
				return XT_CONTINUE;
			}
			/* get location to rewrite IP:port of UDP header, actually we need only first 8 bytes */
			pudph = (void *)(skb->data + piph->ihl*4);
			ASSERT(ct && (ctinfo < IP_CT_IS_REPLY));
		#endif

		}
	}	

#ifdef CONFIG_NETFILTER_XT_TARGET_RTPPROXY_USE_CONNTRACK
	DEBUGPT("do_target:{ct->src: %d.%d.%d.%d:%u, ct->dst: %d.%d.%d.%d:%u} reply:{ct->src: %d.%d.%d.%d:%u, ct->dst: %d.%d.%d.%d:%u}, ctinfo: %u\n",
		NIPQUAD(GET_TUPLE_IP(IP_CT_DIR_ORIGINAL, src)), ntohs(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port),
		NIPQUAD(GET_TUPLE_IP(IP_CT_DIR_ORIGINAL, dst)), ntohs(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.u.udp.port),
		NIPQUAD(GET_TUPLE_IP(IP_CT_DIR_REPLY, src)), ntohs(ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.u.udp.port),
		NIPQUAD(GET_TUPLE_IP(IP_CT_DIR_REPLY, dst)), ntohs(ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u.udp.port),
		ctinfo);
#endif
	DEBUGPT("do_target:skb->iph.src: %d.%d.%d.%d:%u, skb->iph.dst: %d.%d.%d.%d:%u\n",
		NIPQUAD(piph->saddr), ntohs(pudph->source),
		NIPQUAD(piph->daddr), ntohs(pudph->dest)
		);
	/* ip:port must be a port related to gate_a/b */
	/* I hope netfilter locking (against do target destroy) is accomplished by netfilter itself */
	if (ip != info->addr.ip || port < info->addr.port || port >= info->addr.port+2*info->max_sess) {
		DEBUGPT("do_target: session not found, XT_CONTINUE\n");
		return XT_CONTINUE;
	}
	sess_no = (port-info->addr.port) >> 1;
	is_rtcp = (port-info->addr.port) & 0x01; /* TODO: check if may be also even port number */
	sess = info->switchboard->sessions + sess_no;
	
	#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
	if (!skb->tstamp.off_sec && !skb->tstamp.off_usec) {
		__net_timestamp(skb);	
	}
	tstamp_off_sec = skb->tstamp.off_sec;
	#else
	if (!skb->tstamp.tv64) {
		__net_timestamp(skb);	
	}
	{
		struct timeval stamp;
		skb_get_timestamp(skb, &stamp);
		tstamp_off_sec = stamp.tv_sec;
	}
	#endif
	if (!tstamp_off_sec) {
		/* BUG cannot get stamp */
		DEBUGP("do_target: BUG: tstamp_off_sec is 0\n");
		return NF_DROP;
	}

	read_lock(&rw_lock);  /* protect against session end-point changes, i.e. sess[*]->session2 is solid rock */
	sessA = get_sessA(sess);
	if (!sessA) {
		read_unlock(&rw_lock);
		DEBUGPT("do_target: session not established, NF_DROP\n");
		return NF_DROP;
	}
	if (sess->switchboard->so.ha_state == xt_rtpproxy_STANDBY) {
		read_unlock(&rw_lock);
		DEBUGPT("do_target: switchboard in standby, dropping\n");
		return NF_DROP;
	}
	
	spin_lock(&sessA->lock);
	
	DEBUGPT("do_target: session found\n");
	
	switch (TG_PARAMS(hooknum)) {
		case NF_IP_PRE_ROUTING:
		case NF_IP_LOCAL_OUT:
			/* check if session is active */
			switch (sessA->so.state) {
				case xt_rtpproxy_INIT1:
				case xt_rtpproxy_INIT2:
				case xt_rtpproxy_FORWARD1:
				case xt_rtpproxy_FORWARD2:
					break;
				default:
					DEBUGPT("do_target: sessA->state=%d, NF_DROP\n", sessA->so.state);
					spin_unlock(&sessA->lock);
					read_unlock(&rw_lock);
					return NF_DROP;
			}

			/* check if session is still active */
			if (sessA->so.expires < tstamp_off_sec) {
				set_state(sessA, xt_rtpproxy_EXPIRED);
				sessA->so.last_change_stamp = tstamp_off_sec; 
				spin_unlock(&sessA->lock);
				read_unlock(&rw_lock);
				DEBUGPT("do_target: session expired, NF_DROP\n");
				return NF_DROP;
			}
			if (sessA->so.ttl && (sessA->so.created + sessA->so.ttl) < tstamp_off_sec) {
				set_state(sessA, xt_rtpproxy_STALED);
				sessA->so.last_change_stamp = tstamp_off_sec; 
				spin_unlock(&sessA->lock);
				read_unlock(&rw_lock);
				DEBUGPT("do_target: session staled, NF_DROP\n");
				return NF_DROP;
			}

			if (sess->so_dir.stream[is_rtcp].source_learnt.ip != piph->saddr || sess->so_dir.stream[is_rtcp].source_learnt.port != ntohs(pudph->source)) {
				if ( sess->so_dir.stream[is_rtcp].learning >= tstamp_off_sec &&
				    (sess->so_dir.stream[is_rtcp].source_learnt.ip == piph->saddr || (sess->so_dir.stream[is_rtcp].flags & XT_RTPPROXY_SOCKOPT_FLAG_LEARN_PORT_ONLY)== 0)) {
					/* learn source address:port */
					sess->so_dir.stream[is_rtcp].source_learnt.ip = piph->saddr;
					sess->so_dir.stream[is_rtcp].source_learnt.port = ntohs(pudph->source);
					if (!sess->so_dir.always_learn)
						sess->so_dir.stream[is_rtcp].learning = 0;  /* learn only once */
					sessA->so.last_change_stamp = tstamp_off_sec; 
					/* sess->learnt = skb->tstamp; */
				}
				else {
					/* drop unlearnable packet silently */
					spin_unlock(&sessA->lock);
					read_unlock(&rw_lock);
					DEBUGPT("do_target: unlearnable packet, NF_DROP\n");
					return NF_DROP;
				}
			}
			/* rewrite destination address to opposite side (learnt/forced IP:port) */
			if (sess->session2->so_dir.stream[is_rtcp].source_learnt.ip == 0) {
				/* not learnt/received any packet from opposite side, we cannot forward it */
				if (sess->session2->so_dir.stream[is_rtcp].learning >= tstamp_off_sec) {
					/* if we don't missed chance forever to learn opposite address then prolong expires even we cannot forward the packet */
					sessA->so.expires = tstamp_off_sec + sessA->so.expires_timeout;
				}
				spin_unlock(&sessA->lock);
				read_unlock(&rw_lock);
				DEBUGPT("do_target: no packet received from opposite side, NF_DROP\n");
				return NF_DROP;
			}
			switch (sessA->so.state) {
				case xt_rtpproxy_INIT2:
					set_state(sessA, xt_rtpproxy_FORWARD1); /* first forwarded packet */
					sessA->so.last_change_stamp = tstamp_off_sec; 
					break;
				case xt_rtpproxy_FORWARD1:
					if (sess->session2->so_dir.stream[is_rtcp].last_packet_stamp) {
						/* we've already got packet from opposite side (of the same stream!) */
						set_state(sessA, xt_rtpproxy_FORWARD2);
						sessA->so.last_change_stamp = tstamp_off_sec; 
					}
					break;
			}

			new_ip = sess->session2->so_dir.stream[is_rtcp].source_learnt.ip;
			new_port = sess->session2->so_dir.stream[is_rtcp].source_learnt.port;

			sess->so_dir.stat.packets_in++;
			sess->so_dir.stat.bytes_in += ntohs(pudph->len);
//			info->switchboard->stat.traffic.packets_in++;
//			info->switchboard->stat.traffic.bytes_in += ntohs(pudph->len);
			sessA->so.expires = tstamp_off_sec + sessA->so.expires_timeout;
			sess->so_dir.stream[is_rtcp].last_packet_stamp = tstamp_off_sec;

			/* bandwidth throttling */
			if (!sess->so_dir.stream[is_rtcp].throttle.stamp) {
				sess->so_dir.stream[is_rtcp].throttle.stamp = tstamp_off_sec; 
			} else if (sess->so_dir.stream[is_rtcp].throttle.stamp + THROTTLE_INTERVAL <= tstamp_off_sec) {
				sess->so_dir.stream[is_rtcp].throttle.bandwidth = sess->so_dir.stream[is_rtcp].throttle.counter;
				memset(&sess->so_dir.stream[is_rtcp].throttle.counter, 0, sizeof(sess->so_dir.stream[is_rtcp].throttle.counter));
				sess->so_dir.stream[is_rtcp].throttle.stamp = tstamp_off_sec; 
			}
			sess->so_dir.stream[is_rtcp].throttle.counter.packets++;
			sess->so_dir.stream[is_rtcp].throttle.counter.bytes += ntohs(pudph->len);
			if ((sess->so_dir.stream[is_rtcp].throttle.max_bandwidth.packets && sess->so_dir.stream[is_rtcp].throttle.counter.packets > sess->so_dir.stream[is_rtcp].throttle.max_bandwidth.packets) ||
			    (sess->so_dir.stream[is_rtcp].throttle.max_bandwidth.bytes && sess->so_dir.stream[is_rtcp].throttle.counter.bytes > sess->so_dir.stream[is_rtcp].throttle.max_bandwidth.bytes)) {
				/* mark packet */
			#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
		        skb->nfmark |= sess->so_dir.stream[is_rtcp].throttle.mark;
			#else
			skb->mark |= sess->so_dir.stream[is_rtcp].throttle.mark;
			#endif
			sess->so_dir.stream[is_rtcp].throttle.mark_counter.packets++;
			sess->so_dir.stream[is_rtcp].throttle.mark_counter.bytes += ntohs(pudph->len);
			}
			break;

		case NF_IP_POST_ROUTING:
		case NF_IP_LOCAL_IN:
			/* rewrite source address to proxy gate */
			new_ip = sess->session2->so_dir.switchboard.addr.ip;
			new_port = sess->session2->so_dir.stream[is_rtcp].port;

			sess->session2->so_dir.stat.packets_out++;
			sess->session2->so_dir.stat.bytes_out += ntohs(pudph->len);
//			sess->gate2->so.stat.traffic.packets_out++;
//			sess->gate2->so.stat.traffic.bytes_out += ntohs(pudph->len);
			break;

		default: /* never happens */
			spin_unlock(&sessA->lock);
			read_unlock(&rw_lock);
			DEBUGPT("do_target: unexpected hooknum = %d, XT_CONTINUE\n", TG_PARAMS(hooknum));
			return XT_CONTINUE;
	}
	spin_unlock(&sessA->lock);
	read_unlock(&rw_lock);

	if (!skb_make_writable(
	#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
		pskb, 
	#else
		skb,
	#endif
		piph->ihl*4 + sizeof(*pudph))) { /* it changes skb !!! */
		DEBUGPT("do_target: !skb_make_writable\n");
                return NF_DROP;
	}

	#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
	skb = *pskb;
	#endif
	piph = (void *)skb->data;
	pudph = (void *)(skb->data + piph->ihl*4);
	switch (TG_PARAMS(hooknum)) {
		case NF_IP_PRE_ROUTING:
		case NF_IP_LOCAL_OUT:
			old_ip = &piph->daddr;
			old_port = &pudph->dest;
			/* dst_release(skb->dst); idea taken from NAT but does not work
			skb->dst = NULL; */
			break;
		case NF_IP_POST_ROUTING:
		case NF_IP_LOCAL_IN:
			old_ip = &piph->saddr;
			old_port = &pudph->source;
			break;
		default:
			return NF_DROP;  /* never happen */
	}
	#if 0
	if (TG_PARAMS(hooknum) == NF_IP_LOCAL_OUT) {
		/* bug: seems checksum is incorrect for locally generated udp packets, e.g. data not counted, seems also ^0xFFFFed (kernel 2.6.17.8)  */
		pudph->check = 0;
	}
	#endif

	/* rewrite UDP checksum if provided */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
	if (pudph->check) {
		pudph->check = ip_nat_cheat_check(~*old_ip, new_ip, 
				ip_nat_cheat_check(*old_port ^ 0xFFFF, new_port,
				pudph->check));
	}
	/* rewrite also IP header checksum */
	piph->check = ip_nat_cheat_check(~*old_ip, new_ip, piph->check);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
        if (pudph->check || skb->ip_summed == CHECKSUM_PARTIAL) {
	
		pudph->check = nf_proto_csum_update(skb, ~*old_ip, new_ip, pudph->check, 1);
		pudph->check = nf_proto_csum_update(skb, *old_port ^ 0xFFFF, new_port, pudph->check, 0);
	}
	/* rewrite also IP header checksum */
	piph->check = nf_csum_update(~*old_ip, new_ip, piph->check);
#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,6,24)
	if (pudph->check || skb->ip_summed == CHECKSUM_PARTIAL) {
		inet_proto_csum_replace4(&pudph->check, skb, *old_ip, new_ip, 1);
		inet_proto_csum_replace2(&pudph->check, skb, *old_port, new_port, 0);
		if (!pudph->check)
			pudph->check = CSUM_MANGLED_0;
	}
	/* rewrite also IP header checksum */
	csum_replace4(&piph->check, *old_ip, new_ip);
#else
	if (pudph->check || skb->ip_summed == CHECKSUM_PARTIAL) {
		nf_proto_csum_replace4(&pudph->check, skb, *old_ip, new_ip, 1);
		nf_proto_csum_replace2(&pudph->check, skb, *old_port, new_port, 0);
		if (!pudph->check)
			pudph->check = CSUM_MANGLED_0;
	}
	/* rewrite also IP header checksum */
	nf_csum_replace4(&piph->check, *old_ip, new_ip);
#endif

	DEBUGPT("do_target: old_addr: %d.%d.%d.%d:%u, new_addr: %d.%d.%d.%d:%u, udph->check: %u, ip_summed: %d\n", NIPQUAD(*old_ip), ntohs(*old_port), NIPQUAD(new_ip), new_port, pudph->check, skb->ip_summed);
	pudph->check = 0; /* avoid UDP checksum, I'm not able to calculate it correctly */

	*old_port = htons(new_port);
	*old_ip = new_ip;

	/* skb->nfcache |= NFC_ALTERED; */
	DEBUGPT("do_target: NF_ACCEPT\n");
	return NF_ACCEPT;
}

/* recheck session status and calc statistics by state, xt_usercontext_mutex must be acquired and rw_lock must not be acquired,
   return TRUE (sessA) if session is established (and session locked if leave_locked is TRUE), NULL if session teared down */
static inline struct xt_rtpproxy_session * xt_do_session_audit(struct timeval *stamp, struct xt_rtpproxy_session *sess, int leave_locked)
{
	struct xt_rtpproxy_session *sessA;
	sessA = get_sessA(sess);
	if (!sessA)
		return NULL;
	spin_lock_bh(&sessA->lock);
	if (sess->switchboard->so.ha_state != xt_rtpproxy_STANDBY && sess->session2->switchboard->so.ha_state != xt_rtpproxy_STANDBY) {
		switch (sessA->so.state) {
			case xt_rtpproxy_INIT1:
			case xt_rtpproxy_INIT2:
			case xt_rtpproxy_FORWARD1:
			case xt_rtpproxy_FORWARD2:
				if ((__kernel_time_t) sessA->so.expires < stamp->tv_sec) {
					DEBUGP("xt_do_session_audit: expiration sess-id %d/%d, expires: %ld<%ld\n", sess->so_dir.sess_id, sess->session2->so_dir.sess_id, (__kernel_time_t) sessA->so.expires, stamp->tv_sec);
					set_state(sessA, xt_rtpproxy_EXPIRED);
					sessA->so.last_change_stamp = stamp->tv_sec; 
					sess->switchboard->so.stat.counter.expired++;
					if (sess->session2 && sess->session2->switchboard != sess->switchboard) {
							sess->session2->switchboard->so.stat.counter.expired++;
					}	
				}
				if (sessA->so.ttl && (__kernel_time_t) (sessA->so.created + sessA->so.ttl) < stamp->tv_sec) {
					DEBUGP("xt_do_session_audit: staling sess-id %d/%d, stales: %ld<%ld\n", sess->so_dir.sess_id, sess->session2->so_dir.sess_id, (__kernel_time_t) (sessA->so.created + sessA->so.ttl), stamp->tv_sec);
					set_state(sessA, xt_rtpproxy_STALED);
					sessA->so.expires = stamp->tv_sec; /* for correct resurrection */
					sessA->so.last_change_stamp = stamp->tv_sec; 
					sess->switchboard->so.stat.counter.staled++;
					if (sess->session2 && sess->session2->switchboard != sess->switchboard) {
							sess->session2->switchboard->so.stat.counter.staled++;
					}	
				}
				break;
			default:
				break;
		}
		switch (sessA->so.state) {
			case xt_rtpproxy_STALED:
			case xt_rtpproxy_EXPIRED:
			case xt_rtpproxy_DESTROYED:
				if ((__kernel_time_t)(sessA->so.expires + sess->switchboard->so.resurrection_timeout) < stamp->tv_sec) {
					DEBUGP("xt_do_session_audit: destroying sess-id %d/%d, resurrection: %ld<%ld\n", sess->so_dir.sess_id, sess->session2->so_dir.sess_id, (__kernel_time_t) (sessA->so.expires + sess->switchboard->so.resurrection_timeout), stamp->tv_sec);
					spin_unlock_bh(&sessA->lock);
					/* minor non-important race condition, tear down even expires has been prolonged meanwhile */
					write_lock_bh(&rw_lock);
					set_state(sessA, xt_rtpproxy_NONE);
					sess->so.last_change_stamp = stamp->tv_sec; 
					if (sess->session2) {
						sess->switchboard->so.stat.num_session--;
						sess->session2->switchboard->so.stat.num_session--;
						sess->session2->so.last_change_stamp = stamp->tv_sec; 
						sess->session2->session2 = NULL;
						sess->session2 = NULL;
						so_global.total.num_pair_session--;
					}
					write_unlock_bh(&rw_lock);
					return NULL;
				}
			default: ;
		}
	}
	if (!leave_locked) {
		spin_unlock_bh(&sessA->lock);
	}
	return sessA;
}

static inline int copy_from_user2(void** user2, int *in_len, unsigned int buf_size, void* buf) {
	DEBUGP("copy_from_user2(%p, %d, %u)\n", *user2, *in_len, buf_size);
	if (*in_len < buf_size || copy_from_user (buf, *user2, buf_size)) { 
		return 1;
	}
	else {
		*user2 += buf_size; 
		*in_len -= buf_size; 
		return 0;
	}
}

static inline int copy_to_user2(void** user2, unsigned int buf_size, void* buf) {
	DEBUGP("copy_to_user2(%p, %u)\n", *user2, buf_size);
	if (copy_to_user (*user2, buf, buf_size)) { 
		return 1;
	}
	else {
		*user2 += buf_size; 
		return 0;
	}
}

static inline int copy_to_user_len2(void** user2, unsigned int *out_len, unsigned int buf_size, void* buf) {
	DEBUGP("copy_to_user_len2(%p, %u, %u)\n", *user2, *out_len, buf_size);
	if (*out_len < buf_size) {
		*out_len = 0;
		*user2 += buf_size;
		return 0;
	}
	else {
		if (copy_to_user (*user2, buf, buf_size)) { 
			return 1;
		}
		else {
			*out_len -= buf_size;
			*user2 += buf_size; 
			return 0;
		}
	}
}


#define DEBUGP_ERR DEBUGP("error in function '%s' (%s:%d)\n",__FUNCTION__,  __FILE__, __LINE__);					

#define COPY_FROM_USER(user2, in_len, buf) copy_from_user2(&(user2), &(in_len), sizeof(buf), &(buf))
#define COPY_TO_USER(user2, buf) copy_to_user2(&(user2), sizeof(buf), &(buf)) 
#define COPY_TO_USER_LEN(user2, out_len, buf) copy_to_user_len2(&(user2), &(out_len), sizeof(buf), &(buf)) 

/* userspace related communication, xt_usercontext_mutex must be acquired */
static inline struct xt_rtpproxy_switchboard* xt_find_switchboard(struct xt_rtpproxy_switchboard_id *switchboard) {
	struct xt_rtpproxy_switchboard *p;
	DEBUGP("xt_find_switchboard: %d.%d.%d.%d:%u\n", NIPQUAD(switchboard->ip), switchboard->port);
	for (p = switchboard_reg; p; p = p->next) {
		/* chain is sorted by IP, port */
		if (p->so.addr.ip > switchboard->ip) 
			break;
		if (p->so.addr.ip == switchboard->ip) {
			if (p->so.addr.port == switchboard->port) {
				DEBUGP("xt_find_switchboard: found\n");
				return p;
			}
			else if (p->so.addr.port > switchboard->port) {
				break;
			}
		}
	}
	return NULL;
}

static inline struct xt_rtpproxy_session* xt_find_session(struct xt_rtpproxy_sockopt_session *in_sess) {
	struct xt_rtpproxy_session *sess = NULL, *sessA;
	if (!in_sess->dir[0].switchboard.addr.ip && !in_sess->dir[1].switchboard.addr.ip) {
		return NULL;
	}
	if (in_sess->dir[0].switchboard.addr.ip) {
		struct xt_rtpproxy_switchboard *switchboard;
		switchboard = xt_find_switchboard(&in_sess->dir[0].switchboard.addr);
		if (!switchboard) {
			return NULL;
		}
		
		if (in_sess->dir[0].sess_id >= switchboard->so.max_sess) {
			return NULL;
		}
		sess = switchboard->sessions + in_sess->dir[0].sess_id;
		if (!sess->session2) {
			return NULL;
		}
	}
	if (in_sess->dir[1].switchboard.addr.ip) {
		struct xt_rtpproxy_switchboard *switchboard;
		switchboard = xt_find_switchboard(&in_sess->dir[1].switchboard.addr);
		if (!switchboard) {
			return NULL;
		}
		if (in_sess->dir[1].sess_id >= switchboard->so.max_sess) {
			return NULL;
		}
		if (sess && (sess->session2 != switchboard->sessions + in_sess->dir[0].sess_id)) {
			/* redundant identification, but check if matches */
			return NULL;
		}
		sess = switchboard->sessions + in_sess->dir[1].sess_id;
		sess = sess->session2;
	}
	sessA = get_sessA(sess);
	if (sessA && in_sess->sh.created && sessA->so.created != in_sess->sh.created)
		return NULL; /* created does not match */
	return sess;
}

static int xt_do_get_switchboard_config(struct timeval *stamp, void **out_user2, struct xt_rtpproxy_switchboard *switchboard, u_int32_t flags, int reset_only, xt_rtpproxy_sockopt_count *out_session_count) {
	void *out_switchboard_user;
	xt_rtpproxy_sockopt_count out_session_count2;
	int i;

	out_switchboard_user = *out_user2;
	out_session_count2 = 0;
	memset(&switchboard->so.stat.state, 0, sizeof(switchboard->so.stat.state));
	memset(&switchboard->so.stat.traffic, 0, sizeof(switchboard->so.stat.traffic));
	memset(&switchboard->so.stat.throttle, 0, sizeof(switchboard->so.stat.throttle));
	if (!reset_only) {
		if (COPY_TO_USER(*out_user2, switchboard->so)) {
			DEBUGP_ERR;
			return -2;
		}
		if (COPY_TO_USER(*out_user2, out_session_count2)) {
			DEBUGP_ERR;
			return -2;
		}
	}

	for (i=0; i < switchboard->so.max_sess; i++) {
		struct xt_rtpproxy_session *sess, *sessA;
		sess = switchboard->sessions + i;
		
		sessA = xt_do_session_audit(stamp, sess, 1);
		/* update statistics */
		if (sessA) {
			switchboard->so.stat.state[sessA->so.state]++;
		}
		else {
			switchboard->so.stat.state[xt_rtpproxy_NONE]++;
		}
		switchboard->so.stat.traffic.packets_in += sess->so_dir.stat.packets_in;
		switchboard->so.stat.traffic.packets_out += sess->so_dir.stat.packets_out;
		switchboard->so.stat.traffic.bytes_in += sess->so_dir.stat.bytes_in;
		switchboard->so.stat.traffic.bytes_out += sess->so_dir.stat.bytes_out;
		if (sess->so_dir.stream[0].throttle.mark_counter.packets || sess->so_dir.stream[1].throttle.mark_counter.packets) {
			int j;
			switchboard->so.stat.throttle.num_mark_session++;
			for (j=0; j<2; j++) {
				switchboard->so.stat.throttle.mark_counter.packets += sess->so_dir.stream[j].throttle.mark_counter.packets;
				switchboard->so.stat.throttle.mark_counter.bytes += sess->so_dir.stream[j].throttle.mark_counter.bytes;
			}
		}
		if (flags & XT_RTPPROXY_SOCKOPT_FLAG_RESET_SWITCHBOARD_PACKET_STAT) {
			int j;
			memset(&sess->so_dir.stat, 0, sizeof(sess->so_dir.stat));
			for (j=0; j<2; j++) {
				memset(&sess->so_dir.stream[j].throttle.mark_counter, 0, sizeof(sess->so_dir.stream[j].throttle.mark_counter));
			}
		}
		if (!reset_only && (flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS) && sessA && *out_session_count) {
			struct xt_rtpproxy_sockopt_session in_sess;
			in_sess.sh = sessA->so;
			in_sess.dir[0] = sess->so_dir;
			in_sess.dir[1] = sess->session2->so_dir;
			if (sessA) {
				spin_unlock_bh(&sessA->lock);
			}
			if (COPY_TO_USER(*out_user2, in_sess)) {
				DEBUGP_ERR;
				return -2;
			}

			out_session_count2++;
			(*out_session_count)--;			
		}
		else {
			if (sessA) {
				spin_unlock_bh(&sessA->lock);
			}
		}
	} 

	if (!reset_only) {
		if (COPY_TO_USER(out_switchboard_user, switchboard->so)) {
			DEBUGP_ERR;
			return -2;
		}
		if (flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS) {
			if (COPY_TO_USER(out_switchboard_user, out_session_count2)) {
				DEBUGP_ERR;
				return -2;
			}
		}
	}
	if (flags & XT_RTPPROXY_SOCKOPT_FLAG_RESET_SWITCHBOARD_STAT) {
		memset(&switchboard->so.stat.counter, 0, sizeof(switchboard->so.stat.counter));
	}
	return 0;
}
		 
typedef struct xt_rtpproxy_session	*xt_rtpproxy_session_ptr_arr[XT_RTPPROXY_MAX_ALLOC_SESSION];

static int xt_do_allocate_sessions2(
			struct timeval *stamp, struct xt_rtpproxy_switchboard *switchboard, 
			int session_count, int offset, xt_rtpproxy_session_ptr_arr *sessions) {
	int sess_alloc;
	int next_sess_id, n;
	sess_alloc = 0;
	next_sess_id = switchboard->next_sess_id;
	for (n=switchboard->so.max_sess; n > 0; n--) {
		int established;
		(*sessions)[sess_alloc+offset] = switchboard->sessions + next_sess_id;
		established = xt_do_session_audit(stamp, (*sessions)[sess_alloc+offset], 0) != NULL;
		/* no lock needed, we need non-established session, i.e. state is NONE */
		if (next_sess_id < switchboard->so.max_sess-1)
			next_sess_id++;
		else
			next_sess_id = 0;
		if (!established) {
			sess_alloc++;
			if (sess_alloc == session_count) {
				/* all session has been allocated */
				switchboard->next_sess_id = next_sess_id;
				return 0;
			}
		}
	}
	return -1;
}

/* try allocate sessions at one of switchboard taken from userspace memory, no lock may be acquired! */
static int xt_do_allocate_sessions(
			struct timeval *stamp, void **user2, int *in_len, int session_count, 
			xt_rtpproxy_session_ptr_arr *sessions) {
	struct xt_rtpproxy_switchboard_id in_switchboard;
	struct xt_rtpproxy_switchboard *last_switchboard, *switchboard;
	xt_rtpproxy_sockopt_count switchboard_count;
	int ret = -1;

	last_switchboard = NULL;
	if (COPY_FROM_USER(*user2, *in_len, switchboard_count)) {
		DEBUGP_ERR;
		goto brk;
	}	
	DEBUGP("xt_do_allocate_sessions: switchboard_count=%u\n", switchboard_count);
	
	while (switchboard_count > 0) {
		switchboard_count--;
		if (COPY_FROM_USER(*user2, *in_len, in_switchboard)) {
			DEBUGP_ERR;
			goto brk;
		}
		switchboard = xt_find_switchboard(&in_switchboard);
		if (switchboard && switchboard->so.ha_state != xt_rtpproxy_STANDBY) {
			last_switchboard = switchboard;
			if (xt_do_allocate_sessions2(stamp, switchboard, session_count, 0, sessions) == 0) {
				*user2 += switchboard_count*sizeof(in_switchboard);
				*in_len -= switchboard_count*sizeof(in_switchboard);
				return 0;
			}
		}
	}
end:
	if (last_switchboard) {
		last_switchboard->so.stat.counter.alloc_failed++;
	}
	return ret;
brk:
	ret = -2;
	goto end;
}

static unsigned int repl_retrieve_counter = 0;  /* mark session when retrieving repl data to avoid doubled sessions in package */

/* writelock rw_lock must be acquired */
static void xt_do_delete_sessions(struct timeval *stamp, struct xt_rtpproxy_switchboard *switchboard, int check_retrieve_counter) {
	int i;
	/* remove associated gate sessions */
	for (i=0; i < switchboard->so.max_sess; i++) {
		struct xt_rtpproxy_session *sess, *sessA;
		sess = switchboard->sessions + i;
		sessA = get_sessA(sess);
		if (sessA) {
			if (check_retrieve_counter && (sess->retrieve_counter == repl_retrieve_counter || sess->session2->retrieve_counter == repl_retrieve_counter)) {
				continue;
			}
			/* session spin_lock is not needed, write_lock(rw_lock) is stronger */
			set_state(sessA, xt_rtpproxy_NONE);
			sess->so.last_change_stamp = stamp->tv_sec; 
			sess->switchboard->so.stat.num_session--;
			sess->session2->switchboard->so.stat.num_session--;
			sess->session2->so.last_change_stamp = stamp->tv_sec; 
			sess->session2->session2 = NULL;
			sess->session2 = NULL;
			so_global.total.num_pair_session--;
		}			
	}	
}

/* writelock acquired */
static void xt_adjust_ha_state(struct timeval *stamp, struct xt_rtpproxy_switchboard *switchboard) {
	int i;
	switch (switchboard->so.ha_state) {
		case xt_rtpproxy_ACTIVE:
			switchboard->cps.counter = 0;
			/* in case transition STANDBY->ACTIVE we must adjust timeouts because they are not replicated */
			for (i=0; i < switchboard->so.max_sess; i++) {
				struct xt_rtpproxy_session *sess, *sessA, *ss[2];
				int m;
				sess = switchboard->sessions + i;
				sessA = get_sessA(sess);
				if (sessA) {
					switch (sessA->so.state) {
						case xt_rtpproxy_STALED:
						case xt_rtpproxy_EXPIRED:
						case xt_rtpproxy_DESTROYED:
						case xt_rtpproxy_NONE:
							break;
						default:
							sessA->so.expires = stamp->tv_sec + sessA->so.expires_timeout;
							ss[0] = sess;
							ss[1] = sess->session2;
							for (m=0; m<2; m++) {
								int j;
								for (j=0; j<2; j++) {
									ss[m]->so_dir.stream[j].throttle.stamp = 0;
									memset(&ss[m]->so_dir.stream[j].throttle.mark_counter, 0, sizeof(ss[m]->so_dir.stream[j].throttle.mark_counter));
									if (ss[m]->so_dir.stream[j].learning) {
										ss[m]->so_dir.stream[j].learning = stamp->tv_sec + ss[m]->so_dir.stream[j].learning_timeout;
										if (ss[m]->so_dir.stream[j].learning > sessA->so.expires) {
											sessA->so.expires = ss[m]->so_dir.stream[j].learning;
										}
									}
								}
							}
					}
				}
			}
			break;
		default:
			;
	}
}

#define REPL_REC_FLAG_SWITCHBOARD                 0x00000001   /* 2 bits (a, b) */
#define REPL_REC_FLAG_ALWAYS_LEARN                0x00000004   /* 2 bits (a, b) */
#define REPL_REC_FLAG_THROTTLE                    0x00000010   /* 4 bits (rtp-a, rtcp-a, rtp-b, rtcp-b */
#define REPL_REC_FLAG_LEARNING                    0x00000100   /* 4 bits */
#define REPL_REC_FLAG_LAST_PACKET_STAMP           0x00001000   /* 4 bits */
#define REPL_REC_FLAG_LEARNING_TIMEOUT            0x00010000   /* 4 bits */
#define REPL_REC_FLAG_SOURCE_ADDR                 0x00100000   /* 4 bits */
#define REPL_REC_FLAG_DIR_B                       0x01000000
#define REPL_REC_FLAG_EMPTY_SWITCHBOARD           0x02000000
#define REPL_REC_FLAG_DIR_A_B                     0x04000000

/* packed data being replicated active->standby */
/*
    packed struct xt_rtpproxy_sockopt_header.magic;
	
	if (header.magic != 0) {
		packed struct xt_rtpproxy_sockopt_header;  // !magic
		if (header.flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS == 0)
			packed struct xt_rtpproxy_sockopt_in_interval in_interval
	
		struct {
			u_int32_t flags;
			if (flags & REPL_FLAG_SWITCHBOARD_A) {
				struct xt_rtpproxy_switchboard_id switchboard_addr_a;
			}
			else {
				switchboard_addr_a = switchboard_addr_a[record-1];
			}
			if (flags & REPL_FLAG_SWITCHBOARD_B) {
				struct xt_rtpproxy_switchboard_id switchboard_addr_b;
			}
			else {
				switchboard_addr_b = switchboard_addr_a;
			}			
			if (flags & REPL_FLAG_SWITCHBOARD_A) {
				u_int16_t switchboard[0].next_sess_id;
			}

			u_int16_t sess_id[0];
			if (flags & REPL_REC_FLAG_DIR_B)
				u_int16_t sess_id[1];
			u_int8_t state;

			switch (state) {
				case xt_rtpproxy_NONE:
					break;

				case xt_rtpproxy_STALED:
				case xt_rtpproxy_EXPIRED:
				case xt_rtpproxy_DESTROYED:
					u_int32_t expires;
					break;

				default:
					if (dir == 0) {
						u_int32_t created;
						u_int16_t expires_timeout;
						u_int16_t ttl;
						u_int32_t last_change_stamp; // ????
					}

					struct {
						struct {
							if (flags & REPL_REC_FLAG_SOURCE_ADDR_x_yyy) 
								struct xt_rtpproxy_addr source;
								struct xt_rtpproxy_addr source_learnt;
							if (flags & REPL_REC_FLAG_LEARNING_TIMEOUT_x_yyy) 
								u_int16_t learning_timeout;
							if (flags & REPL_REC_FLAG_THROTTLE_x_yyy) {
								struct {	
									struct xt_rtpproxy_throttle_stat max_bandwidth;
									u_int32_t mark;
								} throttle;
							}
						} stream[2];
					} [2];
			}
		} records[record_count];
	}
	else {
		u_int32_t req_mem;
	}
*/

static int xt_do_repl_retrieve_switchboard(struct timeval *stamp, void **out_user, unsigned int *out_len, 
			struct xt_rtpproxy_sockopt_header *in_header,
			struct xt_rtpproxy_sockopt_in_interval *in_interval, 
			struct xt_rtpproxy_switchboard *switchboard, u_int32_t *out_record_count) {
	int i, all;
	struct xt_rtpproxy_switchboard_id last_switchboard_id;
	unsigned int cnt;	
	u_int32_t flags;
	struct xt_rtpproxy_session *sessA;
	if (switchboard->so.ha_state == xt_rtpproxy_STANDBY) {
		return -1;
	}
	cnt = 0;
	all = (in_header->flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS) != 0;
	memset(&last_switchboard_id, 0, sizeof(last_switchboard_id));
	for (i=0; i<switchboard->so.max_sess; i++) {
		struct xt_rtpproxy_session *sess, *sess2, *ss[2];
		int m;
		sess = switchboard->sessions + i;
		if (sess->retrieve_counter == repl_retrieve_counter) {
			/* sess already retieved */
			continue;
		}

		flags = 0;
		sessA = xt_do_session_audit(stamp, sess, 1);
		sess2 = sessA?sessA:sess;
		if (!all) {
			if (sess2->so.last_change_stamp < in_interval->from_stamp ||
			    sess2->so.last_change_stamp > in_interval->to_stamp) {
				goto unlock_continue;
			}
		}
		else {
			if (sess2->so.state == xt_rtpproxy_NONE)
				goto unlock_continue;
		}
		if (!sessA) {
			sess->retrieve_counter = repl_retrieve_counter;
		}
		else {
			flags |= REPL_REC_FLAG_DIR_B;
			if (sess->so_dir.switchboard.addr.ip != sess->session2->so_dir.switchboard.addr.ip ||
			    sess->so_dir.switchboard.addr.port != sess->session2->so_dir.switchboard.addr.port) {
				flags |= (REPL_REC_FLAG_SWITCHBOARD << 1);
			}			
			if (sess->session2->switchboard->so.ha_state == xt_rtpproxy_STANDBY) {
				spin_unlock_bh(&sessA->lock);
				return -1;
			}
			sess->retrieve_counter = sess->session2->retrieve_counter = repl_retrieve_counter;
			if (sess->dir) {
				flags |= REPL_REC_FLAG_DIR_A_B;
			}
		}
		if (cnt == 0) {
			flags |= REPL_REC_FLAG_SWITCHBOARD;
		}

		ss[0] = sess;
		ss[1] = sess->session2;
		for (m=0; m<2; m++) {
			int j;
			for (j=0; j<2; j++) {
				if (ss[m]->so_dir.stream[j].source.ip || ss[m]->so_dir.stream[j].source.port ||
					ss[m]->so_dir.stream[j].source_learnt.ip || ss[m]->so_dir.stream[j].source_learnt.port) {
					flags |= REPL_REC_FLAG_SOURCE_ADDR << (2*m+j);
				}
				if (ss[m]->so_dir.stream[j].learning) {
					flags |= REPL_REC_FLAG_LEARNING << (2*m+j);
				}
				if (ss[m]->so_dir.stream[j].learning_timeout) {
					flags |= REPL_REC_FLAG_LEARNING_TIMEOUT << (2*m+j);
				}
				if (ss[m]->so_dir.stream[j].last_packet_stamp) {
					flags |= REPL_REC_FLAG_LAST_PACKET_STAMP << (2*m+j);
				}
				if (ss[m]->so_dir.stream[j].throttle.mark || ss[m]->so_dir.stream[j].throttle.max_bandwidth.packets || ss[m]->so_dir.stream[j].throttle.max_bandwidth.bytes) {
					flags |= REPL_REC_FLAG_THROTTLE << (2*m+j);
				}
			}	
			if (ss[m]->so_dir.always_learn) {
				flags |= REPL_REC_FLAG_ALWAYS_LEARN << m;
			}
			if ((flags & REPL_REC_FLAG_DIR_B) == 0) 
				break;
		}
	
		if (COPY_TO_USER_LEN(*out_user, *out_len, flags)) {
			DEBUGP_ERR;
			goto unlock_brk;
		}
		for (m=0; m<2; m++) {
			if (flags & (REPL_REC_FLAG_SWITCHBOARD << m)) {
				if (COPY_TO_USER_LEN(*out_user, *out_len, ss[m]->so_dir.switchboard.addr)) {
					DEBUGP_ERR;
					goto unlock_brk;
				}
			}
		}
		if (flags & REPL_REC_FLAG_SWITCHBOARD) {
			if (COPY_TO_USER_LEN(*out_user, *out_len, ss[0]->switchboard->next_sess_id)) {
				DEBUGP_ERR;
				goto unlock_brk;
			}
			if (ss[0]->switchboard->cps.stamp && ss[0]->switchboard->cps.stamp + CPS_INTERVAL <= stamp->tv_sec) {
				ss[0]->switchboard->cps.stamp = stamp->tv_sec;
				ss[0]->switchboard->cps.rate = ss[0]->switchboard->cps.counter;
				ss[0]->switchboard->cps.counter = 0;
			}
			if (COPY_TO_USER_LEN(*out_user, *out_len, ss[0]->switchboard->cps.rate)) {
				DEBUGP_ERR;
				goto unlock_brk;
			}
		}
		if (COPY_TO_USER_LEN(*out_user, *out_len, ss[0]->so_dir.sess_id)) {
			DEBUGP_ERR;
			goto unlock_brk;
		}
		if (flags & REPL_REC_FLAG_DIR_B) {
			if (COPY_TO_USER_LEN(*out_user, *out_len, ss[1]->so_dir.sess_id)) {
				DEBUGP_ERR;
				goto unlock_brk;
			}
		}
		if (COPY_TO_USER_LEN(*out_user, *out_len, sess2->so.state)) {
			DEBUGP_ERR;
			goto unlock_brk;
		}
		switch (sess2->so.state) {
			case xt_rtpproxy_NONE:
				break;
			case xt_rtpproxy_STALED:
			case xt_rtpproxy_EXPIRED:
			case xt_rtpproxy_DESTROYED:
				if (COPY_TO_USER_LEN(*out_user, *out_len, sess2->so.expires)) {
					DEBUGP_ERR;
					goto unlock_brk;
				}
				break;
			default:
				if (COPY_TO_USER_LEN(*out_user, *out_len, sessA->so.created) ||
				    COPY_TO_USER_LEN(*out_user, *out_len, sessA->so.expires_timeout) ||
				    COPY_TO_USER_LEN(*out_user, *out_len, sessA->so.ttl) ) {
					DEBUGP_ERR;
					goto unlock_brk;
				}

				for (m=0; m<2; m++) {
					int j;
					for (j=0; j<2; j++) {
						if (flags & (REPL_REC_FLAG_SOURCE_ADDR << (2*m+j))) {
							if (COPY_TO_USER_LEN(*out_user, *out_len, ss[m]->so_dir.stream[j].source)) { 
								DEBUGP_ERR;
								goto unlock_brk;
							}
							if (COPY_TO_USER_LEN(*out_user, *out_len, ss[m]->so_dir.stream[j].source_learnt)) { 
								DEBUGP_ERR;
								goto unlock_brk;
							}
						}
						if (flags & (REPL_REC_FLAG_LEARNING_TIMEOUT << (2*m+j))) {
							if (COPY_TO_USER_LEN(*out_user, *out_len, ss[m]->so_dir.stream[j].learning_timeout) ) {
								DEBUGP_ERR;
								goto unlock_brk;
							}
						}
						if (flags & (REPL_REC_FLAG_THROTTLE << (2*m+j))) {
							if (COPY_TO_USER_LEN(*out_user, *out_len, ss[m]->so_dir.stream[j].throttle.max_bandwidth) ||
							    COPY_TO_USER_LEN(*out_user, *out_len, ss[m]->so_dir.stream[j].throttle.mark) ) {
								DEBUGP_ERR;
								goto unlock_brk;
							}
						}
					}
					if ((flags & REPL_REC_FLAG_DIR_B) == 0) {
						break; /*for m*/
					}
				}
				break;
		}

		(*out_record_count)++;
		cnt++;
	unlock_continue:
		if (sessA)
			spin_unlock_bh(&sessA->lock);
	}
	if (cnt == 0) {
		/* no active session, report it! report also no change to provide next_sess_id */
		(*out_record_count)++;
		flags = REPL_REC_FLAG_SWITCHBOARD | REPL_REC_FLAG_EMPTY_SWITCHBOARD;
		if (COPY_TO_USER_LEN(*out_user, *out_len, flags) ||
		    COPY_TO_USER_LEN(*out_user, *out_len, switchboard->so.addr) ||
		    COPY_TO_USER_LEN(*out_user, *out_len, switchboard->next_sess_id)) {
			DEBUGP_ERR;
			return -1;
		}
		if (switchboard->cps.stamp && switchboard->cps.stamp + CPS_INTERVAL <= stamp->tv_sec) {
			switchboard->cps.stamp = stamp->tv_sec;
			switchboard->cps.rate = switchboard->cps.counter;
			switchboard->cps.counter = 0;
		}
		if (COPY_TO_USER_LEN(*out_user, *out_len, switchboard->cps.rate)) {
			DEBUGP_ERR;
			return -1;
		}
	}
	return 0;
unlock_brk:
	if (sessA)
		spin_unlock_bh(&sessA->lock);
	return -1;

}

static int xt_do_replicate_sessions(struct timeval *stamp, void *user2, unsigned int in_len, 
					struct xt_rtpproxy_sockopt_header *in_header,
					struct xt_rtpproxy_sockopt_in_interval *in_interval, 
					u_int32_t record_count) {
	int all;
	struct xt_rtpproxy_switchboard *switchboards[2];

	memset(&switchboards, 0, sizeof(switchboards));
	all = (in_header->flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS) != 0;

	while (record_count) {
		u_int8_t state;
		u_int16_t sess_ids[2];
		struct xt_rtpproxy_session *ss[2], *sessA;
		u_int32_t flags;
		int m;

		if (COPY_FROM_USER(user2, in_len, flags)) {
			DEBUGP_ERR;
			return -1;
		}

		if (all && switchboards[0] && (flags & (REPL_REC_FLAG_SWITCHBOARD))) {
			xt_do_delete_sessions(stamp, switchboards[0], 1);
		}

		for (m=0; m<2; m++) {
			if (flags & (REPL_REC_FLAG_SWITCHBOARD << m)) {
				struct xt_rtpproxy_switchboard_id switchboard_id;
				if (COPY_FROM_USER(user2, in_len, switchboard_id)) { 
					DEBUGP_ERR;
					return -1;
				}
				switchboards[m] = xt_find_switchboard(&switchboard_id);
			}
		}
		if (flags & REPL_REC_FLAG_SWITCHBOARD) {
			unsigned int n;
			if (COPY_FROM_USER(user2, in_len, switchboards[0]->next_sess_id)) {
				DEBUGP_ERR;
				return -1;
			}
			if (COPY_FROM_USER(user2, in_len, switchboards[0]->cps.rate)) {
				DEBUGP_ERR;
				return -1;
			}
			/* kind of magic how to minimize danger that calls of non-replicated sessions from former ACTIVE
			   will affect newly created sessions at new ACTIVE */
			n = switchboards[0]->cps.rate * ((switchboards[0]->cps.stamp?(stamp->tv_sec - switchboards[0]->cps.stamp):0) + 1);
			if (n > switchboards[0]->so.max_sess / 2) {
				n = switchboards[0]->so.max_sess / 2;
			}
			if (n + switchboards[0]->next_sess_id < switchboards[0]->so.max_sess) {
				switchboards[0]->next_sess_id += n; 
			}
			else {
				switchboards[0]->next_sess_id = n + switchboards[0]->next_sess_id - switchboards[0]->so.max_sess;
			}
		}
		
		if ((flags & (REPL_REC_FLAG_SWITCHBOARD << 1)) == 0) {
			switchboards[1] = switchboards[0];
		}
		if (flags & REPL_REC_FLAG_EMPTY_SWITCHBOARD) {
			continue;
		}

		if (COPY_FROM_USER(user2, in_len, sess_ids[0])) {
			DEBUGP_ERR;
			return -1;
		}
		if (flags & REPL_REC_FLAG_DIR_B) {
			if (COPY_FROM_USER(user2, in_len, sess_ids[1])) {
				DEBUGP_ERR;
				return -1;
			}
		}
		if (COPY_FROM_USER(user2, in_len, state)) {
			DEBUGP_ERR;
			return -1;
		}
		ss[1] = NULL;
		for (m=0; m<2; m++) {
			if (!switchboards[m] || switchboards[m]->so.ha_state != xt_rtpproxy_STANDBY) {
				DEBUGP("xt_do_replicate_sessions: switchboard not found or ACTIVE\n");
				return -1;				
			}
			if (sess_ids[m] >= switchboards[m]->so.max_sess) {
				DEBUGP("xt_do_replicate_sessions: sessid out of range %d>=%d\n", sess_ids[m], switchboards[m]->so.max_sess);
				return -1;				
			}
			ss[m] = switchboards[m]->sessions + sess_ids[m];
			if ((flags & REPL_REC_FLAG_DIR_B) == 0) {
				break;
			}
		}
		/* destroy sessions if paired at STANDBY differently than provided in package */
		for (m=0; m<2; m++) {
			sessA = get_sessA(ss[m]);
			if (sessA && (ss[m]->session2 != ss[m^1] || ((m ^ ss[m]->dir) != ((flags & REPL_REC_FLAG_DIR_A_B) != 0)) )) {
				if (ss[m]->session2->switchboard->so.ha_state != xt_rtpproxy_STANDBY) {
					DEBUGP("xt_do_replicate_sessions: switchboard is ACTIVE\n");
					return -1;
				}
				DEBUGP("xt_do_replicate_sessions: sess_id=%d/%d, destroying\n", ss[0]->so_dir.sess_id, ss[1]->so_dir.sess_id);
				set_state(sessA, xt_rtpproxy_NONE); /* so.expires not needed for NONE state */
				ss[m]->switchboard->so.stat.num_session--;
				ss[m]->session2->switchboard->so.stat.num_session--;
				ss[m]->so.last_change_stamp = stamp->tv_sec; 
				ss[m]->session2->so.last_change_stamp = stamp->tv_sec; 
				ss[m]->session2->session2 = NULL;
				ss[m]->session2 = NULL;
				so_global.total.num_pair_session--;
			}				
		}

		if (ss[1]) {
			if (ss[1] != ss[0]->session2) {
				for (m=0; m<2; m++) {
					ss[m]->session2 = ss[m^1];
					ss[m]->switchboard->so.stat.num_session++;
					ss[m]->dir = m ^ ((flags & REPL_REC_FLAG_DIR_A_B) != 0);
				}
				so_global.total.num_pair_session++;
			}
			sessA = get_sessA(ss[0]);			
			sessA->lock = SPIN_LOCK_UNLOCKED;
		} else {
			sessA = ss[0];
		}
		sessA->so.state = state;

		switch (state) {
			case xt_rtpproxy_NONE:
				break;
			case xt_rtpproxy_STALED:
			case xt_rtpproxy_EXPIRED:
			case xt_rtpproxy_DESTROYED:
				if (COPY_FROM_USER(user2, in_len, sessA->so.expires)) {
					DEBUGP_ERR;
					return -1;
				}
				break;
			default:
				sessA->so.state = state;
				sessA->so.last_change_stamp = stamp->tv_sec;
				sessA->so.expires = 0;
				if (COPY_FROM_USER(user2, in_len, sessA->so.created) ||
					COPY_FROM_USER(user2, in_len, sessA->so.expires_timeout) ||
					COPY_FROM_USER(user2, in_len, sessA->so.ttl) ) {
					DEBUGP_ERR;
					return -1;
				}
				for (m=0; m<2; m++) {
					int j;
					ss[m]->so_dir.always_learn = (flags & (REPL_REC_FLAG_ALWAYS_LEARN << m)) != 0;
					for (j=0; j<2; j++) {
						ss[m]->so_dir.stream[j].learning = (flags & (REPL_REC_FLAG_LEARNING << (2*m+j)))?stamp->tv_sec:0;
						ss[m]->so_dir.stream[j].last_packet_stamp = (flags & (REPL_REC_FLAG_LAST_PACKET_STAMP << (2*m+j)))?stamp->tv_sec:0;
						if (flags & (REPL_REC_FLAG_SOURCE_ADDR << (2*m+j))) {
							if (COPY_FROM_USER(user2, in_len, ss[m]->so_dir.stream[j].source)) {
								DEBUGP_ERR;
								return -1;
							}
							if (COPY_FROM_USER(user2, in_len, ss[m]->so_dir.stream[j].source_learnt)) {
								DEBUGP_ERR;
								return -1;
							}
						}
						else {
							memset(&ss[m]->so_dir.stream[j].source, 0, sizeof(ss[m]->so_dir.stream[j].source));
							memset(&ss[m]->so_dir.stream[j].source_learnt, 0, sizeof(ss[m]->so_dir.stream[j].source_learnt));
						}
						if (flags & (REPL_REC_FLAG_LEARNING_TIMEOUT << (2*m+j))) {
							if (COPY_FROM_USER(user2, in_len, ss[m]->so_dir.stream[j].learning_timeout) ) {
								DEBUGP_ERR;
								return -1;
							}
						}
						else {
							ss[m]->so_dir.stream[j].learning_timeout = 0;
						}
						if (flags & (REPL_REC_FLAG_THROTTLE << (2*m+j))) {
							if (COPY_FROM_USER(user2, in_len, ss[m]->so_dir.stream[j].throttle.max_bandwidth) ||
								COPY_FROM_USER(user2, in_len, ss[m]->so_dir.stream[j].throttle.mark) ) {
								DEBUGP_ERR;
								return -1;
							}
						}
						else {
							memset(&ss[m]->so_dir.stream[j].throttle, 0, sizeof(ss[m]->so_dir.stream[j].throttle));
						}
					}
					if ((flags & REPL_REC_FLAG_DIR_B) == 0) 
						break;  /* cannot happend */
				}
		}
		record_count--;
	}
	if (all && switchboards[0]) {
		xt_do_delete_sessions(stamp, switchboards[0], 1);
	}
	return 0;
}
										   
/* there is no uniq cmd numbering space, 2 separates for GET and SET */
#define XT_OPT_SET_CMD 0x8000

/* user context code */
static int xt_do_get_ctl(struct sock *sk, int cmd, void *user, int *len)
{
	struct xt_rtpproxy_sockopt_header in_header;
	struct timeval stamp;
	int ret = -EFAULT;
	int in_len;
	void *user2;

	DEBUGP("xt_do_%cet_ctl(sk=%p, cmd=%d, user=%p, len=%u)\n", (cmd&XT_OPT_SET_CMD)?'s':'g', sk, cmd & ~XT_OPT_SET_CMD, user, *len);
	/* acquire lock here/early because we ensapsulate also statistics changes */
	/* we cannot use spin_lock because copy_from/to_user calls sleep! */
	if (down_interruptible(&xt_usercontext_mutex)) {
		ret = -EINTR;
		goto ret_no_unlock;
	}

	if (!capable(CAP_NET_ADMIN)) {
		so_global.stat.perm++;
		ret = -EPERM;
		goto ret_no_unlock;
	}

	user2 = user;
	in_len = *len;
	if (in_len == 0) {
		memset(&in_header, 0, sizeof(in_header));
	}
	else {
		/* do it item by item because unpredictable alignment when replicating */
		if (COPY_FROM_USER(user2, in_len, in_header) ) {
			goto in_brk;
		}
	}
	do_gettimeofday(&stamp);
	DEBUGP("xt_do_get_ctl: stamp.tv_sec=%ld, in_header={.magic=0x%x, .version=0x%x, .flags=0x%x}\n", stamp.tv_sec, in_header.magic, in_header.version, in_header.flags);

	if (in_header.magic != XT_RTPPROXY_SOCKOPT_MAGIC) {
		PRINTK(KERN_WARNING, "xt_get_ctl: bad magic %x!=%x\n", in_header.magic, XT_RTPPROXY_SOCKOPT_MAGIC);
		goto in_brk;
	}
	if (in_header.version != XT_RTPPROXY_SOCKOPT_VERSION) {
		if (cmd == (XT_RTPPROXY_SO_SET_REPLICATE_SESSIONS|XT_OPT_SET_CMD)) {
			/* because of upgradability HA pair implement extra version check here if necessary*/
		}
		PRINTK(KERN_WARNING, "xt_get_ctl: bad version %d!=%d\n", in_header.version, XT_RTPPROXY_SOCKOPT_VERSION);
		goto in_brk;		
	}
	
	switch (cmd) {
		case XT_RTPPROXY_SO_GET_CONFIG: {
			struct xt_rtpproxy_switchboard_id in_switchboard[XT_RTPPROXY_MAX_GET_SWITCHBOARD];
			struct xt_rtpproxy_switchboard *switchboard;
			xt_rtpproxy_sockopt_count switchboard_count, out_switchboard_count, out_session_count, out_switchboard_count2;
			void *out_user2, *out_switchboard_count_user;

			out_user2 = user;

			if (COPY_FROM_USER(user2, in_len, out_switchboard_count)) {
				DEBUGP_ERR;
				goto brk;
			}	
			DEBUGP("xt_do_get_ctl: out_switchboard_count=%u\n", out_switchboard_count);
			if (COPY_FROM_USER(user2, in_len, out_session_count)) {
				DEBUGP_ERR;
				goto brk;
			}
			DEBUGP("xt_do_get_ctl: out_session_count=%u\n", out_session_count);

			if ((in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS)==0) {
				int i;
				if (COPY_FROM_USER(user2, in_len, switchboard_count)) {
					DEBUGP_ERR;
					goto brk;
				}
				DEBUGP("xt_do_get_ctl: switchboard_count=%u\n", switchboard_count);
				if (switchboard_count > XT_RTPPROXY_MAX_GET_SWITCHBOARD) {
					goto in_brk;
				}
				for (i=0; i < switchboard_count; i++) {
					if (COPY_FROM_USER(user2, in_len, in_switchboard[i])) {
						DEBUGP_ERR;
						goto brk;
					}
				}
			}

			DEBUGP("xt_do_get_ctl: so_global={.version=0x%x}\n", so_global.version);
			if (COPY_TO_USER(out_user2, so_global)) {
				DEBUGP_ERR;					
				goto brk;
			}
			out_switchboard_count_user = out_user2;
			out_switchboard_count2 = 0;
			if (COPY_TO_USER(out_user2, out_switchboard_count2)) {
				DEBUGP_ERR;
				goto brk;
			}

			if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS) {
				for (switchboard = switchboard_reg; switchboard; switchboard = switchboard->next) {
					if (xt_do_get_switchboard_config(&stamp, &out_user2, switchboard, in_header.flags, out_switchboard_count2 >= out_switchboard_count, &out_session_count) < 0) {
						goto brk;
					};
					if (out_switchboard_count2 < out_switchboard_count) {
						out_switchboard_count2++;
					}
				}
			}
			else {
				int i;
				for (i=0; i < switchboard_count; i++) {
					switchboard = xt_find_switchboard(in_switchboard+i);
					if (switchboard) {
						if (xt_do_get_switchboard_config(&stamp, &out_user2, switchboard, in_header.flags, out_switchboard_count2 >= out_switchboard_count, &out_session_count) < 0) {
							goto brk;
						}
						if (out_switchboard_count2 < out_switchboard_count) {
							out_switchboard_count2++;
						}
					}
				}
			}
			if (copy_to_user(out_switchboard_count_user, &out_switchboard_count2, sizeof(out_switchboard_count2))) {
				DEBUGP_ERR;
				goto brk;
			}

			*len = out_user2-user;
			ret = 0;
			break;
		}

		case XT_RTPPROXY_SO_GET_ALLOC_SESSIONS: {
		
			xt_rtpproxy_session_ptr_arr sessions1, sessions2, *ss[2];
			xt_rtpproxy_sockopt_count session_count;
			int i, m;

			if (COPY_FROM_USER(user2, in_len, session_count)) {
				DEBUGP_ERR;
				goto brk;
			}
			DEBUGP("xt_do_get_ctl: session_count=%u\n", session_count);
			if (!session_count || session_count > XT_RTPPROXY_MAX_ALLOC_SESSION) {
				goto in_brk;
			}

			if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_SWITCHBOARD_GROUPS) {
				switch (xt_do_allocate_sessions(&stamp, &user2, &in_len, session_count, &sessions1)) {
					case -1: 
						goto in_brk;
					case -2: 
						goto brk;
					default: ;
				}
				switch (xt_do_allocate_sessions(&stamp, &user2, &in_len, session_count, &sessions2)) {
					case -1: 
						goto in_brk;
					case -2: 
						goto brk;
					default: ;
				}
			}
			ss[0] = &sessions1;
			ss[1] = &sessions2;
			for (i=0; i<session_count; i++) {
				struct xt_rtpproxy_session *sessA;
				struct xt_rtpproxy_sockopt_session in_sess;

				if (COPY_FROM_USER(user2, in_len, in_sess)) {
					DEBUGP_ERR;
					goto brk_lock_i_unlock;
				}
				if ((in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_SWITCHBOARD_GROUPS) == 0) {
					/* switchboard explicitly defined in in_sess */
					struct xt_rtpproxy_switchboard *switchboard;
					for (m=0; m<2; m++) {
						switchboard = xt_find_switchboard(&in_sess.dir[m].switchboard.addr);
						if (!switchboard || 
							switchboard->so.ha_state == xt_rtpproxy_STANDBY || 
							xt_do_allocate_sessions2(&stamp, switchboard, 1, i, ss[m]) < 0 ) {

							write_lock_bh(&rw_lock);
							for (i--; i>=0; i--) {
								sessions1[i]->so.state = xt_rtpproxy_NONE;
								sessions1[i]->session2 = NULL;
								sessions2[i]->session2 = NULL;
							}
							write_unlock_bh(&rw_lock);
							goto in_brk;
						}
					}
				}

				write_lock_bh(&rw_lock);
				sessA = (*ss[0])[i];
				sessA->lock = SPIN_LOCK_UNLOCKED;
				sessA->so.state = xt_rtpproxy_NONE;
				sessA->so.last_change_stamp = stamp.tv_sec;
				
				sessA->so.created = stamp.tv_sec;

				if (in_sess.sh.flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_EXPIRES) {
					sessA->so.expires_timeout = in_sess.sh.expires_timeout;
					sessA->so.expires = stamp.tv_sec + in_sess.sh.expires_timeout;
				}
				else {
					sessA->so.expires_timeout = 0;
					sessA->so.expires = 0;   /* may be overwritten by learning timeout */
				}
				sessA->so.ttl = (in_sess.sh.flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_TTL)?in_sess.sh.ttl:0;
				for (m = 0; m<=1; m++) {
					int j;
					(*ss[m])[i]->dir = m^1;
					(*ss[m])[i]->session2 = (*ss[m^1])[i];
					DEBUGP("xt_do_get_ctl: allocating sess_id=%d, created=%u, expires=%u, ttl=%u\n", (*ss[m])[i]->so_dir.sess_id, sessA->so.created, sessA->so.expires, sessA->so.ttl);
					for (j=0; j<2; j++) {
						if (in_sess.dir[m].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR) {
							(*ss[m])[i]->so_dir.stream[j].source = in_sess.dir[m].stream[j].source;
							(*ss[m])[i]->so_dir.stream[j].source_learnt = in_sess.dir[m].stream[j].source;
							(*ss[m])[i]->so_dir.stream[j].flags &= ~XT_RTPPROXY_SOCKOPT_FLAG_LEARN_PORT_ONLY;
							(*ss[m])[i]->so_dir.stream[j].flags |= in_sess.dir[m].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_LEARN_PORT_ONLY;
						} else {
							memset(&(*ss[m])[i]->so_dir.stream[j].source, 0, sizeof((*ss[m])[i]->so_dir.stream[j].source));
							memset(&(*ss[m])[i]->so_dir.stream[j].source_learnt, 0, sizeof((*ss[m])[i]->so_dir.stream[j].source_learnt));
						}

						if (in_sess.dir[m].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_LEARNING_TIMEOUT &&
						    in_sess.dir[m].stream[j].learning_timeout) {

							(*ss[m])[i]->so_dir.stream[j].learning_timeout = in_sess.dir[m].stream[j].learning_timeout;
							(*ss[m])[i]->so_dir.stream[j].learning = stamp.tv_sec + in_sess.dir[m].stream[j].learning_timeout;
							if ((*ss[m])[i]->so_dir.stream[j].learning > sessA->so.expires) {
								sessA->so.expires = (*ss[m])[i]->so_dir.stream[j].learning;
							}	
						} 
						else {
								(*ss[m])[i]->so_dir.stream[j].learning = 0;
								(*ss[m])[i]->so_dir.stream[j].learning_timeout = 0;						
						}
						memset(&(*ss[m])[i]->so_dir.stream[j].throttle, 0, sizeof((*ss[m])[i]->so_dir.stream[j].throttle));
						if (in_sess.dir[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_THROTTLE_MARK) {
							(*ss[m])[i]->so_dir.stream[j].throttle.mark = in_sess.dir[i].stream[j].throttle.mark;
						}
						if (in_sess.dir[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_THROTTLE_BANDWIDTH) {
							(*ss[m])[i]->so_dir.stream[j].throttle.max_bandwidth = in_sess.dir[i].stream[j].throttle.max_bandwidth;
						}

						if ((*ss[m])[i]->so_dir.stream[j].learning || (*ss[m])[i]->so_dir.stream[j].source_learnt.ip) {
							/* maybe both sides initialized ? */
							sessA->so.state = (sessA->so.state == xt_rtpproxy_INIT1) ? xt_rtpproxy_INIT2 : xt_rtpproxy_INIT1;
						}	
						if (in_sess.dir[m].flags & XT_RTPPROXY_SOCKOPT_FLAG_ALWAYS_LEARN) {
							(*ss[m])[i]->so_dir.always_learn = in_sess.dir[m].always_learn;
						}
						memset(&(*ss[m])[i]->so_dir.stat, 0, sizeof((*ss[m])[i]->so_dir.stat));			
					}
				}
					
				if (sessA->so.state == xt_rtpproxy_NONE) { /* we need at least one source addr or learning active */
					goto brk_i_unlock;
				}
				write_unlock_bh(&rw_lock);
			}
			so_global.total.num_pair_session += session_count;
			for (i=0; i<session_count; i++) {
				for (m = 0; m <= 1; m++) {
					/* it's allocated, does not matter if theoretically copy_to_user() fails */
					(*ss[m])[i]->switchboard->so.stat.counter.alloc++;
					(*ss[m])[i]->switchboard->so.stat.num_session++;
					if (!(*ss[m])[i]->switchboard->cps.stamp) {
						(*ss[m])[i]->switchboard->cps.stamp = stamp.tv_sec; 
					} else if ((*ss[m])[i]->switchboard->cps.stamp + CPS_INTERVAL <= stamp.tv_sec) {
						(*ss[m])[i]->switchboard->cps.rate = (*ss[m])[i]->switchboard->cps.counter;
						(*ss[m])[i]->switchboard->cps.counter = 0;
						(*ss[m])[i]->switchboard->cps.stamp = stamp.tv_sec; 
					}
					(*ss[m])[i]->switchboard->cps.counter++;
				}
			}

			user2 = user;
			if (COPY_TO_USER(user2, session_count)) {
				DEBUGP_ERR;
				goto brk;
			}
			for (i=0; i<session_count; i++) {
				struct xt_rtpproxy_sockopt_session out_sess;
				spin_lock_bh(&sessions1[i]->lock);
				out_sess.sh = sessions1[i]->so;
				
				for (m=0; m<2; m++) {
					out_sess.dir[m] = (*ss[m])[i]->so_dir;
					out_sess.dir[m].switchboard.alloc = (*ss[m])[i]->switchboard->so.stat.num_session;
					out_sess.dir[m].switchboard.free = (*ss[m])[i]->switchboard->so.max_sess-out_sess.dir[m].switchboard.alloc;
			
				}
				out_sess.dir[1] = sessions2[i]->so_dir;
				spin_unlock_bh(&sessions1[i]->lock);
				DEBUGP("xt_do_get_ctl: copy_to_user(out_sess[%d]})\n", i);
				if (COPY_TO_USER(user2, out_sess)) {
					DEBUGP_ERR;
					goto brk;
				}
			}
			*len = user2-user;
			ret = 0;
			break;

		brk_lock_i_unlock:
			write_lock_bh(&rw_lock);
		brk_i_unlock:
			for (; i>=0; i--) {
				sessions1[i]->so.state = xt_rtpproxy_NONE;
				sessions1[i]->session2 = NULL;
				sessions2[i]->session2 = NULL;
				sessions2[i]->so.last_change_stamp = stamp.tv_sec;
			}
			write_unlock_bh(&rw_lock);
			goto brk;
		}

		case XT_RTPPROXY_SO_GET_UPDATE_SESSIONS: { 
			xt_rtpproxy_sockopt_count session_count;
			enum xt_rtpproxy_session_state new_state;
			u_int32_t new_expires;			
			int k;
			void *out_user2;
			
			if (COPY_FROM_USER(user2, in_len, session_count)) {
				DEBUGP_ERR;
				goto brk;
			}			
			DEBUGP("xt_do_get_ctl: session_count=%u\n", session_count);
			if (!session_count) {
				goto in_brk;
			}
			out_user2 = user2;
			for (k=0; k<session_count; k++) {
				struct xt_rtpproxy_session *ss[2], *sessA;
				struct xt_rtpproxy_sockopt_session in_sess;
				int i, j;
				
				if (COPY_FROM_USER(user2, in_len, in_sess)) {
					DEBUGP_ERR;
					goto brk;
				}
				in_sess.sh.state = xt_rtpproxy_NONE;
				DEBUGP("xt_do_get_ctl: in_sess[%d]={"
						".addr[A]=0x%x:%u:%d,"
						".addr[B]=0x%x:%u:%d,"
						".sh.created=%u,"
						".sh.expires_timeout=%u,"
						".sh.ttl=%u,"
						".sh.flags=%x,"
						".source[A]={"						
						"RTP={.addr=0x%x:%u, .learning_timeout=%u, .flags=%x},"
						"RTCP={.addr=0x%x:%u, .learning_timeout=%u, .flags=%x},"
						".always_learn=%u,. flags=%x},"
						".source[B]={"
						"RTP={.addr=0x%x:%u, .learning_timeout=%u, .flags=%x},"
						"RTCP={.addr=0x%x:%u, .learning_timeout=%u, .flags=%x},"
						".always_learn=%u, .flags=%x}"
					"}\n",
					k,
					in_sess.dir[0].switchboard.addr.ip, in_sess.dir[0].switchboard.addr.port, in_sess.dir[0].sess_id,
					in_sess.dir[1].switchboard.addr.ip, in_sess.dir[1].switchboard.addr.port, in_sess.dir[1].sess_id,
					in_sess.sh.created,
					in_sess.sh.expires_timeout,
					in_sess.sh.ttl,
					in_sess.sh.flags,
					in_sess.dir[0].stream[0].source.ip, in_sess.dir[0].stream[0].source.port, in_sess.dir[0].stream[0].learning_timeout, in_sess.dir[0].stream[0].flags,
					in_sess.dir[0].stream[1].source.ip, in_sess.dir[0].stream[1].source.port, in_sess.dir[0].stream[1].learning_timeout, in_sess.dir[0].stream[1].flags,
					in_sess.dir[0].always_learn, in_sess.dir[0].flags,
					in_sess.dir[1].stream[0].source.ip, in_sess.dir[1].stream[0].source.port, in_sess.dir[1].stream[0].learning_timeout, in_sess.dir[1].stream[0].flags,
					in_sess.dir[1].stream[1].source.ip, in_sess.dir[1].stream[1].source.port, in_sess.dir[1].stream[1].learning_timeout, in_sess.dir[1].stream[1].flags,
					in_sess.dir[1].always_learn, in_sess.dir[1].flags
				);
				ss[0] = xt_find_session(&in_sess);
				if (!ss[0]) {
					in_sess.sh.flags |= XT_RTPPROXY_SOCKOPT_FLAG_NOT_FOUND;
					goto copy_in_sess;
				}
				sessA = xt_do_session_audit(&stamp, ss[0], 1);
				if (!sessA) {
					/* expired or destroyed, we cannot update such sessions */
					in_sess.sh.flags |= XT_RTPPROXY_SOCKOPT_FLAG_NOT_FOUND;
					goto copy_in_sess;
				}
				ss[1] = ss[0]->session2;

				if (in_sess.sh.flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_INFO) 
					goto unlock_continue;
				if (ss[0]->switchboard->so.ha_state == xt_rtpproxy_STANDBY || ss[1]->switchboard->so.ha_state == xt_rtpproxy_STANDBY) {
					goto unlock_continue;
				}

				sessA->so.last_change_stamp = stamp.tv_sec; 
				if (in_sess.sh.flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_DESTROY) {
					switch (sessA->so.state) {
						case xt_rtpproxy_INIT1:
						case xt_rtpproxy_INIT2:
						case xt_rtpproxy_FORWARD1:
						case xt_rtpproxy_FORWARD2:
							DEBUGP("xt_do_get_ctl: sess_id=%d/%d, destroying\n", ss[0]->so_dir.sess_id, ss[1]->so_dir.sess_id);
							sessA->so.expires = stamp.tv_sec;  /* required for and packet dropping and ressurection */
							set_state(sessA, xt_rtpproxy_DESTROYED);
							ss[0]->switchboard->so.stat.counter.destroyed++;
							if (ss[0]->switchboard != ss[1]->switchboard) {
								ss[1]->switchboard->so.stat.counter.destroyed++;
							}
							break;
						default: ;
					}
					
					goto unlock_continue;
				}
				if (in_sess.sh.flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_EXPIRES) {
					sessA->so.expires_timeout = in_sess.sh.expires_timeout;
				}
				if (in_sess.sh.flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_TTL) {
					sessA->so.ttl = in_sess.sh.ttl;
				}
				new_state = xt_rtpproxy_NONE;
				new_expires = stamp.tv_sec + sessA->so.expires_timeout;
				DEBUGP("xt_do_get_ctl: sess_id=%d(%d), new_expires=%u, new_ttl=%u, updating\n", ss[0]->so_dir.sess_id, ss[1]->so_dir.sess_id, new_expires, sessA->so.ttl);
				for (i=0; i<2; i++) {
					for (j=0; j<2; j++) {
						if (in_sess.dir[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR) {
							if (ss[i]->so_dir.stream[j].source.ip != in_sess.dir[i].stream[j].source.ip ||
							    ss[i]->so_dir.stream[j].source.port != in_sess.dir[i].stream[j].source.port) {
								ss[i]->so_dir.stream[j].source = in_sess.dir[i].stream[j].source;
								ss[i]->so_dir.stream[j].source_learnt = in_sess.dir[i].stream[j].source;
								new_state = xt_rtpproxy_INIT1;
							}
							ss[i]->so_dir.stream[j].flags &= ~XT_RTPPROXY_SOCKOPT_FLAG_LEARN_PORT_ONLY;
							ss[i]->so_dir.stream[j].flags |= in_sess.dir[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_LEARN_PORT_ONLY;
						}

						if (in_sess.dir[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_LEARNING_TIMEOUT) {
							ss[i]->so_dir.stream[j].learning_timeout = in_sess.dir[i].stream[j].learning_timeout;
							if (in_sess.dir[i].stream[j].learning_timeout) {
								ss[i]->so_dir.stream[j].learning = stamp.tv_sec + in_sess.dir[i].stream[j].learning_timeout;
								if (ss[i]->so_dir.stream[j].learning > new_expires)
									new_expires = ss[i]->so_dir.stream[j].learning;
								new_state = xt_rtpproxy_INIT1;
							}
							else
								ss[i]->so_dir.stream[j].learning = 0;
						}
						if (in_sess.dir[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_THROTTLE_MARK) {
							ss[i]->so_dir.stream[j].throttle.mark = in_sess.dir[i].stream[j].throttle.mark;
						}
						if (in_sess.dir[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_THROTTLE_BANDWIDTH) {
							ss[i]->so_dir.stream[j].throttle.max_bandwidth = in_sess.dir[i].stream[j].throttle.max_bandwidth;
						}
					}
					if (in_sess.dir[i].flags & XT_RTPPROXY_SOCKOPT_FLAG_ALWAYS_LEARN) {
						ss[i]->so_dir.always_learn = in_sess.dir[i].always_learn;
					}
				}
				if (new_state == xt_rtpproxy_NONE) {
					if (in_sess.sh.flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_EXPIRES) {
						new_state = sessA->so.state;
						goto skip_init2;
					}
					/* we do not allow session destroying or do not update when no change made */
					goto unlock_continue;
				}

				/* maybe both sides initialized ? */
				for (j=0; j<2; j++) {
					for (i=0; i<2; i++) {
						if (!ss[i]->so_dir.stream[j].learning && !ss[i]->so_dir.stream[j].source_learnt.ip)
							break;
					}
					if (i >= 2) { /* all source IPs known/learning for at least one stream ? */
						new_state = xt_rtpproxy_INIT2;
						break;
					}
				}
			skip_init2:
				sessA->so.expires = new_expires;
				ss[0]->switchboard->so.stat.counter.updated++;
				if (ss[0]->switchboard != ss[1]->switchboard) {
					ss[1]->switchboard->so.stat.counter.updated++;
				}

				/* we change always to INIT state (even already in FORWARD state) */
				set_state(sessA, new_state);
			unlock_continue:
				in_sess.sh = sessA->so;
				for (i=0; i<2; i++) {
					in_sess.dir[i] = ss[i]->so_dir;
					in_sess.dir[i].switchboard.alloc = ss[i]->switchboard->so.stat.num_session;
					in_sess.dir[i].switchboard.free = ss[i]->switchboard->so.max_sess-in_sess.dir[i].switchboard.alloc;
				}
				spin_unlock_bh(&sessA->lock);
			copy_in_sess:
				if (COPY_TO_USER(out_user2, in_sess)) {
					DEBUGP_ERR;
					goto brk;
				}
			}
			*len = out_user2-user;
			ret = 0;
			break;
		}
		case XT_RTPPROXY_SO_SET_DELETE_SWITCHBOARD_SESSIONS|XT_OPT_SET_CMD: { 
			struct xt_rtpproxy_switchboard_id in_switchboard;
			struct xt_rtpproxy_switchboard *switchboard;
			xt_rtpproxy_sockopt_count switchboard_count;

			if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS) {
				write_lock_bh(&rw_lock);
				for (switchboard = switchboard_reg; switchboard; switchboard = switchboard->next) {
					if (switchboard->so.ha_state != xt_rtpproxy_STANDBY) {
						xt_do_delete_sessions(&stamp, switchboard, 0);
					}
				}
				write_unlock_bh(&rw_lock);
			}
			else {
				if (COPY_FROM_USER(user2, in_len, switchboard_count)) {
					DEBUGP_ERR;
					goto brk;
				}
				DEBUGP("xt_do_get_ctl: switchboard_count=%u\n", switchboard_count);
				while (switchboard_count > 0) {
					switchboard_count--;
						
					if (COPY_FROM_USER(user2, in_len, in_switchboard)) {
						DEBUGP_ERR;
						goto brk;
					}
					switchboard = xt_find_switchboard(&in_switchboard);
					if (switchboard && switchboard->so.ha_state != xt_rtpproxy_STANDBY) {
						write_lock_bh(&rw_lock);
						xt_do_delete_sessions(&stamp, switchboard, 0);
						write_unlock_bh(&rw_lock);
					}
				}
			}
			ret = 0;
			break;
		}
		case XT_RTPPROXY_SO_SET_HA_STATE|XT_OPT_SET_CMD: { 
			u_int8_t new_ha_state;
			struct xt_rtpproxy_switchboard_id in_switchboard;
			struct xt_rtpproxy_switchboard *switchboard;
			xt_rtpproxy_sockopt_count switchboard_count;

			if (COPY_FROM_USER(user2, in_len, new_ha_state)) {
				DEBUGP_ERR;
				goto brk;
			}
			if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS) {
				write_lock_bh(&rw_lock);
				for (switchboard = switchboard_reg; switchboard; switchboard = switchboard->next) {
					if (switchboard && switchboard->so.ha_state != new_ha_state) {
						switchboard->so.ha_state = new_ha_state;
						xt_adjust_ha_state(&stamp, switchboard);
					}
				}
				write_unlock_bh(&rw_lock);
			}
			else {
				if (COPY_FROM_USER(user2, in_len, switchboard_count)) {
					DEBUGP_ERR;
					goto brk;
				}
				DEBUGP("xt_do_get_ctl: switchboard_count=%u\n", switchboard_count);
				while (switchboard_count > 0) {
					switchboard_count--;
						
					if (COPY_FROM_USER(user2, in_len, in_switchboard)) {
						DEBUGP_ERR;
						goto brk;
					}
					switchboard = xt_find_switchboard(&in_switchboard);
					if (switchboard && switchboard->so.ha_state != new_ha_state) {
						write_lock_bh(&rw_lock);
						switchboard->so.ha_state = new_ha_state;
						xt_adjust_ha_state(&stamp, switchboard);
						write_unlock_bh(&rw_lock);
					}
				}
			}
			ret = 0;
			break;
		}
		case XT_RTPPROXY_SO_GET_RETRIEVE_SESSIONS: {
			struct xt_rtpproxy_switchboard_id in_switchboard[XT_RTPPROXY_MAX_GET_SWITCHBOARD];
			struct xt_rtpproxy_switchboard *switchboard;
			struct xt_rtpproxy_sockopt_in_interval in_interval;
			xt_rtpproxy_sockopt_count switchboard_count;
			void *out_user2;
			unsigned int out_len;
			void *out_record_count_user2;
			u_int32_t out_record_count;

			out_user2 = user;
			out_len = *len;
			out_record_count = 0;

			if ((in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS)==0) {
				int i;
				if (COPY_FROM_USER(user2, in_len, switchboard_count)) {
					DEBUGP_ERR;
					goto brk;
				}
				DEBUGP("xt_do_get_ctl: switchboard_count=%u\n", switchboard_count);
				if (switchboard_count > XT_RTPPROXY_MAX_GET_SWITCHBOARD) {
					goto in_brk;
				}
				for (i=0; i < switchboard_count; i++) {
					if (COPY_FROM_USER(user2, in_len, in_switchboard[i])) {
						DEBUGP_ERR;
						goto brk;
					}
				}
			}
			else {
				switchboard_count = 0;
			}
			if ((in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS) == 0) {
				if (COPY_FROM_USER(user2, in_len, in_interval)) {
					DEBUGP_ERR;
					goto brk;
				}
				DEBUGP("xt_do_get_ctl: interval {from_stamp=%u, to_stamp=%u}\n", in_interval.from_stamp, in_interval.to_stamp);
				if (!in_interval.to_stamp && !in_interval.from_stamp) {
					in_header.flags |= XT_RTPPROXY_SOCKOPT_FLAG_ALL_SESSIONS;
				}
				if (!in_interval.to_stamp) {
					in_interval.to_stamp = stamp.tv_sec;
				}
			}
			else {
				in_interval.from_stamp = 0;
				in_interval.to_stamp = stamp.tv_sec;
			}
			/* put item by item because of unpredictable alignment */
			if (COPY_TO_USER_LEN(out_user2, out_len, in_header)) {
				DEBUGP_ERR;
				goto brk;
			}
			if (COPY_TO_USER_LEN(out_user2, out_len, in_interval)) {
				DEBUGP_ERR;
				goto brk;
			}
			out_record_count_user2 = out_user2;
			if (COPY_TO_USER_LEN(out_user2, out_len, out_record_count)) {
				DEBUGP_ERR;
				goto brk;
			}
			repl_retrieve_counter++;
			if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_ALL_SWITCHBOARDS) {
				for (switchboard = switchboard_reg; switchboard; switchboard = switchboard->next) {
					if (xt_do_repl_retrieve_switchboard(&stamp, &out_user2, &out_len, &in_header, &in_interval, switchboard, &out_record_count) < 0) {
						goto brk;
					}
				}
			}
			else {
				int i;
				for (i=0; i < switchboard_count; i++) {
					switchboard = xt_find_switchboard(in_switchboard+i);
					if (switchboard) {
						if (xt_do_repl_retrieve_switchboard(&stamp, &out_user2, &out_len, &in_header, &in_interval, switchboard, &out_record_count) < 0) {
							goto brk;
						}
					}
				}
			}
			if (out_user2-user > *len) {
				u_int32_t req_mem;
				req_mem = out_user2-user;
				out_user2 = user;
				in_header.magic = 0;
				if (COPY_TO_USER(out_user2, in_header.magic)) {
					DEBUGP_ERR;
					goto brk;
				}
				if (COPY_TO_USER(out_user2, req_mem)) {
					DEBUGP_ERR;
					goto brk;
				}
			} else {
				if (COPY_TO_USER(out_record_count_user2, out_record_count)) {
					DEBUGP_ERR;
					goto brk;
				}
			}
			*len = out_user2-user;
			ret = 0;
			break;			
		}
		case XT_RTPPROXY_SO_SET_REPLICATE_SESSIONS|XT_OPT_SET_CMD: {
			struct xt_rtpproxy_sockopt_in_interval in_interval;
			u_int32_t record_count;

			if (COPY_FROM_USER(user2, in_len, in_interval)) {
				DEBUGP_ERR;
				goto brk;
			}
			if (COPY_FROM_USER(user2, in_len, record_count)) {
				DEBUGP_ERR;
				goto brk;
			}
			repl_retrieve_counter++;
			/* write_lock_bh(&rw_lock);  not required because only STANDBY switchboards may be affected */
			if (xt_do_replicate_sessions(&stamp, user2, in_len, &in_header, &in_interval, record_count) < 0) {
				/* write_unlock_bh(&rw_lock); */
				goto brk;
			}
			/* write_unlock_bh(&rw_lock); */

			ret = 0;
			break;
		}																   	
		default:												   
			so_global.stat.unknown++;			
			PRINTK(KERN_WARNING, "xt_%s_ctl: unknown request %i\n", cmd&XT_OPT_SET_CMD?"set":"get", cmd&~XT_OPT_SET_CMD);
			ret = -EINVAL;
			goto ret;
	}
	if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_RESET_GLOBAL_STAT) {
		memset(&so_global.stat, 0, sizeof(so_global.stat));
	}
brk:
	if (cmd & XT_OPT_SET_CMD)
		so_global.stat.cmd[(cmd&~XT_OPT_SET_CMD)-XT_RTPPROXY_SO_BASE_CTL]++;
	else {
		DEBUGP("xt_get_ctl: output len: %d\n", *len);
		so_global.stat.cmd[cmd-XT_RTPPROXY_SO_BASE_CTL+XT_RTPPROXY_SO_SET_MAX-XT_RTPPROXY_SO_BASE_CTL+1]++;
	}
in_brk:	/* bad input params */
	if (ret == 0)
		so_global.stat.ok++;
	else
		so_global.stat.fail++;
ret:
	up(&xt_usercontext_mutex);

ret_no_unlock:
	DEBUGP("xt_do_get_ctl: ret=%d\n", ret);
	return ret;
}

static int xt_do_set_ctl(struct sock *sk, int cmd,
                               void *user, u_int32_t len)
{
	return xt_do_get_ctl(sk, cmd|XT_OPT_SET_CMD, user, (int *) &len);
}

/* user-context */
static 
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
	int 
#else
	bool
#endif
		do_checkentry(
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)
				const char *table,
				const void *entryinfo,
				const struct xt_target *target,
				void *targinfo,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
				unsigned int targinfosize,
#endif
				unsigned int hook_mask
#else
				const struct xt_tgchk_param *params
#endif
		)
{
	/*const*/ struct xt_rtpproxy_target_info *info;
	const struct ipt_entry *e; 
	int j;
	unsigned long l1;
	struct xt_rtpproxy_switchboard *sw, **psw;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)
	#define CE_PARAMS(f) f
#else
	#define CE_PARAMS(f) params->f
#endif
	info = CE_PARAMS(targinfo);
	e = CE_PARAMS(entryinfo);

	DEBUGP("do_checkentry(table='%s', e=%p, targinfo=%p, hook_mask=%x)\n", CE_PARAMS(table), e, CE_PARAMS(targinfo), CE_PARAMS(hook_mask));

#if defined(CONFIG_NETFILTER_XT_TARGET_RTPPROXY_USE_CONNTRACK)
#  if defined(CONFIG_NF_CONNTRACK) || defined(CONFIG_NF_CONNTRACK_MODULE)
	if (nf_ct_l3proto_try_module_get(CE_PARAMS(target)->family) < 0) {
		PRINTK(KERN_WARNING, "do_checkentry(): can't load conntrack support for proto=%d\n", CE_PARAMS(target)->family);
		return 0;
	}
#  endif
#endif
	
	if (strcmp(CE_PARAMS(table), "mangle") != 0) {
		PRINTK(KERN_WARNING, "bad table '%s', use the 'mangle' table.\n",
			   CE_PARAMS(table));
		return 0;
	}
	if (CE_PARAMS(hook_mask) & ~(  (1 << NF_IP_PRE_ROUTING)
			    | (1 << NF_IP_LOCAL_OUT)
			    | (1 << NF_IP_POST_ROUTING)
			    | (1 << NF_IP_LOCAL_IN)) ) {
		PRINTK(KERN_WARNING, "do_checkentry(): bad hook_mask=%x\n", CE_PARAMS(hook_mask));
		return 0;
	}

	/* Must specify that it's a UDP packet */
	if (e->ip.proto != IPPROTO_IP) {  /* if specified a protocol parameter */
		if (e->ip.proto != IPPROTO_UDP || (e->ip.invflags & XT_INV_PROTO)) {
			PRINTK(KERN_WARNING, "not valid for non-udp (proto:%c0x%x)\n",(e->ip.invflags & XT_INV_PROTO)?'!':' ', e->ip.proto);
			return 0;
		}
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
	if (targinfosize != XT_ALIGN(sizeof(*info))) {
		PRINTK(KERN_WARNING, "targinfosize %u != %u\n",
		       targinfosize,
		       XT_ALIGN(sizeof(struct xt_rtpproxy_target_info)));
		return 0;
	}
#endif
	if ( !info->addr.port) {
		PRINTK(KERN_WARNING, "begin of port range is too low\n");
		return 0;
	}
	l1 = info->addr.port + 2*info->max_sess;
	if ( l1 > 0xFFFFL) {
		PRINTK(KERN_WARNING, "end of port range is too high\n");
		return 0;
	}

	if (down_interruptible(&xt_usercontext_mutex)) {
		return -EINTR;
	}
	/* check if session is already registered */
	for (info->switchboard = switchboard_reg; info->switchboard; info->switchboard = info->switchboard->next ) {
		if (info->switchboard->so.addr.ip == info->addr.ip) {
			if (info->switchboard->so.addr.port == info->addr.port && info->switchboard->so.max_sess == info->max_sess) {
				DEBUGP("do_checkentry(): found switchboard=%p\n", info->switchboard);
				goto found;
			}
			else {
				unsigned long l2;
				l2 = info->switchboard->so.addr.port + 2*info->switchboard->so.max_sess;
				if ( (l2 > info->addr.port && info->switchboard->so.addr.port < l1) ||
					 (l1 > info->switchboard->so.addr.port && info->addr.port < l2) ) {
					up(&xt_usercontext_mutex);
					PRINTK(KERN_WARNING, "switchboard overlaps another switchboard\n");
					return 0;
				}
			}
		}
	}

	DEBUGP("do_checkentry(): allocating new switchboard\n");
	DEBUGP("do_checkentry(): vmalloc(%u)\n", (unsigned int) sizeof(*info->switchboard));
	info->switchboard = vmalloc(sizeof(*info->switchboard)); /* alternatively kmalloc(, GPF_KERNEL); */
	if (!info->switchboard) {
		PRINTK(KERN_WARNING, "cannot allocate %u bytes\n",
			(unsigned int) sizeof(*info->switchboard));
		return 0;
	}
	DEBUGP("do_checkentry(): allocated info->switchboard=%p\n", info->switchboard);
	memset(info->switchboard, 0, sizeof(*(info->switchboard)));
	info->switchboard->so.max_sess = info->max_sess;
	info->switchboard->so.ha_state = info->ha_state;
	info->switchboard->so.stat.state[xt_rtpproxy_NONE] = info->max_sess;
	info->switchboard->so.addr = info->addr;
	DEBUGP("do_checkentry(): vmalloc(%u)\n", (unsigned int) sizeof(*info->switchboard->sessions)*info->max_sess);
	info->switchboard->sessions = vmalloc(sizeof(*info->switchboard->sessions)*info->max_sess);
	if (!info->switchboard->sessions) {
		PRINTK(KERN_WARNING, "cannot allocate %u bytes\n",
			(unsigned int) sizeof(*info->switchboard->sessions)*info->max_sess);
		vfree(info->switchboard);
		return 0;
	}
	DEBUGP("do_checkentry(): allocated info->switchboard->sessions=%p\n", info->switchboard->sessions);
	memset(info->switchboard->sessions, 0, sizeof(*info->switchboard->sessions)*info->max_sess);
	for (j=0; j<info->max_sess; j++) {
		int i;
			
		info->switchboard->sessions[j].so_dir.sess_id = j;
		info->switchboard->sessions[j].so_dir.switchboard.addr = info->switchboard->so.addr;
		info->switchboard->sessions[j].switchboard = info->switchboard;
		info->switchboard->sessions[j].lock = SPIN_LOCK_UNLOCKED;
		for (i=0; i<=1; i++) {
			info->switchboard->sessions[j].so_dir.stream[i].port = info->addr.port+2*j+i;
		}
	}
	/* add self at the top of registration chain */
	/* rw_lock need not be acquired because softirq need not access switchboard chain */ 
	/* let's sort by ip, port */

	for (psw = &switchboard_reg, sw = NULL; *psw; sw = *psw, psw = &(*psw)->next) {
		if ( (*psw)->so.addr.ip > info->switchboard->so.addr.ip || 
			((*psw)->so.addr.ip == info->switchboard->so.addr.ip && (*psw)->so.addr.port > info->switchboard->so.addr.port) ) {
			break;
		}
	}
	info->switchboard->next = *psw;
	if (*psw) {
		(*psw)->prev = info->switchboard;
	}
	*psw = info->switchboard;
	info->switchboard->prev = sw;

	so_global.total.num_switchboard++;
		
found:
	info->switchboard->usage++;
	DEBUGP("do_checkentry(): info->switchboard->usage=%d\n", info->switchboard->usage);
	up(&xt_usercontext_mutex);
	return 1;
}

/* user-context */
static void do_destroy(
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
	const struct xt_target *target,
#endif
    void *targinfo
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
	, unsigned int targetsize
#endif
#else
	const struct xt_tgdtor_param *params
#endif
					)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)
	#define DE_PARAMS(f) f
#else
	#define DE_PARAMS(f) params->f
#endif
	struct xt_rtpproxy_target_info *info;
	info = DE_PARAMS(targinfo);
	DEBUGP("destroy(targetinfo=%p)\n", DE_PARAMS(targinfo));
	down(&xt_usercontext_mutex);
	write_lock_bh(&rw_lock); /* write lock for switchboard_reg chain to see it interrupt-context consistently */
	info->switchboard->usage--;
	if (!info->switchboard->usage) {
		struct timeval stamp;
		do_gettimeofday(&stamp);
		xt_do_delete_sessions(&stamp, info->switchboard, 0);

		if (info->switchboard->prev) {
			info->switchboard->prev->next = info->switchboard->next;
		}
		else
			switchboard_reg = info->switchboard->next;
		if (info->switchboard->next)
			info->switchboard->next->prev = info->switchboard->prev;
		so_global.total.num_switchboard--;
	}
	else {
		DEBUGP("do_destroy(): info->switchboard->usage=%d\n", info->switchboard->usage);
		info = NULL;  /* do not free switchboard */
	}
	write_unlock_bh(&rw_lock);
	up(&xt_usercontext_mutex);

	if (info) {
		DEBUGP("do_destroy(): vfree(info->switchboard->sessions=%p)\n", info->switchboard->sessions);
		vfree(info->switchboard->sessions);
		DEBUGP("do_destroy(): Deallocated %u bytes\n",
			(unsigned int) sizeof(*info->switchboard->sessions)*info->max_sess);

		DEBUGP("do_destroy(): vfree(info->switchboard=%p)\n", info->switchboard);
		vfree(info->switchboard);
        DEBUGP("do_destroy(): Deallocated %u bytes\n",
			(unsigned int) sizeof(*info->switchboard));
	}
}

static struct xt_target xt_rtpproxy_reg = {
	.name = "RTPPROXY",
	.family = AF_INET,
	.proto = IPPROTO_UDP,
	.target = do_target,
	.targetsize = XT_ALIGN(sizeof(struct xt_rtpproxy_target_info)),
	.checkentry = do_checkentry,
	.destroy = do_destroy,
	.hooks =  (1 << NF_IP_PRE_ROUTING) | (1 << NF_IP_POST_ROUTING) | (1 << NF_IP_LOCAL_OUT) | (1 << NF_IP_LOCAL_IN),
	.table = "mangle",
	.me = THIS_MODULE
};

static struct nf_sockopt_ops xt_rtpproxy_sockopts = {
	.pf = PF_INET,
	.set_optmin = XT_RTPPROXY_SO_BASE_CTL,
	.set_optmax = XT_RTPPROXY_SO_SET_MAX+1,
	.set = xt_do_set_ctl,
	.get_optmin = XT_RTPPROXY_SO_BASE_CTL,
	.get_optmax = XT_RTPPROXY_SO_GET_MAX+1,
	.get = xt_do_get_ctl
};


static int __init init(void)
{
	struct timeval stamp;
	PRINTK(KERN_INFO, "initializing\n");

	switchboard_reg = NULL;
	memset(&so_global, 0, sizeof(so_global));
	so_global.version = XT_RTPPROXY_VERSION_HEX;
	do_gettimeofday(&stamp);
	so_global.started = stamp.tv_sec;
	DEBUGP("init(): stamp.tv_sec=%ld\n", stamp.tv_sec);

	/* Register setsockopt */
	DEBUGP("init(): nf_register_sockopt(%p)\n", &xt_rtpproxy_sockopts);
	if (nf_register_sockopt(&xt_rtpproxy_sockopts) < 0) {
		PRINTK(KERN_ERR, "Can't register sockopts. Aborting\n");
		goto error_cleanup;
	}

	DEBUGP("init(): xt_register_target(%p)\n", &xt_rtpproxy_reg);
	if (xt_register_target(&xt_rtpproxy_reg)) {
		PRINTK(KERN_ERR, "Can't register target. Aborting\n");
		goto error_cleanup2;
	}
	return 0;

error_cleanup2:
	DEBUGP("init(): nf_unregister_sockopt(%p)\n", &xt_rtpproxy_sockopts);
	nf_unregister_sockopt(&xt_rtpproxy_sockopts);
error_cleanup:
	return -EINVAL;
}

static void __exit fini(void)
{
	PRINTK(KERN_INFO, "finalizing\n");
	DEBUGP("fini(): xt_unregister_target(%p)\n", &xt_rtpproxy_reg);
	xt_unregister_target(&xt_rtpproxy_reg);
	DEBUGP("fini(): nf_unregister_sockopt(%p)\n", &xt_rtpproxy_sockopts);
	nf_unregister_sockopt(&xt_rtpproxy_sockopts);
}

module_init(init);
module_exit(fini);

